/* FILE: display.cc                 -*-Mode: c++-*-
 *
 * C++ code for display windows
 *
 */

#include <stdlib.h>
#include <string.h>

#include "display.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

//////////////////////////////////////////////////////////////////////////
// PlotConfiguration structure
//
const ClassDoc PlotConfiguration::class_doc("PlotConfiguration",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.1","16-Jul-1998");
PlotConfiguration::PlotConfiguration()
  : displaystate(1),autosample(0),subsample(0.),size(1.),
    antialias(0),stipple(""),phase(0.),invert(0)
{
  colormap.Setup(256,"Red-Black-Blue");
}

PlotConfiguration& PlotConfiguration::operator=(const PlotConfiguration& pc)
{
  displaystate=pc.displaystate;
  autosample=pc.autosample;
  subsample=pc.subsample;
  size=pc.size;
  antialias=pc.antialias;
  stipple=pc.stipple;
  colorquantity=pc.colorquantity;
  phase=pc.phase;
  invert=pc.invert;
  colormap=pc.colormap;
  return *this;
}

PlotConfiguration::PlotConfiguration(const PlotConfiguration& pc)
  : displaystate(pc.displaystate),autosample(pc.autosample),
    subsample(pc.subsample),size(pc.size),antialias(pc.antialias),
    stipple(pc.stipple),
    colorquantity(pc.colorquantity),phase(pc.phase),invert(pc.invert),
    colormap(pc.colormap)
{}

//////////////////////////////////////////////////////////////////////////
// CoordinateSystem helper function
INT4m CoordinateAngleDifference(CoordinateSystem cs1,CoordinateSystem cs2,
				REAL4m &angle)
{ // Calculates the difference (in degrees, from 0 to 360) of cs1 - cs2, if
  // both are CS_DISPLAY_* coordinates, sets the difference in 'angle' and
  // returns 0;  If cs1 is not a CS_DISPLAY_* c.s., then the return value
  // is 1; if cs2 is not a CS_DISPLAY_* c.s., then the return value is 2.
  // In these latter two cases angle is 0.0 on return.
  REAL4m ang1,ang2;
  angle=0.0;
  switch(cs1) {
  case CS_DISPLAY_STANDARD: ang1=  0; break;
  case CS_DISPLAY_ROT_90:   ang1= 90; break;
  case CS_DISPLAY_ROT_180:  ang1=180; break;
  case CS_DISPLAY_ROT_270:  ang1=270; break;
  default: return 1;  // Illegal coordinates
    // break;
  }
  switch(cs2) {
  case CS_DISPLAY_STANDARD: ang2=  0; break;
  case CS_DISPLAY_ROT_90:   ang2= 90; break;
  case CS_DISPLAY_ROT_180:  ang2=180; break;
  case CS_DISPLAY_ROT_270:  ang2=270; break;
  default: return 2;  // Illegal coordinates
    // break;
  }
  angle=ang1-ang2;
  if(angle<0) angle+=360;
  return 0;
}

//////////////////////////////////////////////////////////////////////////
// Display frame class
const ClassDoc DisplayFrame::class_doc("DisplayFrame",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","14-Dec-1996");
// Recommended display window width, height and cellsize, in pixels:
const REAL4m DisplayFrame::default_arrow_cellsize=REAL4m(20);
const REAL4m DisplayFrame::default_pixel_cellsize=REAL4m(4);
const REAL4m DisplayFrame::default_arrow_ratio=REAL4m(0.85);
const REAL4m DisplayFrame::default_pixel_ratio=REAL4m(1.00);
const REAL4m DisplayFrame::max_display_object_dimension=REAL4m(1024.);

void DisplayFrame::SetMesh(Vf_Mesh *newmesh) {
  mesh=newmesh;
  arrow_valid=pixel_valid=boundary_valid=0;
  REAL8 minmag,maxmag;
  mesh->GetMagHints(minmag,maxmag);
  if(maxmag<=0.0) {
    // Bad hints; set from data
    mesh->GetValueMagSpan(minmag,maxmag);
  }
  // Insure value span is at least 6 orders of magnitude
  if(minmag >= maxmag * 1e-6) {
    minmag = maxmag * 1e-6;
  }
  mesh->SetMagHints(minmag,maxmag);
}

void DisplayFrame::SetPreferredArrowCellsize(REAL4m pixels)
{
  if(pixels>0. && preferred_arrow_cellsize!=pixels) {
    preferred_arrow_cellsize=pixels;
    if(arrow_config.autosample || arrow_config.subsample<0)
      arrow_valid=0;
  }
}

void DisplayFrame::SetPreferredPixelCellsize(REAL4m pixels)
{
  if(pixels>0. && preferred_pixel_cellsize!=pixels) {
    preferred_pixel_cellsize=pixels;
    if(pixel_config.autosample || pixel_config.subsample<0)
      pixel_valid=0;
  }
}

DisplayFrame::DisplayFrame()
{
  mesh=(Vf_Mesh *)NULL;
  tcl_interp=Oc_GlobalInterpreter();
  coords=CS_DISPLAY_STANDARD;
  arrow_valid=pixel_valid=boundary_valid=0;
  mesh_per_pixel=1.0;
  draw_boundary=1;
  boundary_width=1.0;
  background_color="#FFFFFF";  // White
  boundary_on_top=1;
  preferred_arrow_cellsize=default_arrow_cellsize;
  preferred_pixel_cellsize=default_pixel_cellsize;
}

void
DisplayFrame::ChangeCoordinates(CoordinateSystem old_coords,
                                CoordinateSystem new_coords,
                                Nb_List<Vf_DisplayVector> &display_list) const
{
#define MEMBERNAME "ChangeCoordinates(CS,CS,Nb_List<Vf_DisplayVector> &)"
  if(old_coords==new_coords) return;
  if(mesh_per_pixel<=0.0)
    FatalError(-1,STDDOC,"mesh_per_pixel = %g <= 0.0",mesh_per_pixel);
  REAL4m pixel_per_mesh=static_cast<REAL4m>(1.0/mesh_per_pixel);
  Nb_Vec3<REAL4> mesh_min,mesh_max;
  Nb_BoundingBox<REAL4> mesh_range;
  mesh->GetRange(mesh_range);
  mesh_range.GetExtremes(mesh_min,mesh_max);

  // Branch through cases
  if(old_coords==CS_CALCULATION_STANDARD
     && new_coords==CS_DISPLAY_STANDARD) {
    // Convert from mesh coordinates to display coordinates
    for(Vf_DisplayVector* display_vector=display_list.GetFirst();
	display_vector!=(Vf_DisplayVector *)NULL;
	display_vector=display_list.GetNext()) {
      display_vector->location.x -= mesh_min.x;
      display_vector->location.y  = mesh_max.y - display_vector->location.y;
      display_vector->location.z -= mesh_min.z;
      display_vector->location*=pixel_per_mesh;
      display_vector->value.y*=-1;
    }
  }
  else if(old_coords==CS_DISPLAY_STANDARD
     && new_coords==CS_CALCULATION_STANDARD) {
    // Convert from display coordinates to mesh coordinates
    for(Vf_DisplayVector* display_vector=display_list.GetFirst();
	display_vector!=(Vf_DisplayVector *)NULL;
	display_vector=display_list.GetNext()) {
      display_vector->location*=mesh_per_pixel;
      display_vector->location.x += mesh_min.x;
      display_vector->location.y  = mesh_max.y - display_vector->location.y;
      display_vector->location.z += mesh_min.z;
      display_vector->value.y*=-1;
    }
  }
  else if(old_coords==CS_CALCULATION_STANDARD
	  || new_coords==CS_CALCULATION_STANDARD) {
    if(Verbosity>=100) {
      NonFatalError(STDDOC,
	    "Requested coordinate transformation (%d->%d) not yet coded.\n"
	    "  Using 2-step transform (through CS_DISPLAY_STANDARD) instead.",
	    int(old_coords),int(new_coords));
    }
    ChangeCoordinates(old_coords,CS_DISPLAY_STANDARD,display_list);
    ChangeCoordinates(CS_DISPLAY_STANDARD,new_coords,display_list);
  }
  else {  // Flipping between 2 "DISPLAY" coordinate systems
    REAL4m dangle_diff;
    if(CoordinateAngleDifference(new_coords,old_coords,dangle_diff)!=0)
      FatalError(-1,STDDOC,"Non-comparable coordinate systems (%d & %d)",
		 old_coords,new_coords);
    INT4m angle_diff=INT4m(OC_ROUND(dangle_diff));
    Nb_Vec3<REAL4> display_max=mesh_max,vtemp;
    display_max-=mesh_min;
    display_max*=pixel_per_mesh;
    for(Vf_DisplayVector* display_vector=display_list.GetFirst();
	display_vector!=(Vf_DisplayVector *)NULL;
	display_vector=display_list.GetNext()) {
      Nb_Vec3<REAL4> *loc=&(display_vector->location);
      Nb_Vec3<REAL4> *dir=&(display_vector->value);
      switch(angle_diff) {
      case 90:
	vtemp = *loc;
	loc->x = vtemp.y;
	loc->y = display_max.y - vtemp.x;
	vtemp = *dir;
	dir->x =  vtemp.y;
	dir->y = -vtemp.x;
	break;
      case 180:
	loc->x = display_max.x - loc->x;
	loc->y = display_max.y - loc->y;
	dir->x *= -1;
	dir->y *= -1;
	break;
      case 270:
	vtemp = *loc;
	loc->x = display_max.x - vtemp.y;
	loc->y = vtemp.x;
	vtemp = *dir;
	dir->x = -vtemp.y;
	dir->y =  vtemp.x;
	break;
      default:
	FatalError(-1,STDDOC,"%s: Illegal angle_diff (%d)",
		   ErrProgramming,angle_diff);
	break;
      }
    }
  }
#undef MEMBERNAME
}

void
DisplayFrame::ChangeCoordinates(CoordinateSystem old_coords,
				CoordinateSystem new_coords,
				Nb_List< Nb_Vec3<REAL4> > &pt_list) const
{ // Same as previous, but for locations only
#define MEMBERNAME "ChangeCoordinates(CS,CS,Nb_List< Nb_Vec3<REAL4> > &)"
  if(old_coords==new_coords) return;
  if(mesh_per_pixel<=0.0)
    FatalError(-1,STDDOC,"mesh_per_pixel = %g <= 0.0",mesh_per_pixel);
  REAL4m pixel_per_mesh=static_cast<REAL4m>(1.0/mesh_per_pixel);
  Nb_Vec3<REAL4> mesh_min,mesh_max;
  Nb_BoundingBox<REAL4> mesh_range;
  mesh->GetRange(mesh_range);
  mesh_range.GetExtremes(mesh_min,mesh_max);

  // Branch through cases
  if(old_coords==CS_CALCULATION_STANDARD
     && new_coords==CS_DISPLAY_STANDARD) {
    // Convert from mesh coordinates to display coordinates
    for(Nb_Vec3<REAL4>* pt=pt_list.GetFirst();pt!=(Nb_Vec3<REAL4> *)NULL;
	pt=pt_list.GetNext()) {
      pt->x -= mesh_min.x;
      pt->y  = mesh_max.y - pt->y;
      pt->z -= mesh_min.z;
      (*pt) *= pixel_per_mesh;
    }
  }
  else if(old_coords==CS_DISPLAY_STANDARD
     && new_coords==CS_CALCULATION_STANDARD) {
    // Convert from display coordinates to mesh coordinates
    for(Nb_Vec3<REAL4>* pt=pt_list.GetFirst();pt!=(Nb_Vec3<REAL4> *)NULL;
	pt=pt_list.GetNext()) {
      (*pt) *= mesh_per_pixel;
      pt->x += mesh_min.x;
      pt->y  = mesh_max.y - pt->y;
      pt->z += mesh_min.z;
    }
  }
  else if(old_coords==CS_CALCULATION_STANDARD
	  || new_coords==CS_CALCULATION_STANDARD) {
    if(Verbosity>=100) {
      NonFatalError(STDDOC,
	    "Requested coordinate transformation (%d->%d) not yet coded.\n"
	    "  Using 2-step transform (through CS_DISPLAY_STANDARD) instead.",
	    int(old_coords),int(new_coords));
    }
    ChangeCoordinates(old_coords,CS_DISPLAY_STANDARD,pt_list);
    ChangeCoordinates(CS_DISPLAY_STANDARD,new_coords,pt_list);
  }
  else {  // Flipping between 2 "DISPLAY" coordinate systems
    REAL4m dangle_diff;
    REAL4 temp;
    if(CoordinateAngleDifference(new_coords,old_coords,dangle_diff)!=0)
      FatalError(-1,STDDOC,"Non-comparable coordinate systems (%d & %d)",
		 old_coords,new_coords);
    INT4m angle_diff=INT4m(OC_ROUND(dangle_diff));
    Nb_Vec3<REAL4> display_max=mesh_max;
    display_max-=mesh_min;
    display_max*=pixel_per_mesh;
    for(Nb_Vec3<REAL4>* pt=pt_list.GetFirst() ; pt!=(Nb_Vec3<REAL4> *)NULL;
	pt=pt_list.GetNext()) {
      switch(angle_diff) {
      case 90:
	temp=pt->x;
	pt->x=pt->y;
	pt->y=display_max.y-temp;
	break;
      case 180:
	pt->x = display_max.x - pt->x;
	pt->y = display_max.y - pt->y;
	break;
      case 270:
	temp=pt->x;
	pt->x = display_max.x - pt->y;
	pt->y = temp;
	break;
      default:
	FatalError(-1,STDDOC,"%s: Illegal angle_diff (%d)",
		   ErrProgramming,angle_diff);
	break;
      }
    }
  }
#undef MEMBERNAME
}

void DisplayFrame::CoordinatePointTransform
(CoordinateSystem old_coords,
 CoordinateSystem new_coords,
 Nb_Vec3<REAL4> &pt) const
{ // Same as ChangeCoordinates, but for one point only
#define MEMBERNAME "CoordinatePointTransform(CS,CS,Nb_Vec3<REAL4> &)"
  if(old_coords==new_coords) return;
  if(mesh_per_pixel<=0.0)
    FatalError(-1,STDDOC,"mesh_per_pixel = %g <= 0.0",mesh_per_pixel);
  REAL4m pixel_per_mesh=static_cast<REAL4m>(1.0/mesh_per_pixel);
  Nb_Vec3<REAL4> mesh_min,mesh_max;
  Nb_BoundingBox<REAL4> mesh_range;
  mesh->GetRange(mesh_range);
  mesh_range.GetExtremes(mesh_min,mesh_max);

  // Branch through cases
  if(old_coords==CS_CALCULATION_STANDARD
     && new_coords==CS_DISPLAY_STANDARD) {
    // Convert from mesh coordinates to display coordinates
    pt.x -= mesh_min.x;
    pt.y  = mesh_max.y - pt.y;
    pt.z -= mesh_min.z;
    pt *= pixel_per_mesh;
  }
  else if(old_coords==CS_DISPLAY_STANDARD
	  && new_coords==CS_CALCULATION_STANDARD) {
    // Convert from display coordinates to mesh coordinates
    pt *= mesh_per_pixel;
    pt.x += mesh_min.x;
    pt.y  = mesh_max.y - pt.y;
    pt.z += mesh_min.z;
  }
  else if(old_coords==CS_CALCULATION_STANDARD
	  || new_coords==CS_CALCULATION_STANDARD) {
    if(Verbosity>=100) {
      NonFatalError(STDDOC,
	    "Requested coordinate transformation (%d->%d) not yet coded.\n"
	    "  Using 2-step transform (through CS_DISPLAY_STANDARD) instead.",
	    int(old_coords),int(new_coords));
    }
    CoordinatePointTransform(old_coords,CS_DISPLAY_STANDARD,pt);
    CoordinatePointTransform(CS_DISPLAY_STANDARD,new_coords,pt);
  }
  else {  // Flipping between 2 "DISPLAY" coordinate systems
    REAL4m dangle_diff;
    REAL4 temp;
    if(CoordinateAngleDifference(new_coords,old_coords,dangle_diff)!=0)
      FatalError(-1,STDDOC,"Non-comparable coordinate systems (%d & %d)",
		 old_coords,new_coords);
    INT4m angle_diff=INT4m(OC_ROUND(dangle_diff));
    Nb_Vec3<REAL4> display_max=mesh_max;
    display_max-=mesh_min;
    display_max*=pixel_per_mesh;
    switch(angle_diff) {
    case 90:
      temp=pt.x;
      pt.x=pt.y;
      pt.y=display_max.y-temp;
      break;
    case 180:
      pt.x = display_max.x - pt.x;
      pt.y = display_max.y - pt.y;
      break;
    case 270:
      temp=pt.x;
      pt.x = display_max.x - pt.y;
      pt.y = temp;
      break;
    default:
      FatalError(-1,STDDOC,"%s: Illegal angle_diff (%d)",
		 ErrProgramming,angle_diff);
      break;
    }
  }
#undef MEMBERNAME
}

INT4m DisplayFrame::GetDisplayList(SampleInfo &si)
{
  // Use full-size mesh bounding box
  Nb_BoundingBox<REAL4> mesh_box;
  mesh->GetRange(mesh_box);
  return GetDisplayList(mesh_box,si);
}

INT4m
DisplayFrame::GetDisplayList(const Nb_BoundingBox<REAL4>& select_box,
			     SampleInfo &si)
{ // Fills display_list from *mesh using requested step size and
  // color quantity. Use step_request=0 to get _all_ points in mesh.
  //  Import select_box is in mesh units, i.e. CS_CALCULATION_STANDARD
  // coordinate system.
  //  Converts to current display_coords.
  // NOTE: Samples mesh in terms of *mesh cells*, as opposed to
  //       *mesh units*.
  // NOTE: mesh_per_pixel must be set before calling this routine.
#define MEMBERNAME "GetDisplayList"
  if(si.subsample<0)
    FatalError(-1,STDDOC,"%s: step_request=%f",ErrBadParam,si.subsample);
  if(mesh==(Vf_Mesh *)NULL) FatalError(-1,STDDOC,"Mesh pointer not set.");

  // Limit input to intersection of select_box and mesh bounding box
  Nb_BoundingBox<REAL4> mesh_box;
  mesh->GetRange(mesh_box);
  mesh_box.IntersectWith(select_box);

  // Determine stepping rates.  The code below is awkward; It is a
  // result of legacy code/design (or non-design, as the case may be).
  REAL4m mesh_cellsize=mesh->GetApproximateCellWidth();
  if(mesh_cellsize<=0.0) mesh_cellsize=1.0; // Safety
  Nb_Vec3<REAL4> mesh_celldim=mesh->GetApproximateCellDimensions();
  if(mesh_celldim.x<=0.0) mesh_celldim.x=1.0; //Safety
  if(mesh_celldim.y<=0.0) mesh_celldim.y=1.0; //Safety

  // Get list of vectors to display
  REAL4m xstep=mesh_celldim.x*si.subsample;
  REAL4m ystep=mesh_celldim.y*si.subsample;
  REAL4m zstep=0.;
  mesh->GetDisplayList(xstep,ystep,zstep,mesh_box,
		       si.colorquantity,si.phase,si.invert,
		       si.dvlist);
  xstep/=mesh_celldim.x;
  ystep/=mesh_celldim.y;
  si.subsample=OC_MAX(xstep,ystep);
  REAL4m grit=mesh->GetSubsampleGrit()*mesh_cellsize;
  if(0.<si.subsample && si.subsample<grit/2.) si.subsample=grit;

  ChangeCoordinates(CS_CALCULATION_STANDARD,coords,si.dvlist);

  return 0;
#undef MEMBERNAME

}


void DisplayFrame::SetPlotConfiguration(
                const PlotConfiguration& _arrow_config,
                const PlotConfiguration& _pixel_config)
{
  // Copy arrow configuration update arrow display list
  arrow_config=_arrow_config;
  arrow_valid=0;

  // Ditto for pixel configuration
  pixel_config=_pixel_config;
  pixel_valid=0;
}

REAL4m DisplayFrame::GetAutoSampleRate(REAL4m cellsize) const
{ // Returns a recommended subsample rate, for the current mesh at
  // current zoom.
  REAL8m step=cellsize*mesh_per_pixel;
  step=mesh->GetPreferredStepSize(static_cast<REAL4m>(step));
  step/=mesh->GetApproximateCellWidth();
  step=OC_ROUND(step*100.)/100.;  // Make "round" value
  if(step<=1.) step=0.;
  return static_cast<REAL4m>(step);
}

BOOL DisplayFrame::MakeTkColor
(const char* req_color,Nb_DString& tkcolor) const
{ // Converts "color" to Tk 8-bit format, "#XXXXXX", and stores
  // result in color.  If color cannot be interpreted, then
  // a warning message is printed, and tkcolor is unchanged.
  // Returns 1 on success, 0 on failure.
#define MEMBERNAME "MakeTkColor(const char*,Nb_DString&)"
  REAL8m red,green,blue;
  if(Nb_GetColor(req_color,red,green,blue)==0) {
    const char *fmt="Unrecognized color request %s";
    unsigned int tmpsiz = strlen(fmt)+strlen(req_color)+16;
    char *tmpbuf=new char[tmpsiz];
    Oc_Snprintf(tmpbuf,tmpsiz,fmt,req_color);
    NonFatalError(STDDOC,tmpbuf);
    delete[] tmpbuf;
    return 0;
  }

  UINT4 r = (UINT4)OC_ROUND(red*256.);    if(r>255) r=255;
  UINT4 g = (UINT4)OC_ROUND(green*256.);  if(g>255) g=255;
  UINT4 b = (UINT4)OC_ROUND(blue*256.);   if(b>255) b=255;

  char buf[16];
  Oc_Snprintf(buf,sizeof(buf),"#%02X%02X%02X",r,g,b);
  tkcolor=buf;

  return 1;
#undef MEMBERNAME
}

BOOL DisplayFrame::SetBackgroundColor(const char* req_color) {
#define MEMBERNAME "SetBackgroundColor(const char*)"
  return MakeTkColor(req_color,background_color);
#undef MEMBERNAME
}

BOOL DisplayFrame::SetBoundaryColor(const char* req_color) {
#define MEMBERNAME "SetBoundaryColor(const char*)"
  return MakeTkColor(req_color,boundary_color);
#undef MEMBERNAME
}

INT4m DisplayFrame::DrawPixel(const char *canvas_name,
			      const Vf_DisplayVector &dspvec,
			      const Nb_Vec3<REAL4> &pixel_dimens,
			      const char* stipple,
			      BOOL hide) const
{ // Based on code contributed by SLW
#define MEMBERNAME "DrawPixel"
  static char buf[1024];  // *Should* be plenty big enough.
  const Nb_Vec3<REAL4> *postemp=&(dspvec.location);

  // Respect max_display_object_dimension restriction
  REAL4m xsize = pixel_dimens.x;
  if(xsize>max_display_object_dimension) {
    xsize=max_display_object_dimension;
  }
  REAL4m ysize = pixel_dimens.y;
  if(ysize>max_display_object_dimension) {
    ysize=max_display_object_dimension;
  }


  // Color
  const char *color=pixel_config.Color(dspvec.shade);

  // NOTE: Timing tests on Linux 2.0.24 on a i486-DX2/66 indicate that
  //       a substantial time loss is involved inside sprintf doing %f
  //       conversions.  It is 6 or 7 times faster to use %d's instead.

  REAL4m x1,x2,y1,y2;
  int r1,r2,c1,c2;

  x1=(postemp->x)-xsize/2;  x2=x1+xsize;
  r1=int(floor(x1));        r2=int(ceil(x2));
  y1=(postemp->y)-ysize/2;  y2=y1+ysize;
  c1=int(floor(y1));        c2=int(ceil(y2));
  if(hide) {
    Oc_Snprintf(buf,sizeof(buf),
		"%s create rectangle %d %d %d %d -fill %s"
		" -outline {} -tag {pixels p%d}"
		" -stipple {%s} -state hidden",
		canvas_name,r1,c1,r2,c2,color,dspvec.zslice,stipple);
  } else {
    Oc_Snprintf(buf,sizeof(buf),
		"%s create rectangle %d %d %d %d -fill %s"
		" -outline {} -tag {pixels p%d} -stipple {%s}",
		canvas_name,r1,c1,r2,c2,color,dspvec.zslice,stipple);
  }

  return Tcl_Eval(tcl_interp,buf);
#undef MEMBERNAME
}

Nb_Vec3<REAL4> DisplayFrame::GetPixelDimensions() const {
  Nb_Vec3<REAL4> dimens = mesh->GetApproximateCellDimensions();
  dimens *= default_pixel_ratio * pixel_config.size / mesh_per_pixel;
  if(pixel_sample.subsample>1.)
    dimens *= pixel_sample.subsample;
  return dimens;
}

REAL8m DisplayFrame::GetArrowSize() const {
  REAL8m arrowsize=mesh->GetApproximateCellWidth()
    * default_arrow_ratio * arrow_config.size / mesh_per_pixel;
  if(arrow_sample.subsample>1.)
    arrowsize*=arrow_sample.subsample;
  return arrowsize;
}

INT4m DisplayFrame::DrawArrow(const char *canvas_name,
			      const Vf_DisplayVector &dspvec,
                              REAL4m arrowsize,BOOL hide) const
{
#define MEMBERNAME "DrawArrow"
#ifndef SQRT3
#define SQRT3 1.7320508075688772
#endif
  static char buf[4096];  // *Should* be plenty big enough.
  const Nb_Vec3<REAL4> *postemp;
  const Nb_Vec3<REAL4> *vtemp;

  REAL8m x1,y1,x2,y2,arrow_len,proj_len;
  INT4m r0,c0,r1,c1,r2,c2,ioff,a1,a2,a3;
  postemp=&(dspvec.location);
  vtemp=&(dspvec.value);
  REAL8m delx,dely,delz;

  // Line components, rescaled as necessary to respect
  // max_display_object_dimension
  delx = vtemp->x;
  dely = vtemp->y;
  delz = vtemp->z;
  REAL8m rescale = arrowsize;
  REAL8m maxdim = OC_MAX(fabs(delx),fabs(dely));
  if((rescale>1.0 && maxdim>max_display_object_dimension/rescale) ||
     (rescale<=1.0 && maxdim*rescale>max_display_object_dimension)  ) {
    rescale = max_display_object_dimension/maxdim;
  }
  delx *= rescale;
  dely *= rescale;
  if((rescale>1.0
      && fabs(delz)/16>max_display_object_dimension/rescale) ||
     (rescale<=1.0
      && rescale*fabs(delz)/16>max_display_object_dimension)  ) {
    // delz too big; truncate to 16*max_display_object_dimension to
    // protect against overflow.
    delz = 16*max_display_object_dimension;
  } else {
    delz *= rescale;
  }
  arrow_len = sqrt(delx*delx + dely*dely + delz*delz);
  if(arrow_len<REAL4_EPSILON) return TCL_OK;  // This is a zero vector
  proj_len=sqrt(delx*delx+dely*dely);
  // Arrowhead components
  a1=(int)OC_ROUND(proj_len*0.3);
  a2=(int)OC_ROUND(proj_len*0.4);
  a3=(int)OC_ROUND(arrow_len*0.15);
  if(a3<5) a1=a2;
  // Attributes
  int linewidth=(int)OC_ROUND(a3*0.6);
  if(linewidth<1) linewidth=1;

  const char* hidestr="";
  if(hide) {
    hidestr="-state hidden";
  }

  const char* pcolor1=arrow_config.Color(dspvec.shade);

  // NOTE: Timing tests on Linux 2.0.24 on a i486-DX2/66 indicate that
  //       a substantial time loss is involved inside sprintf doing
  //       %f conversions.  It is 6 or 7 times faster to use %d's instead
  //       of sprintf(buf,"%s create line %6.2f %6.2f %6.2f %6.2f"
  //	             "-arrow last -fill %s",canvas_name,x1,y1,x2,y2,
  //                 colormap.TkColor[color]);

  if(proj_len>=arrow_len*0.15) { // Standard arrow
    x1=(postemp->x)-delx/2;   x2=x1+delx;
    r1=int(OC_ROUND(x1));     r2=int(OC_ROUND(x2));
    y1=(postemp->y)-dely/2;   y2=y1+dely;
    c1=int(OC_ROUND(y1));     c2=int(OC_ROUND(y2));
    Oc_Snprintf(buf,sizeof(buf),
            "%s create line %d %d %d %d -tag {arrows a%d}"
            " -arrow last -arrowshape {%d %d %d}"
            " -width %d -fill %s %s",
            canvas_name,r1,c1,r2,c2,dspvec.zslice,
            a1,a2,a3,linewidth,pcolor1,hidestr);
  }  else { // Small projection with undefined direction.  Draw a dot
    REAL4m backshade=dspvec.shade+0.5f;  // Maybe there's a better way
    backshade-=static_cast<REAL4m>(floor(backshade));  /// to do this?
    const char* pcolor2=arrow_config.Color(backshade);
    r0=(int)OC_ROUND(postemp->x); c0=(int)OC_ROUND(postemp->y);
    if(vtemp->z<0) { // Arrow pointing into-plane
      ioff=(int)OC_ROUND(REAL4m(a3)/2.);
      r1=r0-ioff;  c1=c0-ioff;
      r2=r0+ioff;  c2=c0+ioff;
      int cmdlen =
	Oc_Snprintf(buf,sizeof(buf),
		    "%s create rect %d %d %d %d -tag {arrows a%d}"
		    " -fill %s -outline %s %s",
		    canvas_name,r1,c1,r2,c2,dspvec.zslice,
		    pcolor1,pcolor1,hidestr);
      if(a3>6) {
	ioff=(int)OC_ROUND(REAL4m(a3)/6.);
	r1=r0-ioff;  c1=c0-ioff;
	r2=r0+ioff;  c2=c0+ioff;
	Oc_Snprintf(buf+cmdlen,sizeof(buf)-cmdlen,
		    "; %s create rect %d %d %d %d -tag {arrows a%d}"
		    " -fill %s -width 0 %s",
		    canvas_name,r1,c1,r2,c2,dspvec.zslice,
		    pcolor2,hidestr);
      }
    } else { // Arrow pointing out-of-plane
      ioff=(int)OC_ROUND(REAL4m(a3)*(SQRT2/2.));
      r1=r0-ioff;  c1=c0-ioff;
      r2=r0+ioff;  c2=c0+ioff;
      int cmdlen =
	Oc_Snprintf(buf,sizeof(buf),
		    "%s create poly %d %d %d %d %d %d %d %d"
		    " -tag {arrows a%d} -fill %s -outline %s"
		    " -width %d %s",
		    canvas_name,r1,c0,r0,c2,r2,c0,r0,c1,dspvec.zslice,
		    pcolor1,pcolor1,linewidth/3,hidestr);
      // Note: The return quantity of sprintf is OS dependent.
      if(a3>6) {
	Oc_Snprintf(buf+cmdlen,sizeof(buf)-cmdlen,
		   "; %s create line %d %d %d %d -tag {arrows a%d}"
		   " -fill %s -width %d %s"
		   "; %s create line %d %d %d %d -tag {arrows a%d}"
		   " -fill %s -width %d %s",
		   canvas_name,r1,c0,r2,c0,
		   dspvec.zslice,pcolor2,linewidth/3,hidestr,
		   canvas_name,r0,c2,r0,c1,
		   dspvec.zslice,pcolor2,linewidth/3,hidestr);
      }
    }
  }
  return Tcl_Eval(tcl_interp,buf);
#undef MEMBERNAME
}

INT4m DisplayFrame::DrawLine(const char *canvas_name,
			     const Nb_Vec3<REAL4> &Pt1,
			     const Nb_Vec3<REAL4> &Pt2,
			     const char *tags)
{
  static char buf[1024];  // *Should* be plenty big enough.
  if(tags==NULL || tags[0]=='\0') {
    Oc_Snprintf(buf,sizeof(buf),
		"%s create line %6.2f %6.2f %6.2f %6.2f",
		canvas_name,Pt1.x,Pt1.y,Pt2.x,Pt2.y);
  } else {
    Oc_Snprintf(buf,sizeof(buf),
		"%s create line %6.2f %6.2f %6.2f %6.2f -tag {%s}",
		canvas_name,Pt1.x,Pt1.y,Pt2.x,Pt2.y,tags);
  }
  if( Tcl_Eval(tcl_interp,buf) != TCL_OK) return TCL_ERROR;
  return TCL_OK;
}

BOOL DisplayFrame::SetCoordinates(CoordinateSystem new_coords) 
{
    if(new_coords==coords) return 0;
    coords=new_coords;
    arrow_valid=pixel_valid=boundary_valid=0;
    return 1;
}
 
const char* DisplayFrame::GetVecColor(const Nb_Vec3<REAL4>& v) const
{
  // Access to the underlying Vf_Mesh::GetVecShade function.  See that
  // routine for details on shade determination.  The mapping from
  // shade to color depends on the state of arrow_config.displaystate
  // and pixel_config.displaystate.  Pixel config colorquantity and
  // colormap are used, unless pixels are turned off and arrows are
  // turned on, in which case arrrow config colorquantity and colormap
  // are used.
  REAL4m shade;
  const char* color="";
  if(!pixel_config.displaystate && arrow_config.displaystate) {
    // Use arrow_config data
    shade=mesh->GetVecShade(arrow_config.colorquantity,
			    arrow_config.phase,arrow_config.invert,
			    v);
    if(shade>=0.) color=arrow_config.Color(shade);
  } else {
    // Use pixel_config data
    shade=mesh->GetVecShade(pixel_config.colorquantity,
			    pixel_config.phase,pixel_config.invert,
			    v);
    if(shade>=0.) color=pixel_config.Color(shade);
  }
  return color;
}

INT4m DisplayFrame::Render(const char *canvas_name,BOOL hide)
{
#define MEMBERNAME "Render"
  INT4m error_code;

  // Check initialization
  if(tcl_interp==NULL)
    FatalError(-1,STDDOC,"tcl_interp not initialized.");
  if(mesh==NULL)
    FatalError(-1,STDDOC,"mesh not initialized.");

  // Check that current coordinate system is displayable.
  if(coords!=CS_DISPLAY_STANDARD &&
     coords!=CS_DISPLAY_ROT_90   &&
     coords!=CS_DISPLAY_ROT_180  &&
     coords!=CS_DISPLAY_ROT_270    ) {
    FatalError(-1,STDDOC,
	       "Can't draw display list in current (%d) coordinate system",
	       coords);
  }

  Tcl_SavedResult saved;
  Tcl_SaveResult(tcl_interp, &saved);
  // Set canvas background color
  const char *fmt="%s configure -background %s";
  char *buf=new char[strlen(fmt)+strlen(canvas_name)
		    +background_color.Length()+1];
  sprintf(buf,"%s configure -background %s",canvas_name,
	  background_color.GetStr());
  if((error_code=Tcl_Eval(tcl_interp,buf)) != TCL_OK)
    goto done;
  delete[] buf;

  // Draw pixel map, if requested
  if(pixel_config.displaystate) {
    if(!pixel_valid) {
      pixel_sample.subsample=pixel_config.subsample;
      if(pixel_config.autosample || pixel_sample.subsample<0.)
	pixel_sample.subsample=GetAutoSampleRate(preferred_pixel_cellsize);
      pixel_sample.colorquantity=pixel_config.colorquantity;
      pixel_sample.phase=pixel_config.phase;
      pixel_sample.invert=pixel_config.invert;
      GetDisplayList(pixel_sample);
      pixel_valid=1;
    }
    Nb_Vec3<REAL4> pixel_dimens = GetPixelDimensions();
    if(coords == CS_DISPLAY_ROT_90 || coords == CS_DISPLAY_ROT_270) {
      REAL4 temp=pixel_dimens.x;
      pixel_dimens.x = pixel_dimens.y;
      pixel_dimens.y = temp;
    }
    Vf_DisplayVector *dspvec;
    for(dspvec=pixel_sample.dvlist.GetFirst();
	dspvec!=(Vf_DisplayVector *)NULL;
	dspvec=pixel_sample.dvlist.GetNext()) {
      if((error_code=DrawPixel(canvas_name,*dspvec,pixel_dimens,
			       pixel_config.stipple,hide))
	 != TCL_OK) {
	goto done;
      }
    }
  }

  // Draw arrows, if requested
  if(arrow_config.displaystate) {
    if(!arrow_valid) {
      arrow_sample.subsample=arrow_config.subsample;
      if(arrow_config.autosample || arrow_sample.subsample<0.)
	arrow_sample.subsample=GetAutoSampleRate(preferred_arrow_cellsize);
      arrow_sample.colorquantity=arrow_config.colorquantity;
      arrow_sample.phase=arrow_config.phase;
      arrow_sample.invert=arrow_config.invert;
      GetDisplayList(arrow_sample);
      arrow_valid=1;
    }
    REAL4m arrowsize=static_cast<REAL4m>(GetArrowSize());
    Vf_DisplayVector *dspvec;
    for(dspvec=arrow_sample.dvlist.GetFirst();
	dspvec!=(Vf_DisplayVector *)NULL;
	dspvec=arrow_sample.dvlist.GetNext()) {
      if((error_code=DrawArrow(canvas_name,*dspvec,arrowsize,hide))
	 != TCL_OK) {
	goto done;
      }
    }
  }

  // Draw mesh edge box, if requested
  if(draw_boundary) {
    if(!boundary_valid) {
      mesh->GetBoundaryList(boundary_list);
      ChangeCoordinates(CS_CALCULATION_STANDARD,coords,boundary_list);
      boundary_valid=1;
    }
    Nb_Vec3<REAL4> *bdry1,*bdry2;
    if((bdry1=boundary_list.GetFirst())!=(Nb_Vec3<REAL4> *)NULL) {
      while((bdry2=boundary_list.GetNext())!=(Nb_Vec3<REAL4> *)NULL) {
	if((error_code=DrawLine(canvas_name,*bdry1,*bdry2,"bdry"))!=TCL_OK) {
	  goto done;
	}
	bdry1=bdry2;
      }
    }
  }

  error_code = TCL_OK;
done:
  Tcl_RestoreResult(tcl_interp, &saved);
  return error_code;
#undef MEMBERNAME
}

INT4m DisplayFrame::FillBitmap(BOOL reverse_layers,OommfBitmap& bitmap)
{
  Nb_BoundingBox<REAL4> display_box; // Empty box
  return FillBitmap(display_box,display_box,reverse_layers,bitmap);
}

INT4m
DisplayFrame::FillBitmap(const Nb_BoundingBox<REAL4>& arrow_box,
			 const Nb_BoundingBox<REAL4>& pixel_box,
			 BOOL reverse_layers,OommfBitmap& bitmap)
{ // Writes output into bitmap.  Imports arrow_box and pixel_box are in
  // mesh coordinates.  Empty boxes are interpreted as no restraint,
  // i.e., fill all points.  If reverse_layers is true, then the
  // rendering order of the layers is reversed, with the "top" (i.e.,
  // largest z) first, and the "bottom" last.
#define MEMBERNAME "FillBitmap"
  // Check initialization
  if(mesh==NULL)
    FatalError(-1,STDDOC,"mesh not initialized.");

  // Check that current coordinate system is displayable.
  if(coords!=CS_DISPLAY_STANDARD &&
     coords!=CS_DISPLAY_ROT_90   &&
     coords!=CS_DISPLAY_ROT_180  &&
     coords!=CS_DISPLAY_ROT_270    ) {
    FatalError(-1,STDDOC,
	       "Can't draw display list in current (%d) coordinate system",
	       coords);
  }

  // Clear and set background color
  bitmap.Erase(OommfPackedRGB((const char *)background_color));

  // Get pixel data, if needed
  BOOL clip_pixels=0;
  Nb_Vec3<REAL4> pixelbox_minpt,pixelbox_maxpt;
  Nb_Vec3<REAL4> pixel_dimens;
  if(pixel_config.displaystate && !pixel_valid) {
    pixel_sample.subsample=pixel_config.subsample;
    if(pixel_config.autosample || pixel_sample.subsample<0.)
      pixel_sample.subsample=GetAutoSampleRate(preferred_pixel_cellsize);
    pixel_sample.colorquantity=pixel_config.colorquantity;
    pixel_sample.phase=pixel_config.phase;
    pixel_sample.invert=pixel_config.invert;
    if(pixel_box.IsEmpty()) {
      GetDisplayList(pixel_sample);
    } else {
      GetDisplayList(pixel_box,pixel_sample);
    }
    Nb_BoundingBox<REAL4> bdry_box,data_box;  REAL8m data_pad;
    GetBoundingBoxes(bdry_box,data_box,data_pad);
    data_box.ExpandWith(bdry_box); // Ignore data_pad
    data_box.GetExtremes(pixelbox_minpt,pixelbox_maxpt);
    clip_pixels = !data_box.IsEmpty();
    pixel_dimens = GetPixelDimensions();
    if(coords == CS_DISPLAY_ROT_90 || coords == CS_DISPLAY_ROT_270) {
      REAL4 temp=pixel_dimens.x;
      pixel_dimens.x = pixel_dimens.y;
      pixel_dimens.y = temp;
    }
    pixel_valid=1;
  }

  // Get arrow data, if needed
  if(arrow_config.displaystate && !arrow_valid) {
    arrow_sample.subsample=arrow_config.subsample;
    if(arrow_config.autosample || arrow_sample.subsample<0.)
      arrow_sample.subsample=GetAutoSampleRate(preferred_arrow_cellsize);
    arrow_sample.colorquantity=arrow_config.colorquantity;
    arrow_sample.phase=arrow_config.phase;
    arrow_sample.invert=arrow_config.invert;
    if(arrow_box.IsEmpty()) {
      GetDisplayList(arrow_sample);
    } else {
      GetDisplayList(arrow_box,arrow_sample);
    }
    arrow_valid=1;
  }

  // Contruct pixel and arrow level slice arrays
  INT4m slice_count = mesh->GetZsliceCount();
  if(slice_count<1) {
    slice_count=1;  // Safety; Presumably this is an empty mesh
  }
  Nb_List<Vf_DisplayVector> *pixel_level_sample = NULL;
  Nb_List<Vf_DisplayVector> *arrow_level_sample = NULL;
  if(pixel_config.displaystate) {
    pixel_level_sample = new Nb_List<Vf_DisplayVector>[slice_count];
    Vf_DisplayVector *dspvec;
    for(dspvec=pixel_sample.dvlist.GetFirst();
	dspvec!=(Vf_DisplayVector *)NULL;
	dspvec=pixel_sample.dvlist.GetNext()) {
      INT4m level = dspvec->zslice;
      if(level>=0 && level<slice_count)
	pixel_level_sample[level].Append(*dspvec);
    }
  }
  if(arrow_config.displaystate) {
    arrow_level_sample = new Nb_List<Vf_DisplayVector>[slice_count];
    Vf_DisplayVector *dspvec;
    for(dspvec=arrow_sample.dvlist.GetFirst();
	dspvec!=(Vf_DisplayVector *)NULL;
	dspvec=arrow_sample.dvlist.GetNext()) {
      INT4m level = dspvec->zslice;
      if(level>=0 && level<slice_count)
	arrow_level_sample[level].Append(*dspvec);
    }
  }

  // Mesh edge box on bottom (i.e., draw first), if requested
  if(draw_boundary && !boundary_on_top) {
    if(!boundary_valid) {
      mesh->GetBoundaryList(boundary_list);
      ChangeCoordinates(CS_CALCULATION_STANDARD,coords,boundary_list);
      boundary_valid=1;
    }
    PlanePoint pt;
    Nb_List<PlanePoint> vlist;
    const Nb_Vec3<REAL4> *bdry;
    Nb_List_Index< Nb_Vec3<REAL4> > key;
    for(bdry=boundary_list.GetFirst(key);bdry!=NULL;
	bdry=boundary_list.GetNext(key)) {
      pt.x=OC_ROUND(bdry->x);
      pt.y=OC_ROUND(bdry->y);
      vlist.Append(pt);
    }
    int miter=1;
    int antialias=0;
    OommfPackedRGB color(boundary_color);
    bitmap.DrawPolyLine(vlist,boundary_width,color,
			miter,antialias);
  }

  // Loop through levels, drawing first pixels and then arrows.
  // If boundary is requested, that gets layered on last
  for(INT4m i=0;i<slice_count;i++) {

    INT4m level=i;
    if(reverse_layers) level = slice_count - 1 - i;
    
    // Pixels
    if(pixel_config.displaystate) {
      Vf_DisplayVector *dspvec;
      Nb_List<Vf_DisplayVector>& pixlist = pixel_level_sample[level];
      for(dspvec=pixlist.GetFirst();
	  dspvec!=(Vf_DisplayVector *)NULL;
	  dspvec=pixlist.GetNext()) {
	if(dspvec->value.MagSq()<REAL4_EPSILON) continue;
	// This is a zero vector; skip.
	REAL4m x1,x2,y1,y2;
	x1=(dspvec->location.x)-pixel_dimens.x/2;  x2=x1+pixel_dimens.x-1;
	y1=(dspvec->location.y)-pixel_dimens.y/2;  y2=y1+pixel_dimens.y-1;
	/// Unlike the Tk canvas widget, OommfBitmap DrawFilledRectangle
	/// includes both corners.
	if(clip_pixels &&
	   x1<pixelbox_maxpt.x && pixelbox_minpt.x<x2
	   && y1<pixelbox_maxpt.y && pixelbox_minpt.y<y2) {
	  // Don't clip if whole draw rectangle is outside mesh range
	  if(x1<pixelbox_minpt.x) x1=pixelbox_minpt.x;
	  if(x2>pixelbox_maxpt.x) x2=pixelbox_maxpt.x;
	  if(y1<pixelbox_minpt.y) y1=pixelbox_minpt.y;
	  if(y2>pixelbox_maxpt.y) y2=pixelbox_maxpt.y;
	}
	bitmap.DrawFilledRectangle(x1,y1,x2,y2,
				   pixel_config.PackedColor(dspvec->shade),
				   pixel_config.stipple);
      }
    }

    // Arrows
    if(arrow_config.displaystate) {
      REAL8m arrowsize=GetArrowSize();
      Vf_DisplayVector *dspvec;
      Nb_List<Vf_DisplayVector>& arrlist = arrow_level_sample[level];
      for(dspvec=arrlist.GetFirst();
	  dspvec!=(Vf_DisplayVector *)NULL;
	  dspvec=arrlist.GetNext()) {
	REAL8m arrmag=dspvec->value.MagSq();
	if(arrmag<REAL4_EPSILON) continue; // This is a zero vector; skip.
	arrmag=sqrt(arrmag);
	REAL8m arrmaginv=1./arrmag;
	REAL8m xcos=dspvec->value.x*arrmaginv;
	REAL8m ycos=dspvec->value.y*arrmaginv;
	REAL8m zcos=dspvec->value.z*arrmaginv;
	REAL8m xpos=OC_ROUND(dspvec->location.x); // Round for nicer
	REAL8m ypos=OC_ROUND(dspvec->location.y); // display
	if(fabs(zcos)<0.985) {
	  // Angle with xy-plane smaller than 80 degrees => draw arrow
	  bitmap.DrawFilledArrow(xpos,ypos,
				 arrowsize*arrmag,xcos,ycos,zcos,
				 arrow_config.PackedColor(dspvec->shade),
				 arrow_config.antialias);
	} else {
	  REAL4m foreshade=dspvec->shade;
	  REAL4m backshade=foreshade+0.5f;   // Maybe there's a better way
	  backshade-=static_cast<REAL4m>(floor(backshade)); // to do this?
	  if(zcos>0) {
	    // Vector is almost complete out-of-plane.  Draw diamond-plus
	    bitmap.DrawDiamondWithCross(xpos,ypos,arrowsize*arrmag,
					arrow_config.PackedColor(foreshade),
					arrow_config.PackedColor(backshade));
	  } else {
	    // Vector is almost complete in-to-plane.  Draw dotted-square
	    bitmap.DrawSquareWithDot(xpos,ypos,arrowsize*arrmag,
				     arrow_config.PackedColor(foreshade),
				     arrow_config.PackedColor(backshade));
	  }
	}
      }
    }

  }

  // Mesh edge box on top (i.e., draw last), if requested
  if(draw_boundary && boundary_on_top) {
    if(!boundary_valid) {
      mesh->GetBoundaryList(boundary_list);
      ChangeCoordinates(CS_CALCULATION_STANDARD,coords,boundary_list);
      boundary_valid=1;
    }
    PlanePoint pt;
    Nb_List<PlanePoint> vlist;
    const Nb_Vec3<REAL4> *bdry;
    Nb_List_Index< Nb_Vec3<REAL4> > key;
    for(bdry=boundary_list.GetFirst(key);bdry!=NULL;
	bdry=boundary_list.GetNext(key)) {
      pt.x=OC_ROUND(bdry->x);
      pt.y=OC_ROUND(bdry->y);
      vlist.Append(pt);
    }
    int miter=1;
    int antialias=0;
    OommfPackedRGB color(boundary_color);
    bitmap.DrawPolyLine(vlist,boundary_width,color,
			miter,antialias);
  }

  if(arrow_level_sample!=NULL) delete[] arrow_level_sample;
  if(pixel_level_sample!=NULL) delete[] pixel_level_sample;

  return TCL_OK;
#undef MEMBERNAME
}

// NOTE: "zoom" in the next several routines is in units of "pixels per
//       approx cell size".
REAL4m DisplayFrame::GetZoom(void) const
{
  REAL4m mesh_per_cell=mesh->GetApproximateCellWidth();
  if(mesh_per_cell<=0.) mesh_per_cell=1.0;  // Safety
  return mesh_per_cell/mesh_per_pixel;
}

REAL4m DisplayFrame::SetZoom(REAL4m zoom)
{ // Returns zoom value
  REAL4m mesh_per_cell=mesh->GetApproximateCellWidth();
  if(mesh_per_cell<=0.) mesh_per_cell=1.0;  // Safety

  if(REAL4_EPSILON<zoom && zoom<1/REAL4_EPSILON) {
    mesh_per_pixel=mesh_per_cell/zoom;
    arrow_valid=pixel_valid=boundary_valid=0;
  }

  return mesh_per_cell/mesh_per_pixel;
}

REAL4m DisplayFrame::SetZoom(REAL4m width,REAL4m height)
{ // Adjusts mesh_per_pixel scaling factor so that entire
  // mesh will just fit inside width and height.  Marks display
  // lists as out-of-date.  New zoom value is returned.
  // NOTE: Setting either width or height <=0 will cause
  //  that extent to be ignored.  If both are <=0, then
  //  no change is made.

  REAL8m new_mesh_per_pixel=mesh_per_pixel;

  if(new_mesh_per_pixel<=0.) new_mesh_per_pixel=1.0; // Just to be safe.

  Nb_BoundingBox<REAL4> bdry,data;
  REAL8m data_pad;

  GetBoundingBoxes(bdry,data,data_pad);

  REAL8m xscale= -1.0, yscale= -1.0;
  if(width>0.) {
    if(!bdry.IsEmpty()) xscale=bdry.GetWidth()/width;
    if(width>2*data_pad)
      xscale=OC_MAX(xscale,data.GetWidth()/(width-2*data_pad));
  }
  if(height>0.) {
    if(!bdry.IsEmpty()) yscale=bdry.GetHeight()/height;
    if(height>2*data_pad)
      yscale=OC_MAX(yscale,data.GetHeight()/(height-2*data_pad));
  }
  REAL8m scale=OC_MAX(xscale,yscale);

  if(scale>0.) new_mesh_per_pixel*=scale;

  REAL4m mesh_per_cell=mesh->GetApproximateCellWidth();
  if(mesh_per_cell<=0.) mesh_per_cell=1.0;  // Safety
  REAL8m zoom=mesh_per_cell/new_mesh_per_pixel; // User value

  // Make zoom a "round" (%.2f) value
  if(zoom>0.01) {
    zoom=OC_ROUND(zoom*100.)/100.;
    /// Arguably this should be "floor" instead of "round," but the
    /// latter appears more stable with respect to dependence of
    /// outcome on initial value of mesh_per_pixel.  Moreover, there
    /// is now (June-1998) some slack in the calculation of whether
    /// scroll bars are required.
    new_mesh_per_pixel=static_cast<REAL8m>(mesh_per_cell)/zoom;
  }

  if(mesh_per_pixel!=static_cast<REAL4m>(new_mesh_per_pixel)) {
    mesh_per_pixel=static_cast<REAL4m>(new_mesh_per_pixel);
    MarkAllSampleListsInvalid();
  }

  return static_cast<REAL4m>(zoom);
}

void DisplayFrame::GetBoundingBoxes(Nb_BoundingBox<REAL4> &bdry,
				    Nb_BoundingBox<REAL4> &data,
				    REAL8m &data_pad) const
{ // Returns bounding boxes for display, in pixels (at current zoom):
  //   'bdry' is the bounding box for the boundary
  //   'data' is the bounding box for the data
  //   'data_pad' is the margin (in pixels) that needs to be added
  //     to the data bounding box.
  // The first two scale with mesh_per_pixel (i.e., current zoom).  The
  // last is independent of the zoom levels.
  // NB: Either of the returning boxes, in particular bdry, may
  //  be an "empty" box, in which case max<min.

  Nb_Vec3<REAL4> min,max;

  // Check for an empty mesh
  Nb_BoundingBox<REAL4> bbox;
  mesh->GetRange(bbox);
  if(bbox.IsEmpty()) {
    data.Reset();
    bdry.Reset();
    data_pad=0.;
    return;
  }

  // Calculate bounding box for data points
  mesh->GetDataRange(data);
  data.GetExtremes(min,max);
  Nb_List< Nb_Vec3<REAL4> > datalist;
  datalist.SetFirst(min); datalist.Append(max);
  ChangeCoordinates(CS_CALCULATION_STANDARD,coords,datalist);
  min= *(datalist.GetFirst()); max= *(datalist.GetNext());
  data.SortAndSet(min,max);

  data_pad=OC_MAX(preferred_arrow_cellsize,preferred_pixel_cellsize)/2.;

  // Create boundary bounding box by expanding for each point in the
  // boundary list, one at a time.  Use supplemental list (instead of
  // boundary_list) so we don't have to worry about updating boundary_list
  // and loosing the 'const' property of this method.
  Nb_List< Nb_Vec3<REAL4> > bdrylist;
  mesh->GetBoundaryList(bdrylist);
  ChangeCoordinates(CS_CALCULATION_STANDARD,coords,bdrylist);
  bdry.Reset();
  for(Nb_Vec3<REAL4>* bp=bdrylist.GetFirst() ; bp!=NULL ;
      bp=bdrylist.GetNext()) {
    bdry.ExpandWith(*bp);
  }
}

Nb_BoundingBox<REAL4> DisplayFrame::GetDisplayBox() const
{ // Returns bounding box for display, in pixels (at current zoom)
  // Box extents will be (0,0,0) x (0,0,0) in empty case.
  Nb_Vec3<REAL4> min,max;
  Nb_BoundingBox<REAL4> bdry,data,total_box;
  REAL8m wide_data_pad;
  GetBoundingBoxes(bdry,data,wide_data_pad);
  REAL4m data_pad = static_cast<REAL4>(wide_data_pad);

  // Add padding to data box.
  if(!data.IsEmpty()) {
    data.GetExtremes(min,max);
    min.x-=data_pad;    max.x+=data_pad;
    min.y-=data_pad;    max.y+=data_pad;
    // Padding z with data_pad doesn't make too much sense,
    // since data_pad is derived from in-plane cell sizes.
    // So instead just pad z with z component of base cell
    // dimension.
    Nb_Vec3<REAL4> celldim = mesh->GetApproximateCellDimensions();
    REAL4m half_zstep
      = static_cast<REAL4m>(celldim.z/(2.0*mesh_per_pixel));
    min.z -= half_zstep;    max.z += half_zstep;
    total_box.Set(min,max);
  } else {
    total_box.Reset();  // Just for clarity
  }

  // Expand as necessary to include boundary.
  total_box.ExpandWith(bdry);
  if(!data.IsEmpty()) {
    // Don't change z-depth for sake of boundary
    Nb_Vec3<REAL4> temp_min,temp_max;
    total_box.GetExtremes(temp_min,temp_max);
    temp_min.z=min.z;
    temp_max.z=max.z;
    total_box.Set(temp_min,temp_max);
  }

  // Special handling for empty case.
  if(total_box.IsEmpty()) {
    Nb_Vec3<REAL4> mesh_min(0,0,0),mesh_max(0,0,0);
    total_box.Set(mesh_min,mesh_max);
  }

  return total_box;
}

REAL8m DisplayFrame::GetValueScaling() const {
  REAL8m minmag,maxmag;
  mesh->GetMagHints(minmag,maxmag);
  REAL8m scale = mesh->GetDisplayValueScale()*maxmag;
  if(scale<=0.0) scale=1.0;  // Provide safe initial value.
  return scale;
}

INT4m DisplayFrame::SetValueScaling(REAL8m newscaling) {
  if(newscaling<=0.0) return 1;
  INT4m errcode=1;
  REAL8m minmag,maxmag;
  mesh->GetMagHints(minmag,maxmag);
  if(fabs(newscaling)<FLT_MAX*fabs(maxmag)) {
    errcode=0;
    newscaling/=maxmag;
    if(newscaling!=mesh->GetDisplayValueScale()) {
      errcode=mesh->SetDisplayValueScale(newscaling);
      arrow_valid=pixel_valid=0;
    }
  }
  return errcode;
}
