#!/bin/sh
# FILE: odtcols.tcl
#
# This Tcl script takes an .odt (OOMMF Data Table) file on stdin, and
# prints to stdout the requested columns, which are specified numerically
# starting with 1.
#
# Example Usage: tclsh odtcols.tcl 6 1 3 10 <infile.odt >outfile.odt
#
#    v--Edit here if necessary \
exec tclsh "$0" ${1+"$@"}

# Check for usage request
if {[llength $argv]<1 || [lsearch -regexp $argv {-+h.*}]>=0} {
    puts stderr {Usage: tclsh odtcols.tcl [-h] [-s]\
	    [-w colwidth] [-f format]\
	    col [col ...] <infile >outfile}
    exit 1
}

# Check for file summary request
set show_summary 0
set cmdindex [lsearch -regexp $argv {-+s.*}]
if {$cmdindex>=0} {
    set show_summary 1
    set argv [lreplace $argv $cmdindex $cmdindex]
}

# Tweak stdio streams
fconfigure stdin -buffering full -buffersize 30000
fconfigure stdout -buffering full  -buffersize 30000

# Pull column width requests out from $argv
set colwidth 15
while {[set width_index [lsearch -regexp $argv {-+w.*}]]>=0} {
    set colwidth [lindex $argv [expr $width_index+1]]
    set argv [lreplace $argv $width_index [expr $width_index+1]]
}

# Pull data format requests out from $argv
set data_format "%-${colwidth}s"
while {[set fmt_index [lsearch -regexp $argv {-+f.*}]]>=0} {
    set data_format [lindex $argv [expr $fmt_index+1]]
    set argv [lreplace $argv $fmt_index [expr $fmt_index+1]]
}

# Get column request from $argv
set colspecs $argv
if {[llength $colspecs]==0} {set colspecs {*}} ;# Default is full list

# Proc to substitute column glob-style selection expressions
# to numerical indices.
proc GenerateColumnIndexList { subset full } {
    # Do case insensitive matching.  It might be preferable to
    # match uppercase in subset only against uppercase in full,
    # but to match lowercase in subset against either uppercase
    # or lowercase in full.
    set subset [string tolower $subset]
    set full [string tolower $full]
    set indices {}
    foreach id $subset {
	if {[regexp -- {^[0-9]+$} $id] && $id<[expr {[llength $full]}]} {
	    # id is a number; assume this is a direct index request
	    lappend indices $id
	} else {
	    # Otherwise, try a glob-style match
	    for {set i 0} {$i<[llength $full]} {incr i} {
		if {[string match $id [lindex $full $i]]} {
		    lappend indices $i
		}
	    }
	}
    }
    return $indices
}

# Header list request
if {$show_summary} {
    proc DumpHeaders {column_subset column_list units_list \
	    table_number table_title row_count} {
	global colwidth
	if {$table_number>1} { puts "" } ;# Table separator
	if {[string match {} $table_title]} {
	    puts [format "*** Table %2d ***" $table_number]
	} else {
	    puts [format "*** Table %2d: $table_title ***" $table_number]
	}
	if {[llength column_list]<1} {
	    if {$row_count>0} {
		puts "WARNING: No column header line detected."
	    }
	} else {
	    puts [format "Column %-${colwidth}s %s" Label Units]
	    set index 0
	    if {[llength $column_subset]==0} {
		# List all columns
		foreach c $column_list u $units_list {
		    set c [list $c]
		    set u [list $u]
		    puts [format "%4d:  %-${colwidth}s %s" $index $c $u]
		    incr index
		}
	    } else {
		# Dump column subset
		foreach index $column_subset {
		    set c [list [lindex $column_list $index]]
		    set u [list [lindex $units_list $index]]
		    puts [format "%4d:  %-${colwidth}s %s" $index $c $u]
		    incr index
		}
		puts  "Columns:   [llength $column_list]"
	    }
	    puts      "Data rows: $row_count"		  
	}
    }

    set row_count 0
    set table_count 1
    set table_label {}
    set column_list {}
    set units_list {}
    set columns {}
    set line {}
    while {[gets stdin newline]>=0} {
	# Is new line a continuation of previous line?
	if {![string match {} $line]} {
	    # Yes: strip any leading #'s
	    set newline [string trimleft $newline "#"]
	}
	# Is new line continued on next read?
	if {[regexp -- {^(.*)\\$} $newline dummy match]} {
	    # Yes: append w/o trailing slash to current line
	    # and immediately read next line.
	    append line $match
	    continue
	}
	# Append new line to current line
	append line $newline

	# Replace tabs (if any) with spaces.  This simplifies regexp
	# processing below
	regsub -all -- "\t" $line " " line

	# Process current line
	if {[string match "#" [string index $line 0]]} {
	    # Comment line
	    if {[regexp -nocase -- {^# +columns} $line]} {
		# Column header line
		set column_list [concat $column_list [lrange $line 2 end]]
		set columns [GenerateColumnIndexList $colspecs $column_list]
	    } elseif {[regexp -nocase -- {^# +units} $line]} {
		set units_list [concat $units_list [lrange $line 2 end]]
	    } elseif {[regexp -nocase -- {^# +table +start} $line]} {
		# New table.  Copy title string, if any
		if {[regexp {^[^:]*: *(.*[^ ]) *$} $line dummy new_title]} {
		    set table_label $new_title
		}
	    } elseif {[regexp -nocase -- {^# +title} $line]} {
		# New title
		if {[regexp {^[^:]*: *(.*[^ ]) *$} $line dummy new_title]} {
		    set table_label $new_title
		}
	    } elseif {[regexp -nocase -- {^# +table +end} $line]} {
		# End of current table
		DumpHeaders $columns $column_list $units_list \
			$table_count $table_label $row_count
		set row_count 0
		set table_label {}
		set column_list {}
		set units_list {}
		incr table_count
	    } else {
		# Plain or otherwise unprocessed comment line; skip
	    }
	} else {
	    # Data line
	    incr row_count
	}
	set line {}
    }
    if {![string match {} $column_list] || $row_count>0} {
	DumpHeaders $columns $column_list $units_list \
		$table_count $table_label $row_count
	incr table_count
    }
    if {$table_count>2} {
	puts [format "\nTotal number of tables: %2d" [expr $table_count-1]]
    }
    exit 0
}

# Collimated output routines
proc ColumnDump { list {leader " "}} {
    global colwidth default_format
    puts -nonewline $leader
    foreach elt $list {
	puts -nonewline [format " %-${colwidth}s" [list $elt]]
    }
    puts {}
}

proc ColumnDataDump { list {leader " "}} {
    global colwidth data_format
    puts -nonewline $leader
    foreach elt $list {
	if {[catch {set entry [format $data_format $elt]}]} {
	    # Bad data.  Fill field with stars
	    puts stderr "ERROR Data doesn't match format: \"$elt\""
	    set entry {}
	    for {set i 0} {$i<$colwidth} {incr i} {append entry "*"}
	}
	puts -nonewline [format " %-${colwidth}s" $entry]
    }
    puts {}
}


# Loop through input lines
set line {}
set columns {}
while {[gets stdin newline]>=0} {
    # Is new line a continuation of previous line?
    if {![string match {} $line]} {
	# Yes: strip any leading #'s
	set newline [string trimleft $newline "#"]
    }
    # Is new line continued on next read?
    if {[regexp -- {^(.*)\\$} $newline dummy match]} {
	# Yes: append w/o trailing slash to current line
	# and immediately read next line.
	append line $match
	continue
    }
    # Append new line to current line
    append line $newline

    # Remove leading and trailing whitespace
    set line [string trim $line]

    # Replace tabs (if any) with spaces.  This simplifies regexp
    # processing below
    regsub -all -- "\t" $line " " line

    # Process current line
    if {[string match {} $line]} {
	# Blank line
	puts {}
    } elseif {[string match "#" [string index $line 0]]} {
	# Comment line
	if {[regexp -nocase -- {^# +columns} $line]} {
	    # Column header line
	    set columns [GenerateColumnIndexList $colspecs \
		    [lrange $line 2 end]]
	    set outlist {}
	    foreach i $columns {
		lappend outlist [lindex $line [expr $i+2]]
	    }
	    puts "# Columns: \\"
	    ColumnDump $outlist "#"
	} elseif {[regexp -nocase -- {^# +units} $line]} {
	    set outlist {}
	    foreach i $columns {
		lappend outlist [lindex $line [expr $i+2]]
	    }
	    puts "# Units: \\"
	    ColumnDump $outlist "#"
	} else {
	    # Plain comment line
	    puts $line
	}
    } else {
	# Data line
	set outlist {}
	foreach i $columns {
	    lappend outlist [lindex $line $i]
	}
	ColumnDataDump $outlist " "
    }
    set line {}
}

