/* FILE: director.cc                 -*-Mode: c++-*-
 *
 * 3D solver director class.
 *
 */

#undef REPORT_CONSTRUCT_TIMES  // #define this to get timing report on stderr

#include <algorithm>
#include <vector>
#include <string>
#include "oc.h"
#include "nb.h"
#include "director.h"
#include "driver.h"
#include "simstate.h"
#include "ext.h"
#include "energy.h"
#include "util.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "threevector.h"

#ifdef REPORT_CONSTRUCT_TIMES
#include "nb.h"
#endif

/* End includes */

// The following two routines make sure that all Oxs_Director
// instances are destroyed as part of a Tcl_Exit().
void Oxs_Director::ExitProc(ClientData clientData)
{
    Oxs_Director *d = (Oxs_Director *) clientData;
    delete d;
}

Oxs_Director::Oxs_Director(Tcl_Interp *i) 
  : problem_count(0),problem_id(0),
    interp(i),
    random_seed_set(0), random_seed(0),
    driver(NULL),
    simulation_state_reserve_count(0)
{
  // Register cleanup handler
  Tcl_CreateExitHandler((Tcl_ExitProc *) ExitProc, (ClientData) this);

  // Make sure tcl_precision is full
  Tcl_Eval(interp,Oc_AutoBuf("set tcl_precision 17"));
}

int Oxs_Director::ProbReset()
{ // Sets micromag problem to initial state.
  // Returns TCL_OK on success, TCL_ERROR otherwise.
  int errcode = TCL_OK;

  // If seed has been specified, then reset
  // random number generators.
  if(random_seed_set) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),"%s RandomSeed %lu",
		mif_object.c_str(),
		static_cast<unsigned long>(random_seed));
    Tcl_SavedResult saved;
    Tcl_SaveResult(interp,&saved);
    errcode = Tcl_Eval(interp,buf);
    if(errcode!=TCL_OK) {
      Tcl_DiscardResult(&saved);
      return errcode;
    }
    Tcl_RestoreResult(interp,&saved);
  }

  // Run Init() on all Oxs_Ext's
  vector<Oxs_Ext*>::iterator it = ext_obj.begin();
  try {
    while(it!=ext_obj.end() && errcode==TCL_OK) {
      if(!((*it)->Init())) errcode = TCL_ERROR;
      it++;
    }
  } catch (Oxs_Ext::Error& err) {
    Tcl_AppendResult(interp,
		     "Oxs_Ext error"
		     " in Init() function of ",
		     (*it)->InstanceName(),
		     " --- ",
		     err.c_str(),(char *)NULL);
    errcode = TCL_ERROR;
  } catch (Oc_Exception& err) {
    Tcl_AppendResult(interp,
		     "Oxs_Ext error"
		     " in Init() function of ",
		     (*it)->InstanceName()," --- ",
		     err.MessageType()," --- ",
		     err.MessageText().c_str(),
		     (char *)NULL);
    errcode = TCL_ERROR;
  }

  return errcode;
}

int Oxs_Director::ProbRun(vector<OxsRunEvent>& results,
			  string& errmsg)
{ // On success, fills results with event list and returns 0.
  // Otherwise, return is >0, and errmsg is filled with a message
  // describing the error.

  int errcode=0;

  if(driver==NULL) {
    errmsg = string("Problem not initialized (no driver identified)");
    errcode=1;
  } else {
    try {
      driver->Run(results);
    } catch (Oxs_Ext::Error& err) {
      errcode=2;
      errmsg = string("Oxs_Ext error inside Run() function of driver ")
	+ string(driver->InstanceName())
	+ string(" --- ") + string(err);
    } catch (Oc_Exception& err) {
      errcode=3;
      errmsg = string("Error inside Run() function of driver ")
	+ string(driver->InstanceName()) + string(" --- ")
	+ string(err.MessageType()) + string(" --- ")
	+ err.MessageText();
    } catch (EXCEPTION& err) {
      errcode = 4;
      if(errmsg.length()>0) errmsg += string(" --- ");
      errmsg += string("Library error thrown from inside"
		       " Run() function of driver ")
	+ string(driver->InstanceName()) + string(" --- ")
	+ string(err.what());
#ifndef NDEBUG
    } catch (...) {
      errcode=5;
      errmsg = string("Unrecognized error thrown from inside"
		      " Run() function of driver ")
	+ string(driver->InstanceName());
#endif // NDEBUG
    }
  }

  return errcode;
}

int Oxs_Director::SetStage(int requestedStage, string& result)
{
  if(driver==NULL) {
    result = "Problem not initialized (no driver identified)";
    return TCL_ERROR;
  }

  if(requestedStage<0) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"Invalid stage request: %d; stage must be non-negative.",
		requestedStage);
    result = buf;
    return TCL_ERROR;
  }
  // Should we also check for requestedStage too large?

  int errcode = TCL_ERROR;
  try {
    driver->SetStage(static_cast<UINT4m>(requestedStage));
    errcode = TCL_OK;
  } catch (Oxs_Ext::Error& err) {
    result = string("Oxs_Ext error inside SetStage function of driver ")
      + string(driver->InstanceName())
      + string(" --- ") + string(err);
  } catch (Oc_Exception& err) {
    result = string("Error inside SetStage function of driver ")
      + string(driver->InstanceName()) + string(" --- ")
      + string(err.MessageType()) + string(" --- ")
      + err.MessageText();
#ifndef NDEBUG
  } catch (...) {
    result = string("Unrecognized error thrown from inside"
		    " SetStage function of driver ")
      + string(driver->InstanceName());
#endif // NDEBUG
  }
  return errcode;
}

int Oxs_Director::GetStage(int& stage, string& result)
{
  if(driver==NULL) {
    result = "Problem not initialized (no driver identified)";
    return TCL_ERROR;
  }

  int errcode = TCL_ERROR;
  try {
    stage = static_cast<int>(driver->GetStage());
    errcode = TCL_OK;
  } catch (Oxs_Ext::Error& err) {
    result = string("Oxs_Ext error inside GetStage function of driver ")
      + string(driver->InstanceName())
      + string(" --- ") + string(err);
  } catch (Oc_Exception& err) {
    result = string("Error inside GetStage function of driver ")
      + string(driver->InstanceName()) + string(" --- ")
      + string(err.MessageType()) + string(" --- ")
      + err.MessageText();
#ifndef NDEBUG
  } catch (...) {
    result = string("Unrecognized error thrown from inside"
		    " GetStage function of driver ")
      + string(driver->InstanceName());
#endif // NDEBUG
  }
  return errcode;
}

int Oxs_Director::GetNumberOfStages(int& stage_count, string& result)
{
  if(driver==NULL) {
    result = "Problem not initialized (no driver identified)";
    return TCL_ERROR;
  }

  int errcode = TCL_ERROR;
  try {
    stage_count = static_cast<int>(driver->GetNumberOfStages());
    errcode = TCL_OK;
  } catch (Oxs_Ext::Error& err) {
    result = string("Oxs_Ext error inside"
		    " GetNumberOfStages function of driver ")
      + string(driver->InstanceName())
      + string(" --- ") + string(err);
  } catch (Oc_Exception& err) {
    result = string("Error inside GetNumberOfStages function of driver ")
      + string(driver->InstanceName()) + string(" --- ")
      + string(err.MessageType()) + string(" --- ")
      + err.MessageText();
#ifndef NDEBUG
  } catch (...) {
    result = string("Unrecognized error thrown from inside"
		    " GetNumberOfStages function of driver ")
      + string(driver->InstanceName());
#endif // NDEBUG
  }
  return errcode;
}

int Oxs_Director::GetMif(string& result)
{
  if(problem_id == 0) {
    result = "no problem loaded";
    return TCL_ERROR;
  }
  result = mif_object;
  return TCL_OK;
}

int Oxs_Director::ProbInit(const char* filename)
{ // Creates an Oxs_Mif object in the master interp, and
  // has it source the problem specified by filename, which
  // should be in MIF 2.x format.  The Oxs_Mif object is
  // deleted by the director when the problem is released,
  // but no sooner.  Note that the Oxs_Mif object creates
  // an slave interpreter in its constructor, which it holds
  // until it deletes it in its destructor.
  int i,error_code;
  string cmd;

  Release();
  simulation_state_reserve_count=0;  // Safety

  // Create mif object; side effect: creates a slave interpreter
  error_code = Tcl_Eval(interp,Oc_AutoBuf("Oxs_Mif New _"));
  if(error_code!=TCL_OK) return error_code;
  mif_object = string(Tcl_GetStringResult(interp));
  Tcl_ResetResult(interp); // Make frequent calls to Tcl_ResetResult
  /// to insure that we have a clean slate on which to build error
  /// messages generated by C++ exceptions.

  // Read file into Oxs_Mif object
  cmd = mif_object + string(" ReadMIF ")
    + string(" [list ") + string(filename) + string(" ]");
  error_code = Tcl_Eval(interp,Oc_AutoBuf(cmd.c_str()));
  if(error_code!=TCL_OK) return error_code;
  Tcl_ResetResult(interp);

  // Get list of spec blocks in mif object
  cmd = mif_object + string(" GetSpecKeys");
  error_code=Tcl_Eval(interp,Oc_AutoBuf(cmd.c_str()));
  if(error_code!=TCL_OK) return error_code;
  string speckeys = Tcl_GetStringResult(interp);
  Tcl_ResetResult(interp);
  Oxs_SplitList arglist;
  error_code = arglist.Split(speckeys);
  if(error_code!=TCL_OK) return error_code;
  Tcl_ResetResult(interp);

  // Create Oxs_Ext object for each spec block, storing pointer
  // in ext_obj, and mapping on id in ext_map.
  string bad_specs;
#ifdef REPORT_CONSTRUCT_TIMES
  Nb_StopWatch watch;
  Oc_TimeVal cpu,wall;
  fprintf(stderr,
	  "#    TIMING (seconds)               "
	  "                CPU        Elapsed\n");
#endif
  for(i=0;i<arglist.Count();i++) {
    cmd = mif_object + string(" GetSpecValue [list ")
      + string(arglist[i]) + string(" ]");
    error_code=Tcl_Eval(interp,Oc_AutoBuf(cmd.c_str()));
    if(error_code!=TCL_OK) return error_code;
    string argstr = Tcl_GetStringResult(interp);
    Tcl_ResetResult(interp);

    Oxs_Ext* spec_ptr=NULL;
    try {
#ifdef REPORT_CONSTRUCT_TIMES
      watch.Reset();
      watch.Start();
#endif
      spec_ptr=Oxs_Ext::MakeNew(arglist[i],this,
				argstr.c_str());
#ifdef REPORT_CONSTRUCT_TIMES
      watch.Stop();
      watch.GetTimes(cpu,wall);
      fprintf(stderr,"# %40s  %12.3f %12.3f\n",
	      arglist[i],double(cpu),double(wall));
#endif
      string idname=spec_ptr->InstanceName();
      if(ext_map.find(idname)!=ext_map.end()) {
	// Name already in use.  Append to bad spec list and delete.
	delete spec_ptr;
	bad_specs += string("\n ") + string(arglist[i])
	  + string(" (ID name ") + idname + string("already in use.)");
      } else {
	ext_map[idname]=spec_ptr;  // Store mapping
	// Store in ordered list
	ext_obj.push_back(spec_ptr);

	// If this is an Oxs_Energy object, then append it to
	// the energy_obj list as well.
	Oxs_Energy* energy_ptr
	  = dynamic_cast<Oxs_Energy*>(spec_ptr);
	if(energy_ptr!=NULL)
	  energy_obj.push_back(energy_ptr);

	// Set driver data member if this is the first driver object
	if(driver==NULL) {
	  Oxs_Driver* driver_ptr
	    = dynamic_cast<Oxs_Driver*>(spec_ptr);
	  if(driver_ptr!=NULL) driver=driver_ptr;
	}
      }
    } catch (Oxs_Ext::UnregisteredType) {
      // Drop exception, but append to bad_spec list for later reporting.
      spec_ptr=NULL;
      bad_specs += string("\n  ") + string(arglist[i]);
    } catch (Oxs_Ext::Error& err) {
      Tcl_AppendResult(interp,
		       "Oxs_Ext initialization error"
		       " in construction of ",
		       arglist[i],
		       ": ",
		       err.c_str(),(char *)NULL);
      if(bad_specs.length()>0) {
	Tcl_AppendResult(interp,"\nUnrecognized Specify Blocks---",
			 bad_specs.c_str(),(char *)NULL);
      }
      Release(); // Delete any Oxs_Ext objects already constructed
      return TCL_ERROR;
    } catch (Oc_Exception& err) {
      Tcl_AppendResult(interp,
		       "Oxs_Ext initialization error"
		       " in construction of ",
		       arglist[i]," --- ",
		       err.MessageType(),": ",
		       err.MessageText().c_str(),(char *)NULL);
      if(bad_specs.length()>0) {
	Tcl_AppendResult(interp,"\nUnrecognized Specify Blocks---",
			 bad_specs.c_str(),(char *)NULL);
      }
      Release(); // Delete any Oxs_Ext objects already constructed
      return TCL_ERROR;
    }
  }

  // Checks for invalid input
  if(bad_specs.length()>0 || driver==NULL) {
    BOOL no_driver = 1;
    if(driver!=NULL) no_driver = 0;
    if(bad_specs.length()>0) {
      Tcl_AppendResult(interp,"Unrecognized Specify Blocks---",
		       bad_specs.c_str(),"\n",(char *)NULL);
    }
    if(no_driver) {
      Tcl_AppendResult(interp,"No driver specified\n",(char *)NULL);
    }
    Release(); // Delete any Oxs_Ext objects already constructed
    return TCL_ERROR;
  }

  // Check to see if random number generator seed has set
  // in mif_object.  If so, record the fact via the director
  // member variable random_seed_set, and create a new seed.
  cmd = mif_object + string(" GetRandomSeed");
  error_code=Tcl_Eval(interp,Oc_AutoBuf(cmd.c_str()));
  if(error_code!=TCL_OK) return error_code;
  const char* pseed = Tcl_GetStringResult(interp);
  Tcl_ResetResult(interp);
  // We don't care what the original seed was, only
  // if it was set or not.  If set, then grab a new seed
  // from the random number generator.
  if(*pseed == '\0') {
    // Empty string => Seed not set
    random_seed_set=0;
    random_seed=0;
  } else {
    random_seed_set=1;
    random_seed = static_cast<UINT4m>(floor(Oc_UnifRand() * UINT_MAX));
  }

  // Run Init() on all Oxs_Ext's
  if(ProbReset()!=TCL_OK) {
    Release(); // Delete any Oxs_Ext objects already constructed
    return TCL_ERROR;
  }

  problem_name = string(filename);
  problem_id = ++problem_count;
  return TCL_OK;
}

// The following file-local not_null function is used with find_if
// on output_obj to determine if there are any registered output
// objects.
static bool not_null(const Oxs_Output* obj) { return obj!=NULL; }

void Oxs_Director::Release()
{
  problem_id=0;
  problem_name.erase();

  driver=NULL;  // Make sure we don't use this pointer anymore.

  // Release all Oxs_Ext's, from the back forward.  The
  // order is important so that objects at the back of the
  // list can release any dependency locks held on objects
  // at the front of the list.
  vector<Oxs_Ext*>::reverse_iterator ext_it = ext_obj.rbegin();
  while(ext_it!=ext_obj.rend()) {
    delete (*ext_it);
    ++ext_it;
  }
  // Release *all* pointers to these objects.
  ext_obj.clear();
  ext_map.clear();
  energy_obj.clear();

  // Release simulation states.  We don't care about the order.
  vector<Oxs_SimState*>::iterator sim_it = simulation_state.begin();
  while(sim_it!=simulation_state.end()) {
    delete (*sim_it);
    ++sim_it;
  }
  simulation_state.clear();
  simulation_state_reserve_count=0;

  // Delete MIF object
  Tcl_SavedResult saved;
  Tcl_SaveResult(interp,&saved);
  if(mif_object.length()>0) {
    string cmd = mif_object + string(" Delete");
    Tcl_ResetResult(interp);
    if(Tcl_Eval(interp,Oc_AutoBuf(cmd.c_str()))!=TCL_OK) {
      Tcl_DiscardResult(&saved);
      OXS_THROW(Oxs_BadResourceDealloc,
		"Error deleting MIF object during problem release.");
    }
    mif_object.erase();
  }
  Tcl_RestoreResult(interp,&saved);

  // Check that all output objects have deregistered
  vector<Oxs_Output*>::const_iterator output_it
    = find_if(output_obj.begin(),output_obj.end(),not_null);
  if(output_it!=output_obj.end()) {
    OXS_THROW(Oxs_BadResourceDealloc,
	     "Not all output objects deregistered upon problem release.");
  }
  output_obj.clear(); // Reset output object registration list

}

Oxs_Director::~Oxs_Director()
{
  Release();
  Tcl_DeleteExitHandler((Tcl_ExitProc *) ExitProc, (ClientData) this);
}

Tcl_Interp* Oxs_Director::GetMifInterp() const
{
  Tcl_Interp* mif_interp = NULL;
  Tcl_SavedResult saved;
  Tcl_SaveResult(interp, &saved);
  string cmd = mif_object + string(" Cget -mif_interp");
  if(Tcl_Eval(interp,Oc_AutoBuf(cmd.c_str()))==TCL_OK) {
    mif_interp = Tcl_GetSlave(interp,Tcl_GetStringResult(interp));
  }
  Tcl_RestoreResult(interp, &saved);
  return mif_interp;
}

int Oxs_Director::ReadImageFile
(const char* filename,
 string& image) const
{  // Returns 0 on success, and the export "image" is filled with a
  // string representation of the image.  This representation mirrors
  // that of a PPM P3 (text) image file, but without the leading "P3"
  // magic string or comments.  See the Tcl proc Oxs_ReadFile for
  // details.  On failure, the return value is 1, and "image" is
  // filled with an error message.
  BOOL err = 0;
  Tcl_SavedResult saved;
  Tcl_SaveResult(interp, &saved);
  string cmd = string("Oxs_ReadFile {")
    + string(filename) + string("} image");
  if(Tcl_Eval(interp,Oc_AutoBuf(cmd.c_str()))!=TCL_OK) {
    err=1;
  }
  image = Tcl_GetStringResult(interp);
  Tcl_RestoreResult(interp, &saved);
  return err;
}

int Oxs_Director::GetLogFilename(string& logname) const
{ // Returns 0 on success, and the export "logname" is filled the
  // name of the log file, as currently registered with FileLogger.
  // On failure, the return value is 1, and logname is set to an
  // empty string.
  BOOL err = 0;
  Tcl_SavedResult saved;
  Tcl_SaveResult(interp, &saved);
  if(Tcl_Eval(interp,Oc_AutoBuf("FileLogger GetFile"))!=TCL_OK) {
    err=1;
    logname = "";
  } else {
    logname = Tcl_GetStringResult(interp);
    if(logname.length()<1) err=1;
  }
  Tcl_RestoreResult(interp, &saved);
  return err;
}


// The next 3 routines query and get attributes from driver.  Return
// value is 0 on success, 1 if driver is not initialized.
// GetDriverAttibuteValue returns 2 if label is not in attributes list.
int Oxs_Director::GetDriverInstanceName(string& name) const
{
  if (driver == NULL) return 1;
  name = driver->InstanceName();
  return 0;
}
int Oxs_Director::GetDriverAttributes(vector<string>& labels) const
{
  if(driver==NULL) return 1;
  driver->GetAttributeLabels(labels);
  return 0;
}
int Oxs_Director::GetDriverAttributeValue(const string& label,
					  string& value) const
{
  if(driver==NULL) return 1;
  try {
    value = driver->GetAttribute(label);
  } catch (...) {
    return 2; // "label" not in attributes list
  }
  return 0;
}


Oxs_Ext*
Oxs_Director::FindExtObject(const string& id) const
{ // Tries to find a match of id against entries in ext_map.
  // If "id" contains a ":" at any position other than the
  // first character, then it must match exactly against
  // the key.  If the first character of "id" is ":", then
  // it is compared against the back portion of each key.
  // If "id" does not include a ":" separator, then it is
  // compared first against the front portion of each key,
  // and failing that it is compared against the back portion
  // of each key.  An invalid key is returned if a unique
  // match is not obtained.

  UINT4m match_count=0;
  Oxs_Ext* extobj=NULL;
  map<string,Oxs_Ext*>::const_iterator it;

  enum MATCH_TYPES { FRONT, BACK } match_type;
  if(id[0]==':')                      match_type = BACK;
  else if(id.find(':')==string::npos) match_type = FRONT;
  else {
    // Full exact match required.
    // Note: We use ext_map.find(id) rather than
    // ext_map[id] because the latter will insert a
    // dummy element into the map if 'id' is not in
    // the map.
    it = ext_map.find(id);
    if( it == ext_map.end() )
      return NULL; // No match
    return it->second;
  }

  if(match_type==FRONT) {
    it = ext_map.begin();
    while(it!=ext_map.end()) {
      const string& test_key = it->first;
      int matched=0;
      // STL defines substr versions of string.compare(), but
      // they seem to be missing from some implementations.
      string front = test_key.substr(0,id.size());
      if(front.compare(id)==0) {
	matched=1;
      }
      if(matched) {
	if(match_count==0) extobj = it->second;
	++match_count;
      }
      ++it;
    }
  }

  if(match_type==BACK || (match_type==FRONT && match_count==0) ) {
    it = ext_map.begin();
    while(it!=ext_map.end()) {
      const string& test_key = it->first;
      int matched=0;
      // STL defines substr versions of string.compare(), but
      // they seem to be missing from some implementations.
      string::size_type index=test_key.find(':');
      if(index!=string::npos) {
	if(match_type==FRONT) ++index; // Skip over ':'
	string back = test_key.substr(index);
	if(back.compare(id)==0) {
	  matched=1;
	}
      }
      if(matched) {
	if(match_count==0) extobj = it->second;
	++match_count;
      }
      ++it;
    }
  }

  if(match_count==1) return extobj;
  return NULL;
}

string Oxs_Director::ListExtObjects() const
{
  string result;
  vector<Oxs_Ext*>::const_iterator it = ext_obj.begin();
  if(it!=ext_obj.end()) {
    result.append((*it)->InstanceName());
    while((++it)!=ext_obj.end()) {
      result.append("\n");
      result.append((*it)->InstanceName());
    }
  }
  return result;
}

string Oxs_Director::ListEnergyObjects() const
{
  string result;
  vector<Oxs_Energy*>::const_iterator it = energy_obj.begin();
  if(it!=energy_obj.end()) {
    result.append((*it)->InstanceName());
    while((++it)!=energy_obj.end()) {
      result.append("\n");
      result.append((*it)->InstanceName());
    }
  }
  return result;
}

// Routines to register and deregister simulation outputs.
// NOTE: All objects registering output are expected to
// automatically deregister as a consequence of the object
// destructions occuring in Oxs_Director::Release().
void Oxs_Director::RegisterOutput(Oxs_Output* obj)
{
  // Check that object not already registered
  vector<Oxs_Output*>::iterator it
    = find(output_obj.begin(),output_obj.end(),obj);
  if(it != output_obj.end()) {
    // Object already registered
    string msg = string("Output object <")
      + obj->OwnerName() + string(",") 
      + obj->OutputName() + string("> already registered.");
    OXS_THROW(Oxs_BadIndex,msg.c_str());
  }

  // Otherwise, append to end of output_obj list
  output_obj.push_back(obj);
}

void Oxs_Director::DeregisterOutput(const Oxs_Output* obj)
{
  vector<Oxs_Output*>::iterator it
    = find(output_obj.begin(),output_obj.end(),obj);
  if(it == output_obj.end()) {
    OXS_THROW(Oxs_BadIndex,"Attempted deregistration of "
	     "non-registered output object.");
  }
  *it = NULL;
}

BOOL Oxs_Director::IsRegisteredOutput(const Oxs_Output* obj) const
{
  vector<Oxs_Output*>::const_iterator it
    = find(output_obj.begin(),output_obj.end(),obj);
  if(it == output_obj.end()) return 0;
  return 1;
}

const char*
Oxs_Director::MakeOutputToken(unsigned long index,unsigned long probid) const
{
  static Oc_AutoBuf buf;
  buf.SetLength(4*sizeof(unsigned long)+8);
  sprintf(static_cast<char *>(buf),"%lX %lX",index,probid);
  return static_cast<char *>(buf);
}

Oxs_Output*
Oxs_Director::InterpretOutputToken(const char* token) const
{
  unsigned long index=0,probid=0;
  char* endptr;
  index = strtol(token,&endptr,16);
  if(endptr!=NULL) {
    probid = strtol(endptr+1,&endptr,16);
  }
  if(endptr==NULL) {
    OXS_THROW(Oxs_BadParameter,
	     "Invalid token passed to"
	     " Oxs_Director::InterpretOutputToken");
  }
  if(probid!=problem_id) {
    OXS_THROW(Oxs_BadParameter,
	     "Problem ID mismatch on token passed to"
	     " Oxs_Director::InterpretOutputToken");
  }
  if(index>=output_obj.size()) {
    OXS_THROW(Oxs_BadIndex,
	     "Out-of-range index in token passed to"
	     " Oxs_Director::InterpretOutputToken");
  }
  Oxs_Output* obj = output_obj[index];
  if(obj==NULL) {
    OXS_THROW(Oxs_BadParameter,
	     "Token refers to deregistered output object in"
	     " Oxs_Director::InterpretOutputToken");
  }
  return obj;
}

struct OxsDirectorOutputKey {
public:
  size_t index;
  INT4m priority;
  OxsDirectorOutputKey() : index(0), priority(0) {}
  /// Default constructor required by STL containers.
  OxsDirectorOutputKey(size_t index_,INT4m priority_)
    : index(index_), priority(priority_) {}
};

bool operator<(const OxsDirectorOutputKey& a,
	       const OxsDirectorOutputKey& b)
{ 
  return a.priority<b.priority
    || (a.priority == b.priority && a.index<b.index);
}

// The following 2 functions are declared just to keep MSVC++ 5.0 happy
bool operator>(const OxsDirectorOutputKey&, const OxsDirectorOutputKey&);
bool operator==(const OxsDirectorOutputKey&, const OxsDirectorOutputKey&);

void Oxs_Director::ListOutputObjects(vector<string> &outputs) const
{
  // Create list of output objects sorted first by priority and
  // second by registration order.
  vector<OxsDirectorOutputKey> keylist;
  for(size_t i=0;i<output_obj.size();++i) {
    const Oxs_Output* obj = output_obj[i];
    if(obj!=NULL) {
      keylist.push_back(OxsDirectorOutputKey(i,obj->Priority()));
    }
  }
  sort(keylist.begin(),keylist.end());

  // Fill export "outputs" with sorted list
  outputs.clear();
  vector<OxsDirectorOutputKey>::const_iterator it;
  for(it=keylist.begin();it!=keylist.end();++it) {
    outputs.push_back(MakeOutputToken(it->index,problem_id));
  }
}

Oxs_Output* Oxs_Director::FindOutputObject(const char* regexp) const
{ // C++ interface for searching the output object list.  This routine
  // matches the import regexp against the output objects list.  If a
  // unique match is found, then a pointer to that object is returned.
  // Otherwise, a NULL pointer is returned.  The match is performed
  // using the Tcl regexp engine.  Since this routine will typically
  // check through the entire list of output objects, ideally this
  // routine is called after problem initialization, and the result is
  // cached until the problem is released.

  // Get internal representation for regexp
  Tcl_Interp* mif_interp = GetMifInterp();
  if(mif_interp==NULL) {
    OXS_THROW(Oxs_BadPointer,"MIF interpreter not initialized in "
	     "Oxs_Director::FindOutputObject()");
  }
  Tcl_RegExp rp = Tcl_RegExpCompile(mif_interp,Oc_AutoBuf(regexp));
  if(rp==NULL) {
    string msg = "Invalid regular expression in "
      "Oxs_Director::FindOutputObject(): ";
    msg += string(Tcl_GetStringResult(interp));
    OXS_THROW(Oxs_BadParameter,msg);
  }
  Oxs_Output* matched_obj=NULL;
  size_t size = output_obj.size();
  for(size_t i=0;i<size;i++) {
    Oxs_Output* obj = output_obj[i];
    if(obj==NULL) continue; // Deregistered object
    string output_name = obj->LongName();
    Oc_AutoBuf ab = output_name.c_str();
    int match_result 
      = Tcl_RegExpExec(mif_interp,rp,(char*)ab,(char*)ab);
    if(match_result==0) continue; // No match
    if(match_result<0) { // Error
      string msg = "Regular expression match error in "
	"Oxs_Director::FindOutputObject(): ";
      msg += string(Tcl_GetStringResult(interp));
      OXS_THROW(Oxs_TclError,msg);
    }
    // Match found
    if(matched_obj!=NULL) { // More than one matched found
      return NULL;
    }
    matched_obj = obj; // First match against this regexp
  }
  return matched_obj;
}

int Oxs_Director::Output(const char* output_token,
			 Tcl_Interp* use_interp,
			 int argc,CONST84 char** argv)
{
  if(driver==NULL) {
    Tcl_AppendResult(use_interp,"No driver specified\n",(char *)NULL);
    return TCL_ERROR;
  }
  const Oxs_SimState* state=driver->GetCurrentState();
  if(state==NULL) {
    Tcl_AppendResult(use_interp,"Current state not initialized\n",
		     (char *)NULL);
    return TCL_ERROR;
  }
  Oxs_Output* obj = InterpretOutputToken(output_token);

  int errcode = TCL_ERROR;
  string errmsg;
  try {
    errcode = obj->Output(state,use_interp,argc,argv);
  } catch (Oxs_Ext::Error& err) {
    errmsg = string("Error evaluating output \"")
      + obj->OutputName()
      + string("\" of object \"")
      + obj->OwnerName()
      + string("\": ")
      + string(err);
    errcode = TCL_ERROR; // Safety
  } catch (Oc_Exception& err) {
    errmsg = string("Error evaluating output \"")
      + obj->OutputName()
      + string("\" of object \"")
      + obj->OwnerName()
      + string("\": ")
      + string(err.MessageType()) + string(" --- ")
      + err.MessageText();
    errcode = TCL_ERROR; // Safety
  } catch (EXCEPTION& err) {
    errmsg = string("Error evaluating output \"")
      + obj->OutputName()
      + string("\" of object \"")
      + obj->OwnerName()
      + string("\": ")
      + string(err.what());
    errcode = TCL_ERROR; // Safety
#ifndef NDEBUG
  } catch (...) {
    errmsg = string("Unrecognized error evaluating output \"")
      + obj->OutputName()
      + string("\" of object \"")
      + obj->OwnerName()
      + string("\".");
    errcode = TCL_ERROR; // Safety
#endif // NDEBUG
  }

  if(errcode == TCL_ERROR && errmsg.size()>0) {
    Tcl_AppendResult(use_interp,errmsg.c_str(),(char *)NULL);
  }

  return errcode;
}

int
Oxs_Director::OutputCacheRequestIncrement(const char* output_token,
					  INT4m incr) const
{
  Oxs_Output* obj = InterpretOutputToken(output_token);
  if (obj->CacheRequestIncrement(incr)) return TCL_OK;
  return TCL_ERROR;
}

int Oxs_Director::OutputNames(const char* output_token,
       string& owner_name,string& output_name,
       string& output_type,string& output_units) const
{
  Oxs_Output* obj = InterpretOutputToken(output_token);
  owner_name   = obj->OwnerName();
  output_name  = obj->OutputName();
  output_type  = obj->OutputType();
  output_units = obj->OutputUnits();
  return TCL_OK;
}

// Energy object index access
Oxs_Energy* Oxs_Director::GetEnergyObj(UINT4m i) const {
  if(i >=  energy_obj.size()) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"Request for energy object (%u) past end of list (%u)",
		(unsigned int)i,(unsigned int)energy_obj.size());
    OXS_THROW(Oxs_BadIndex,buf);
  }
  return energy_obj[i];
}


// Collect stage requests from all registered Oxs_Ext objects.
void Oxs_Director::ExtObjStageRequestCounts
(unsigned int& min,
 unsigned int& max) const
{
  min=0;
  max=UINT_MAX;  // Defaults
  vector<Oxs_Ext*>::const_iterator it = ext_obj.begin();
  while(it != ext_obj.end()) {
    unsigned int tmin,tmax;
    (*it)->StageRequestCount(tmin,tmax);
    if(tmin>min) min=tmin;
    if(tmax<max) max=tmax;
    ++it;
  }
}


// SIMULATION STATE MANAGEMENT: During construction, each Oxs_Ext object
// that wants to create and/or hold simulation states should register
// its intent with the director using ReserveSimulationStateRequest().
// The states are actually constructed only as needed, but the count
// obtained via reserve calls is used as an upper bound to protect
// against memory leaks swallowing up all of system memory.  Objects
// obtain keys (handles) to new states with the GetNewSimulationState()
// call.  The key initially holds a write lock on the object.  The
// caller can change the lock state, make copies, etc., as desired.  The
// GetNewSimulationState() routine scans through the list of states held
// in the simulation_state array, looking for a pre-created state that
// has no locks on it.  If one is found, a key is returned for that
// state.  If all existing states in the simulation_state array are
// locked, then a new state is created (subject to the max state count
// obtained from ReserveSimulationStateRequest() calls).  In all cases
// the returned key holds a write lock on the state.  The states are
// actually deleted only inside the director Release() command, after
// destruction of all Oxs_Ext objects.
//   NOTE: The key passed in to GetNewSimulationState() is allowed to
// hold a lock on an Oxs_SimState, *IF* that state was obtained from a
// previous call to GetNewSimulationState().  On entry,
// GetNewSimulationState() releases any lock currently held by the
// import key, but then immediately tests if the corresponding
// Oxs_SimState is then free.  If so, then GetNewSimulationState()
// resets the key to the same object.  Therefore, if the chance is good
// that the input key holds the sole lock on an object, then the lock
// should *not* be released before calling GetNewSimulationState().

void Oxs_Director::ReserveSimulationStateRequest
(UINT4m reserve_count)
{
  simulation_state_reserve_count+=reserve_count;
}

void Oxs_Director::GetNewSimulationState
(Oxs_Key<Oxs_SimState>& write_key)
{
  // Check first to see if import write_key is currently holding a lock.
  // If so, release it, and immediately check to see if that SimState
  // is available.
  const Oxs_SimState* csptr = write_key.GetPtr();
  if(csptr!=NULL) {
    write_key.Release();
    if(csptr->ReadLockCount()==0 && csptr->WriteLockCount()==0 ) {
      Oxs_SimState* sptr = const_cast<Oxs_SimState*>(csptr);
      sptr->Reset();
      write_key.Set(sptr);
      write_key.GetWriteReference();
      return;
    }
  }

  // Otherwise, go to the simulation_state list, and try to find
  // an unlocked state, working from the back forward.
  vector<Oxs_SimState*>::reverse_iterator it
    = simulation_state.rbegin();
  while(it!=simulation_state.rend()) {
    Oxs_SimState* sptr = *it;
    if(sptr->ReadLockCount()==0 && sptr->WriteLockCount()==0) {
      sptr->Reset();
      write_key.Set(sptr);
      write_key.GetWriteReference();
      return;
    }
    ++it;
  }

  // If we get to here, then all existing states have locks
  // on them.  So create a new state.
  if(simulation_state.size()>=simulation_state_reserve_count) {
    OXS_THROW(Oxs_BadResourceAlloc,
	     "Simulation state reserve count exhausted.");
  }
  Oxs_SimState* newstate = new Oxs_SimState;
  simulation_state.push_back(newstate);
  write_key.Set(newstate);
  write_key.GetWriteReference();
}


// #define DEVELOP_TEST to make DoDevelopTest active.  It may
// then be accessed from the Oxs solver console (use '-tk 1 -console'
// in Oxsii launch command) via the 'Oxs_DirectorDevelopTest count'
// command.
#define DEVELOP_TEST

#ifndef DEVELOP_TEST
// Default NOP (inactive) code with error return
int Oxs_Director::DoDevelopTest
(const string& /* in */,
 string& /* out */)
{
  return 1;
}
#else // DEVELOP_TEST
// Working code; should return 0 on success, >0 on error
int Oxs_Director::DoDevelopTest(const string& in,string& out)
{ 
  INT4m count = atoi(in.c_str());
  if(count<1) return 1;

  if(driver==NULL) return 2;
  const Oxs_SimState* state = driver->GetCurrentState();
  if(state==NULL) return 3;

  REAL8m mindot = DBL_MAX;
  for(INT4m i=0;i<count;i++) {
    mindot = state->mesh->MinNeighborDot(state->spin,*(state->Ms));
  }
  if(mindot >  1) mindot =  1;
  if(mindot < -1) mindot = -1;
  char buf[256];
  Oc_Snprintf(buf,sizeof(buf),"%.17g",acos(mindot)*(180.0/PI));
  out = string(buf);
  return 0;
}
#endif // DEVELOP_TEST

#ifdef DEVELOP_TEST
#undef DEVELOP_TEST
#endif // DEVELOP_TEST


