/* FILE: director.h                 -*-Mode: c++-*-
 *
 * 3D solver director class.
 *
 */

#ifndef _OXS_DIRECTOR
#define _OXS_DIRECTOR

#include <map>
#include <string>
#include <vector>

#include "oc.h"
#include "key.h"
#include "simstate.h"
#include "mesh.h"
#include "meshvalue.h"
#include "threevector.h"
#include "output.h"
#include "energy.h"	// Needed by MSVC++ 5

OC_USE_STD_NAMESPACE;  // Specify std namespace, if supported

/* End includes */

class Oxs_Ext;       // Forward declarations
class Oxs_Energy;
class Oxs_Driver;

// Supplemental types
enum OxsRunEventTypes { OXS_STEP_EVENT,
			OXS_STAGE_DONE_EVENT,
			OXS_RUN_DONE_EVENT };
struct OxsRunEvent {
public:
  OxsRunEventTypes event;
  Oxs_ConstKey<Oxs_SimState> state_key;
  OxsRunEvent();
  OxsRunEvent(const OxsRunEventTypes& event_,
	      const Oxs_ConstKey<Oxs_SimState>& state_key_)
    : event(event_), state_key(state_key_) {
    state_key.GetReadReference();  // Set and hold read lock
  }
};
/// IMPORTANT NOTE: OxsRunEvent holds a read lock on its associated
/// state from the time of the OxsRunEvent's creation until its
/// destruction.
BOOL operator<(const OxsRunEvent&, const OxsRunEvent&);
BOOL operator>(const OxsRunEvent&, const OxsRunEvent&);
BOOL operator==(const OxsRunEvent&, const OxsRunEvent&);

// 3D solver director
class Oxs_Director {
private:
  static void ExitProc(ClientData);

  UINT4m problem_count; // Number of problems loaded across history
  /// of this instance.  Initialized to 0, then incremented to 1 when
  /// first problem is loaded, and incremented by 1 more each time
  /// a new problem is loaded.

  UINT4m problem_id; // Current problem count id.  Set to 0
  /// if no problem is loaded.  Otherwise should equal problem_count.
  /// NB: ProbRun is disabled if problem_id==0.

  string problem_name; // Name of MIF file providing current
  /// problem.  Might want to include some additional identifying
  /// information, such as perhaps a CRC?

  Tcl_Interp* const interp;

  // MIF object, which holds mif_interp, a slave of interp.  This is
  // created inside ProbInit(), and is held until Release()
  string mif_object;

  // Random number generator.  Keep track if initialized with a seed,
  // and if so, what that seed was.
  BOOL random_seed_set;
  UINT4m random_seed;

  void Release();

  vector<Oxs_Ext*> ext_obj;  // Ordered list of instantiated Oxs_Ext
  /// objects.  This list "owns" the objects, and is responsible for
  /// their destruction.
  map<string,Oxs_Ext*> ext_map; // Mapping of Ext names to objects
  vector<Oxs_Energy*> energy_obj; // Ordered list of energy terms.  This
  // is a sublist of ext_obj.

  vector<Oxs_Output*> output_obj; // List of all registered output objects
  /// When an output object deregisters, its slot in output_obj is
  /// replaced with a NULL pointer.  This makes it easy for external
  /// code (in particular, Tcl script) to refer to a registered object---
  /// once an object is registered its index (i.e., position in the
  /// output_obj list) never changes.  The output_obj list is reset each
  /// time a new problem is loaded.  However, this is a poor algorithm
  /// if during one problem many output objects register and deregister,
  /// since the size of output_obj will grow accordingly.  At some point
  /// a different indexing scheme should replace this.  Perhaps a hash
  /// lookup?

  const char* MakeOutputToken(unsigned long index,unsigned long probid) const;
  Oxs_Output* InterpretOutputToken(const char* token) const;

  Oxs_Driver* driver; // Unique driver object.  This can be changed
  /// to a container if in the future we want to support multiple
  /// drivers in one problem run.

  // Array of held simulation states.  See discussion below concerning
  // ReserveSimulationStateRequest() and GetNewSimulationState().  This
  // array "owns" the simulations states, i.e., is responsible for their
  // destruction.
  UINT4m simulation_state_reserve_count;
  vector<Oxs_SimState*> simulation_state;

  // NB: Because ext_map and energy_obj hold duplicate pointers to
  // objects stored in ext_obj, one has to be careful to release all
  // pointers at the same time.  This is handled inside the Release()
  // function.  We could store Oxs_Keys in ext_map and energy_obj
  // instead of pointers; this would provide notification of hanging
  // pointers, but since Oxs_Director is the owner of these objects, at
  // present I don't think this is necessary. -mjd

public:
  Oxs_Director(Tcl_Interp* interp_);
  ~Oxs_Director();  // Non-virtual.  Oxs_Director is not intended to
  // have children.

  int VerifyInterp(Tcl_Interp* i) { return (i==interp ? 1 : 0); }

  int ProbInit(const char* filename);
  /// Import mif must be the name of an existing MIF object
  /// in the Tcl interpreter 'interp' passed to the constructor
  /// of *this.

  const char* GetProblemName() const { return problem_name.c_str(); }

  int ProbReset();
  int ProbRelease() { Release(); return TCL_OK; }

  int ProbRun(vector<OxsRunEvent>& results,string& errmsg);
  /// On success, fills results with event list and returns 0.
  /// Otherwise, return is >0, and errmsg is filled with a message
  /// describing the error.

  // Both SetStage and GetStage return TCL_OK on success.
  // On error, TCL_ERROR is returned and errmsg holds
  // a diagnostic message.
  int SetStage(int requestedStage, string& errmsg);
  int GetStage(int& stage, string& errmsg);
  int GetNumberOfStages(int& stage_count, string& errmsg);

  int GetMif(string& result);

  Tcl_Interp* GetMifInterp() const;

  int ReadImageFile(const char* filename,string& image) const;
  /// Returns 0 on success, and the export "image" is
  /// filled with a string representation of the image.
  /// This representation mirrors that of a PPM P3 (text)
  /// image file, but without the leading "P3" magic string
  /// or comments.  See the Tcl proc Oxs_ReadImage for details.
  /// On failure, the return value is 1, and image is filled
  /// with an error message.

  int GetLogFilename(string& logname) const;
  /// Returns name of log file registered with FileLogger
  /// in the main Tcl interpreter.

  // Routines to query and get attributes from driver.
  // Return value is 0 on success, 1 if driver is not
  // initialized.  GetDriverAttibuteValue returns 2
  // if label is not in attributes list.
  int GetDriverInstanceName(string& name) const;
  int GetDriverAttributes(vector<string>& labels) const;
  int GetDriverAttributeValue(const string& label,
			      string& value) const;

  Oxs_Ext* FindExtObject(const string& id) const;
  /// Returns a pointer to the element of ext_map with instance id
  /// 'id'.  Returns NULL if the requested object is not
  /// found.  For convenience, some partial matches are allowed.  See
  /// the routine implementation for details.
  ///    It is recommended that the client place a dep lock on the
  /// object if it is to be used for an extended period.  All such locks
  /// must be released before the corresponding object in ext_map can be
  /// deleted.  This means that locks held by ext_map objects to other
  /// ext_map objects should be only to objects earlier in the ext_obj
  /// list than themselves.  Perhaps there should be some type of access
  /// control on this function?

  string ListExtObjects() const;  // Mainly for debugging.
  string ListEnergyObjects() const;  // Mainly for debugging.

  // The following output (de)registration routines throw exceptions
  // on failure. NOTE: All objects registering output are expected to
  // automatically deregister as a consequence of the object destructions
  // occuring in Oxs_Director::Release().  See also output_map object
  // above.
  void RegisterOutput(Oxs_Output*);
  void DeregisterOutput(const Oxs_Output*);
  BOOL IsRegisteredOutput(const Oxs_Output*) const;
  void ListOutputObjects(vector<string> &outputs) const;
  int Output(const char* output_token,
	     Tcl_Interp* interp,
	     int argc,CONST84 char** argv);
  int OutputCacheRequestIncrement(const char* output_token,
				  INT4m incr) const;
  int OutputNames(const char* output_token,
		  string& owner_name,string& output_name,
		  string& output_type,string& output_units) const;

  Oxs_Output* FindOutputObject(const char* regexp) const;
  /// C++ interface for searching the output object list.  This routine
  /// matches the import regexp against the output objects list.  If a
  /// unique match is found, then a pointer to that object is returned.
  /// Otherwise, a NULL pointer is returned.  The match is performed
  /// using the Tcl regexp engine.  Since this routine will typically
  /// check through the entire list of output objects, ideally this
  /// routine is called after problem initialization, and the result is
  /// cached until the problem is released.


  // Energy object index access
  UINT4m GetEnergyObjCount() const { return energy_obj.size(); }
  Oxs_Energy* GetEnergyObj(UINT4m i) const;

  // Collect stage requests from all registered Oxs_Ext objects.
  void ExtObjStageRequestCounts(unsigned int& min,
				unsigned int& max) const;

  // SIMULATION STATE MANAGEMENT: During construction, each Oxs_Ext
  // object that wants to create and/or hold simulation states should
  // register its intent with the director using
  // ReserveSimulationStateRequest().  The states are actually
  // constructed only as needed, but the count obtained via reserve
  // calls is used as an upper bound to protect against memory leaks
  // swallowing up all of system memory.  Objects obtain keys (handles)
  // to new states with the GetNewSimulationState() call.  The key
  // initially holds a write lock on the object.  The caller can change
  // the lock state, make copies, etc., as desired.  The
  // GetNewSimulationState() routine scans through the list of states
  // held in the simulation_state array, looking for a pre-created state
  // that has no locks on it.  If one is found, a key is returned for
  // that state.  If all existing states in the simulation_state array
  // are locked, then a new state is created (subject to the max state
  // count obtained from ReserveSimulationStateRequest() calls).  In all
  // cases the returned key holds a write lock on the state.  The states
  // are actually deleted only inside the director Release() command,
  // after destruction of all Oxs_Ext objects.
  //   NOTE: The key passed in to GetNewSimulationState() is allowed to
  // hold a lock on an Oxs_SimState, *IF* that state was obtained from a
  // previous call to GetNewSimulationState().  On entry,
  // GetNewSimulationState() releases any lock currently held by the
  // import key, but then immediately tests if the corresponding
  // Oxs_SimState is then free.  If so, then GetNewSimulationState()
  // resets the key to the same object.  Therefore, if the chance is
  // good that the input key holds the sole lock on an object, then the
  // lock should *not* be released before calling
  // GetNewSimulationState().
  void ReserveSimulationStateRequest(UINT4m reserve_count);
  void GetNewSimulationState(Oxs_Key<Oxs_SimState>& write_key);

  // Hook for develop tests
  int DoDevelopTest(const string& in,string& out);

};

#endif // _OXS_DIRECTOR
