/* FILE: ext.h                 -*-Mode: c++-*-
 *
 * Base OXS extension class.  This is an abstract class.
 *
 */

#ifndef _OXS_EXT
#define _OXS_EXT

#include <map>
#include <string>
#include <vector>

#include "oc.h"
#include "oxsexcept.h"
#include "lock.h"
#include "threevector.h"
#include "util.h"

OC_USE_STD_NAMESPACE;  // Specify std namespace, if supported

/* End includes */

////////////////////////////////////////////////////////////////////////
// Macro that builds a generic child constructor, and provides a static
// Oxs_ExtRegistrationBlock variable to force automatic class
// registration.  This macro should be invoked at the top of each
// Oxs_Ext child class *.cc class definition file that requires
// registration, i.e., any class for which object instantiation is
// desired.
//   The child constructor produced by this macro has the name
//
//          OxsExtNewChild##classname
//
// where "classname" is the macro argument, and the name of the Oxs_Ext
// child class being registered.
//   This macro also provides definitions for the class public member
// function ClassName().  This is used instead of an RTTI call, because
// we want to be able to associate this name reported by the class with
// entries in the input MIF file.  The RTTI "typeid().name()" function
// returns an implementation dependent string, which is not what we want.
#define OXS_EXT_REGISTER(classname)                               \
static Oxs_Ext*                                                   \
OxsExtNewChild##classname(const char* instancename,               \
			  Oxs_Director* newdtr,                   \
			  const char* argstr)                     \
{ return static_cast<Oxs_Ext*>(                                   \
     new classname(instancename,newdtr,argstr)); }                \
const char* classname::ClassName() const { return #classname; }   \
static Oxs_ExtRegistrationBlock oxsExt##classname##RegBlock(      \
#classname,OxsExtNewChild##classname)

class Oxs_Ext;       // Forward declarations
class Oxs_Director;

typedef Oxs_Ext* Oxs_ExtChildConstructFunc(
   const char* instancename, // Child instance id string
   Oxs_Director* dtr,        // App director
   const char* argstr);      // MIF input block parameters

// Child class registration block
struct Oxs_ExtRegistrationBlock {
public:
  string classid;  // Class name
  Oxs_ExtChildConstructFunc * construct;  // Constructor

  // Standard constructor.  Performs automatic registration
  // by calling Oxs_Ext::Register();
  Oxs_ExtRegistrationBlock(const char* name,
			   Oxs_ExtChildConstructFunc* ctr);

  // Default constructor, for use by STL "map" container.
  Oxs_ExtRegistrationBlock():classid(""),construct(NULL) {}
};

////////////////////////////////////////////////////////////////////////
// Macros to construct and downcast an Oxs_Ext references
// NOTE: Some of the error handling assumes the code using
//       these functions is a child of Oxs_Ext, in particular,
//       that *this exists and can be cast to const *Oxs_Ext.
#define OXS_GET_EXT_OBJECT(params,subtype,obj)                        \
   { Oxs_OwnedPointer<Oxs_Ext> _goeo_tmp;                             \
     try {                                                            \
       FindOrCreateObject(params,_goeo_tmp);                          \
     } catch (Oxs_Ext::Error& msg) {                                  \
       /* Special case handling */                                    \
       if(strcmp(#subtype,"Oxs_ScalarField")==0 &&                    \
	  params.size()==1) {                                         \
	 /* Try to create uniform scalar field */                     \
	 BOOL err;                                                    \
	 Nb_Atof(params[0].c_str(),err);                              \
	 if(err) throw; /* No luck */                                 \
	 vector<string> tmp_params;                                   \
	 vector<string> tmp_sublist;                                  \
	 tmp_params.push_back("Oxs_UniformScalarField:");             \
	 tmp_sublist.push_back("value");                              \
	 tmp_sublist.push_back(params[0]);                            \
	 tmp_params.push_back(Oxs_MergeList(&tmp_sublist));           \
	 FindOrCreateObject(tmp_params,_goeo_tmp);                    \
      } else if(strcmp(#subtype,"Oxs_VectorField")==0 &&              \
	  params.size()==3) {                                         \
	 /* Try to create uniform vector field */                     \
	 vector<string> tmp_params;                                   \
	 vector<string> tmp_sublist;                                  \
	 vector<string> tmp_veclist;                                  \
	 tmp_params.push_back("Oxs_UniformVectorField:");             \
	 tmp_sublist.push_back("vector");                             \
	 tmp_veclist.push_back(params[0]);                            \
	 tmp_veclist.push_back(params[1]);                            \
	 tmp_veclist.push_back(params[2]);                            \
	 tmp_sublist.push_back(Oxs_MergeList(&tmp_veclist));          \
	 tmp_params.push_back(Oxs_MergeList(&tmp_sublist));           \
	 FindOrCreateObject(tmp_params,_goeo_tmp);                    \
       } else {                                                       \
	 throw;  /* No luck */                                        \
       }                                                              \
     }                                                                \
     if(_goeo_tmp.IsOwner()) {                                        \
        obj.SetAsOwner(dynamic_cast<subtype *>(_goeo_tmp.GetPtr()));  \
        _goeo_tmp.ReleaseOwnership();                                 \
     } else {                                                         \
        obj.SetAsNonOwner(dynamic_cast<subtype *>(_goeo_tmp.GetPtr())); \
     }                                                                \
     if(obj.GetPtr()==NULL) {                                         \
        string msg=string("Failed downcast: ")                        \
          + string(_goeo_tmp->InstanceName())                         \
          + string(" is not of type ") + string(#subtype)             \
          + string(" in source code file " OC_STRINGIFY(__FILE__))    \
          + string(", line " OC_STRINGIFY(__LINE__));                 \
        throw Oxs_Ext::Error(this,msg);                               \
     }                                                                \
   }

#define OXS_GET_INIT_EXT_OBJECT(name,subtype,obj)                     \
   { vector<string> _gieo_params;                                     \
     FindRequiredInitValue(name,_gieo_params);                        \
     try {                                                            \
        OXS_GET_EXT_OBJECT(_gieo_params,subtype,obj);                 \
     } catch (Oxs_Ext::Error& err) {                                  \
        err.Prepend(string("Error processing label \"")               \
           + name + string("\" --- "));                               \
        throw;                                                        \
     } catch (Oc_Exception& err) {                                    \
        err.Prepend(string("Error processing label \"")               \
           + name + string("\" --- "));                               \
        throw;                                                        \
     } catch (EXCEPTION& err) {                                       \
       string msg                                                     \
	 = string("System library error processing label \"")         \
	 + name + string("\" --- ") + string(err.what());             \
       throw Oxs_Ext::Error(this,msg);                                \
     } catch (...) {                                                  \
       string msg = string("Unrecognized error processing label \"")  \
	 + name + string("\".");                                      \
       throw Oxs_Ext::Error(this,msg);                                \
     }                                                                \
     DeleteInitValue(name);                                           \
   }

////////////////////////////////////////////////////////////////////////
// Oxs_Ext base class
class Oxs_Ext : public Oxs_Lock {
  // Perhaps the inheritance from Oxs_Lock should be private,
  // but then exposing the Oxs_Lock member functions is awkward.
  friend struct Oxs_ExtRegistrationBlock; // Makes ::Register() calls.
private:
  // List (actually, "map")  of all registered classes.  The actual
  // map is stored as a function static variable of RegClasses().
  // This insures the map is initialized before it is used, which
  // is difficult to insure otherwise, because Oxs_Ext child classes
  // call Oxs_Ext::Register during static initialization.  (See
  // Section 9.4.1 of "The C++ Programming Lanuage," Bjarne Stroustrup,
  // 3rd edition for a discussion of nonlocal variable initialization
  // order.)
  static map<string,Oxs_ExtRegistrationBlock>& RegClasses();

  static void Register(const Oxs_ExtRegistrationBlock & regblk);

  const string instanceId;     // Object instance name

  // Disable copy constructor and assignment operators.  Don't
  // provide definitions for these!
  Oxs_Ext(const Oxs_Ext&);
  Oxs_Ext& operator=(const Oxs_Ext&);

  // Helper function for extended constructor (see below).
  // Throws exception if 'key' is already in init_strings;
  // otherwise adds key+value to init_strings.
  void AddNewInitStringPair(const string& key,const string& value);

protected:
  Oxs_Director* director;  // Controlling director object

  // Base Oxs_Ext constructor.  Call this from a child class if
  // the child wants to parse the MIF argstr on its own. The
  // extended Oxs_Ext constructor farther below is available to
  // provide automatic MIF argstr parsing.
  Oxs_Ext(const char* idname,Oxs_Director* dtr);

  // Helper routine to extract className from an instanceName
  static string ExtractClassName(const string& name);

  // Helper routine to extract instance tail from
  // instanceName = className + instanceTail.
  static string ExtractInstanceTail(const string& name);

  // Routine to check that instanceId and classId are consistent.
  // Throws an error if they aren't.  This is called inside
  // Oxs_Ext::MakeNew().
  BOOL ValidateIdStrings() const;

  // Support code for child construction.  This code is activated
  // by the use of the constructor that has the extended
  // parameter list including 'const char* argstr'.  This
  // constructor requires that argstring be a Tcl list of pairs,
  // where the first element in each pair is an id string, and
  // the second is a value string.  The id string is mapped to
  // the key entry in init_strings, and the value string is
  // stored as the value corresponding to that key.  If the child
  // MIF class argstr does not fit this format, then use the
  // base, non-extended Oxs_Ext constructor.
  Oxs_Ext(const char* idname,Oxs_Director* dtr,const char* argstr);

  map<string,string> init_strings; // Initialization strings.

  void CheckInitValueParamCount(const string& key,UINT4m count) const;
  /// Throws an exception if key is not in init_strings, or if the
  /// associated value is not a valid Tcl list, or if the number
  /// of parameters in the value is not exactly count.

  BOOL HasInitValue(const string& key) const;
  /// Returns 1 if key is in init_strings, 0 otherwise.

  BOOL FindInitValue(const string& key,string& value) const;
  BOOL FindInitValue(const string& key,vector<string>& params) const;
  /// Returns the value associated with key, as either the original
  /// raw string or as a parameter list resulting from sending
  /// the value string through Tcl_SplitList.  The return value is
  /// 1 if key is in init_strings, 0 otherwise.  The params version
  /// will throw an exception if the value is not a valid Tcl list.
  ///   NB: Unlike the Get*InitValue routines, the Find*InitValue
  /// routines do no remove key+value from init_strings.  The
  /// caller must make an explicit call to DeleteInitValue to
  /// do this.

  void FindRequiredInitValue(const string& key,string& value) const;
  void FindRequiredInitValue(const string& key,vector<string>& params) const;
  /// Same as FindInitValue routines, but throws an exception if key is
  /// not in init_strings.  The params version will throw an exception
  /// if the value is not a valid Tcl list.

  BOOL DeleteInitValue(const string& key); // Removes the specified
  /// key+value pair from init_strings.  Returns 1 on success, 0
  /// if key was not found in init_strings.

  UINT4m GetInitValueCount() {
    return static_cast<UINT4m>(init_strings.size());
  }

  void VerifyAllInitArgsUsed(); // Throws an exception if
  /// GetInitValueCount() != 0.  Use this in combination with
  /// DeleteInitValue() to check MIF input block for superfluous
  /// parameters.  This should be called only in the actual
  /// instantiating class (not a parenting class).

  // The following functions are provided as a convenience to child
  // classes, because they occur frequently in practice.  They perform
  // a FindInitValue() call on key, and convert the result to the
  // appropriate type.  An exception is thrown if the lookup fails.
  // On success, the key+value is removed from init_strings (i.e.,
  // DeleteInitValue(key) is built into these functions).  The return
  // is by value, so no clean-up is required.
  REAL8m GetRealInitValue(const string& key);
  ThreeVector GetThreeVectorInitValue(const string& key);
  INT4m GetIntInitValue(const string& key);
  UINT4m GetUIntInitValue(const string& key);
  string GetStringInitValue(const string& key);

  // The next group are duplicates of the above, except return
  // default_value instead of throwing an error if key is not
  // in init_strings.
  REAL8m GetRealInitValue(const string& key,REAL8m default_value);
  ThreeVector GetThreeVectorInitValue(const string& key,
				      const ThreeVector& default_value);
  INT4m GetIntInitValue(const string& key,INT4m default_value);
  UINT4m GetUIntInitValue(const string& key,UINT4m default_value);
  string GetStringInitValue(const string& key,
			    const string& default_value);

  // Support for reading fixed-length lists (vectors)
  void GetRealVectorInitValue(const string& key,UINT4m length,
			      vector<REAL8m>& values);
  /// The value associated with key should be a list of "length"
  /// real values.  This function converts the value into a vector
  /// of REAL8m's, stores the results in export values, and
  /// removes the key-value pair from init_strings.  Throws an
  /// exception if key in not in init_strings, of if any of the
  /// values cannot be completed converted to a REAL8m, or if the
  /// parameter list associated with key does not have "length"
  /// entries.

  // Support functions for reading nested lists
  void GetGroupedStringListInitValue(const string& key,
				     vector<string>& expanded_list);
  /// The value associated with key should be a grouped list of strings.
  /// This function loads the strings into the export vector<string>
  /// expanded_list, and removes the key-value pair from init_strings.
  /// An exception is thrown if key in not in init_strings, if the
  /// associated value does not contain at least one element, of if there
  /// is a structural problem with the grouped list.  Elements of the
  /// associated value may either be single string items, or else grouped
  /// sublists of the form {x1 x2 ... xn R} where x1 through xn are
  /// strings, and R is an integer repeat count.  The sublists are
  /// expanded recursively, so, for example, xi can also be either an
  /// individual value or a grouped sublist of the same form.  As an
  /// example, the value "fred {barney {wilma betty 3} 2}" would be
  /// expanded into "fred barney wilma betty wilma betty wilma betty
  /// barney wilma betty wilma betty wilma betty".  Note that there is no
  /// group count expected or allowed at the top level, i.e., to get the
  /// expanded list "fred fred fred" an extra level of grouping is
  /// necessary, i.e., "{fred 3}" is correct, not "fred 3".
  ///   IMPORTANT NOTE: For this function to work properly, each string
  /// item must be distinguishable from a multi-item list.  In particular,
  /// this function shouldn't be used with filenames, because a filename
  /// with an embedded space looks like a two item list.  We could allow
  /// some escape mechanism, for example, perhaps a repeat count=0 means
  /// concat the earlier list items and treat them as a single item with
  /// no grouping, but this is not presently implemented.
  void GetGroupedRealListInitValue(const string& key,
				   vector<REAL8m>& expanded_list);
  void GetGroupedIntListInitValue(const string& key,
				  vector<INT4m>& expanded_list);
  void GetGroupedUIntListInitValue(const string& key,
				   vector<UINT4m>& expanded_list);
  /// Analogous to GetGroupedStringListInitValue(), these routines
  /// return vectors of REAL8m's, INT4m's and UINT4m's, respectively.
  /// Exceptions are thrown for the same type of grouped list structural
  /// problems as GetGroupedStringListInitValue(), but in addition an
  /// exception is thrown if any element in the expanded list cannot be
  /// completely converted to the specified target type.

  void FindOrCreateObject(const vector<string>& param,
			  Oxs_OwnedPointer<Oxs_Ext>& obj) const;
  /// If param.size()==1, then queries the director for an already
  /// existing, registered Oxs_Ext object by the name param[0].
  /// In this case, the returned obj value points to the pre-existing
  /// object, and ownership is set false.  Otherwise, uses param to
  /// generate a new Oxs_Ext object, and obj points to that with
  /// ownership set true.  In this case it is the responsibility of the
  /// calling routine to see to the new object's eventual destruction.
  /// On failure, this routine will throw an exception.  In no case is
  /// the return obj value a NULL pointer.

public:
  struct Error {
  protected:
    string msg;
  public:
    Error(const char* errmsg)
      : msg("Oxs_Ext ERROR: "
	    + string(errmsg)) {}

    Error(const Oxs_Ext* obj,const char* errmsg)
      : msg("Oxs_Ext ERROR in object "
	    + string(obj->InstanceName())
	    + ": "
	    + string(errmsg)) {}

    Error(const Oxs_Ext* obj,const string& errmsg)
      : msg("Oxs_Ext ERROR in object "
	    + string(obj->InstanceName())
	    + ": "
	    + errmsg) {}

    operator string() const { return msg; }

    const char* c_str() const { return msg.c_str(); }

    void Prepend(const string& prefix) {
      msg = prefix + msg;
    }

  };

  // A couple specializations of Oxs_Ext::Error for MakeNew().  We
  // don't need Error(const Oxs_Ext*, const char*) constructor support,
  // because there is no *this available inside static MakeNew().
  struct UnregisteredType : public Error {
    UnregisteredType(const char* errmsg) : Error(errmsg) {}
  };
  struct ConstructFailure : public Error {
    ConstructFailure(const char* errmsg) : Error("") {
      msg = errmsg; // Pass errmsg up without prefixing
      /// with an "Oxs_Ext ERROR" tag.  To see why this
      /// is desirable, check the MakeNew() source---the
      /// only place ConstructFailure occurs it is as a
      /// rethrow of an Oxs_Ext::Error.
    }
  };


  const char* InstanceName() const { return instanceId.c_str(); }
  virtual const char* ClassName() const =0;    // Child class name

  virtual BOOL Init() =0;  // This can be called at any time to reset
  /// an Ext object to its initial, post constructor state.  The
  /// constructor should set all static object state information, and
  /// then call this function to initialize any state information that
  /// might change during the lifetime of the object.  It is incumbent
  /// upon the constructor to save whatever information is necessary for
  /// Init() to completely reset itself.  Return value of 1 indicates
  /// success, 0 indicates error.  Perhaps this should be changed to
  /// void return, with an exception thrown on error?

  static Oxs_Ext* MakeNew(const char* idstring, // Child object name
			  Oxs_Director* dtr,    // App director
			  const char* argstr);  // MIF input block params
  /// NOTE 1: The safe interpreter is guaranteed valid throughout the
  /// life of the Oxs_Ext object.  This is the same interpreter used
  /// by the Oxs_Mif object to source the input MIF file.  MakeNew()
  /// throws an Oxs_Ext::UnregisteredType exception if idstring
  /// requests an unknown class type, or an Oxs_Ext::ConstructFailure
  /// exception if object construction failed.  The latter is most
  /// likely due to bad argstr.  Otherwise, the return value is
  /// guaranteed to be non-NULL, and to point to a object of the
  /// specified type.  It is the responsibility of the client to delete
  /// the object when finished with it.

  static string ListRegisteredChildClasses(); // This is intended
  /// for debugging purposes only.

  virtual void StageRequestCount(unsigned int& min,
				 unsigned int& max) const;
  // Number of stages object wants.  UINT_MAX indicates infinity.
  // If min is larger than max, then an error may be raised by the
  // driver, depending on the value of its parameter stage_count_check.
  // If checking is disabled, then the number of stages run will
  // be the largest "min" value returned by all StageRequestCount
  // calls, so in the case of non-fatal stage count inconsistencies
  // inside an Oxs_Ext count, the min value should be set to the best
  // stage count guess.  In this case, "max" doesn't really matter,
  // as long as it is less than min.
  //  The default implementation returns (0,UINT_MAX), which means
  // no restriction.

  virtual ~Oxs_Ext();
};


#endif // _OXS_EXT
