/* FILE: simstate.h                 -*-Mode: c++-*-
 *
 * Simulation state class.
 *
 */

#ifndef _OXS_SIMSTATE
#define _OXS_SIMSTATE

#include "oc.h"
#include "lock.h"
#include "mesh.h"
#include "meshvalue.h"
#include "threevector.h"
#include <vector>

/* End includes */

class Oxs_SimState : public Oxs_Lock {
  // Perhaps the inheritance from Oxs_Lock should be private, but then
  // exposing the Oxs_Lock member functions is awkward.
  // Note: To maintain a clean dependency hierarchy and avoid cyclic
  //  dependencies, Oxs_SimState objects should not contain any keys,
  //  because Oxs_SimState objects are held by Oxs_Ext objects which may
  //  have locks places on them by other Oxs_Ext objects.  Therefore,
  //  Oxs_SimState objects are placed at the root of the dependency
  //  hierarchy, and are deleted by Oxs_Director only after all Oxs_Ext
  //  objects are destroyed.  It is therefore the responsibility of each
  //  object holding a lock on an Oxs_SimState object to place locks on
  //  or otherwise guarantee the correctness of all pointers held within
  //  the Oxs_SimState object.
private:
  mutable map<string,REAL8m> derived_data;
public:
  Oxs_SimState();
  ~Oxs_SimState();

  void Reset(); // Called by Oxs_Director to reuse state structure.
  /// Perhaps this should be private, and Oxs_Director made friend?
  /// Regardless, the (potentially large) Oxs_MeshValue array "spin"
  /// is not touched, for reasons of efficiency.  However, for
  /// debugging checks one might want to wipe "spin" too.

  // The following elements should perhaps be wrapped up in a
  // STL map array, and indexed by name.
  UINT4m previous_state_id; // 0 if no previous state
  UINT4m iteration_count;  // Total number of distinct spin configurations
  UINT4m stage_number;     // Stage index
  UINT4m stage_iteration_count; // Number of iterations for this stage
  REAL8m stage_start_time; // Time index corresponding to current stage
  /// with stage_iteration_count == 0
  REAL8m stage_elapsed_time; // Time elapsed in current stage.
  /// NOTE: total elapsed time = stage_start_time + stage_elapsed_time.
  REAL8m last_timestep; // Seconds between this state and previous state.

  // It is the responsibility of any object using this Oxs_SimState
  // object to insure the validity of these pointers.
  const Oxs_MeshValue<REAL8m>* Ms;
  const Oxs_MeshValue<REAL8m>* Ms_inverse; // 1/Ms
  const Oxs_Mesh* mesh;

  Oxs_MeshValue<ThreeVector> spin;

  // Status done/not_done cache.  The driver Done and StageDone methods
  // should set these fields appropriately.  Caching done status can
  // help protect against cache thrashing for derived data stored outside
  // of the state.  Perhaps this should be moved into the DerivedData area?
  enum SimStateStatus { UNKNOWN, DONE, NOT_DONE };
  mutable SimStateStatus stage_done;
  mutable SimStateStatus run_done;

  // AddDerivedData and GetDerivedData return 1 if operation
  // was successful, 0 otherwise.  In particular, AddDerivedData
  // fails if name is already used.
  BOOL AddDerivedData(const char* name,REAL8m value) const;
  BOOL GetDerivedData(const char* name,REAL8m& value) const;
  void ListDerivedData(vector<string>& names) const;
  void ClearDerivedData();

  // IMPORTANT NOTICE:  Any changes to Oxs_SimState's public interface
  // should be reflected in Oxs_Driver and children state initialization
  // functions.
};

#endif // _OXS_SIMSTATE
