/* FILE: threevector.cc        -*-Mode: c++-*-
 *
 * Simple 3D vector class.  *All* client code should use the
 * ThreeVector typedef in case this class gets moved into an
 * OOMMF extension library.
 *
 */

#include <math.h>

#include "nb.h"
#include "threevector.h"

/* End includes */


const Oxs_ThreeVector operator^(const Oxs_ThreeVector& v,
				const Oxs_ThreeVector& w)
{ // Cross product
  REAL8m x = v.y * w.z  -  w.y * v.z;
  REAL8m y = w.x * v.z  -  v.x * w.z;
  REAL8m z = v.x * w.y  -  w.x * v.y;
  return Oxs_ThreeVector(x,y,z);
}

Oxs_ThreeVector& Oxs_ThreeVector::operator^=
(const Oxs_ThreeVector& w)
{ // Cross product.  Implemented by calling the binary operator,
  // because in-place computation doesn't save much (we need to
  // save results of the first two equations in temporaries).
  *this = (*this ^ w);
  return *this;
}

void Oxs_ThreeVector::SetMag(REAL8m mag)
{ // Adjusts size to "mag".  We assume MagSq() doesn't over/underflow.
  // This is probably okay for our intended uses, but if we want to
  // bullet-proof we should do some preemptive prescaling.
  // Cf. James L. Blue, "A Portable Fortran Program to Find the
  // Euclidean Norm of a Vector," ACM Transactions on Mathematical
  // Software, 4, 15-23 (1978).
  REAL8m magsq=MagSq();
  if(magsq<=0) {
    Random(mag);
  } else {
    REAL8m mult = mag/sqrt(magsq);
    x*=mult; y*=mult; z*=mult;
  }
}

void Oxs_ThreeVector::MakeUnit()
{ // Conceptually equivalent to SetMag(1.0), but
  // with enhanced accuracy.
  REAL8m magsq=MagSq();
  if(magsq<=0) {
    Random(1.0);
    magsq=MagSq();
  }
  REAL8m error = 1-magsq;
  if(fabs(error)<=2*REAL8_EPSILON) {
    // In general, we can't do better than this, so we might as
    // well stop.  Moreover, kicking out on this condition helps
    // limit "chatter," i.e., without this check repeated calls
    // into this routine will yield slightly different results,
    // causing (x,y,z) to wander around like a drunk looking for
    // his house keys.
    return;
  }
  else if(fabs(error)>CUBE_ROOT_REAL8_EPSILON) {
    // Big error; fall back to making a sqrt evaluation.
    // REAL8_EPSILON is about 2e-16 with IEEE floating pt,
    // so CUBE_ROOT_REAL8_EPSILON is 6e-6.  See mjd NOTES II,
    // 10-Feb-2001, p 91.
    REAL8m mult = 1.0/sqrt(magsq);
    x*=mult; y*=mult; z*=mult;
    magsq=MagSq();
    error = 1-magsq;
  }

  // Do one Halley step (see mjd's NOTES II, 8-Feb-2001, p82)
  // to improve accuracy to <= 2*REAL8_EPSILON
  REAL8m adj = 2*error/(1+3*magsq);
  x += adj*x;   y += adj*y;   z += adj*z;
  /// v += adj*v produces smaller | ||v||^2 - 1 | than
  /// v *= (1+adj).
}

void Oxs_ThreeVector::Random(REAL8m mag)
{ // Makes a random vector of size "mag"
  REAL8m r=2*Oc_UnifRand()-1;
  REAL8m theta=r*PI;

  REAL8m costheta=1.0,sintheta=0.0;

  if(fabs(r)>0.25 && fabs(r)<0.75) {
    // cosine is small and accurate
    costheta=cos(theta);
    sintheta=sqrt(OC_MAX(0.,1.-OC_SQ(costheta)));
    if(r<0) sintheta *= -1;
  } else {
    // sine is small and accurate
    sintheta=sin(theta);
    costheta=sqrt(OC_MAX(0.,1.-OC_SQ(sintheta)));
    if(fabs(r)>0.5) costheta *= -1;
  }

  REAL8m cosphi=1-2*Oc_UnifRand();
  REAL8m sinphi=sqrt(OC_MAX(0.,1.-OC_SQ(cosphi)));

  x=mag*sinphi*costheta;
  y=mag*sinphi*sintheta;
  z=mag*cosphi;
}

