/* FILE: cgevolve.cc                 -*-Mode: c++-*-
 *
 * Concrete minimization evolver class, using conjugate gradients
 *
 */

#include <float.h>

#include "nb.h"
#include "director.h"
#include "mindriver.h"
#include "simstate.h"
#include "cgevolve.h"
#include "key.h"
#include "threevector.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_CGEvolve);

/* End includes */

// Constructor
Oxs_CGEvolve::Oxs_CGEvolve(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_MinEvolver(name,newdtr,argstr),
    energy_calc_count(0),
    cycle_count(0),
    base_direction_max_mag(0.),
    cache_state_id(0),
    bracket_flag(0),bracket_count(0),
    scaled_minimum_bracket_step(0.),
    scaled_maximum_bracket_step(0.),
    bracket_start_step(0.),
    line_minimum_flag(0),line_minimum_count(0),
    last_minimum_reduction_ratio(0.0),
    next_to_last_minimum_reduction_ratio(0.0),
    total_H_field_output(-5),
    mxHxm_output(-5),
    energy_calc_count_output(-5),
    cycle_count_output(-5),
    max_mxHxm_output(-5),
    delta_E_output(-5),
    total_energy_output(-5),
    bracket_count_output(-5),
    line_min_count_output(-5)
{
  // Process arguments

  gradient_reset_count = GetUIntInitValue("gradient_reset_count",42);

  REAL8m min_ang = GetRealInitValue("minimum_bracket_step",0.05);
  min_ang *= PI/180.; // Convert from degrees to radians
  minimum_bracket_step = tan(min_ang);

  REAL8m max_ang = GetRealInitValue("maximum_bracket_step",10);
  max_ang *= PI/180.; // Convert from degrees to radians
  maximum_bracket_step = tan(max_ang);

  line_minimum_relwidth = GetRealInitValue("line_minimum_relwidth",1e-6);

  VerifyAllInitArgsUsed();
}

BOOL Oxs_CGEvolve::Init()
{
  // Setup output.  Note: MSVC++ 6.0 requires fully qualified
  // member function names.
  total_H_field_output.Setup(this,InstanceName(),"H","A/m",1,
              &Oxs_CGEvolve::UpdateDerivedFieldOutputs,director);
  mxHxm_output.Setup(this,InstanceName(),"mxHxm","A/m",1,
	      &Oxs_CGEvolve::UpdateDerivedFieldOutputs,director);
  max_mxHxm_output.Setup(this,InstanceName(),"Max mxHxm","A/m",0,
	      &Oxs_CGEvolve::UpdateDerivedOutputs,director);
  total_energy_output.Setup(this,InstanceName(),"Total energy","J",0,
              &Oxs_CGEvolve::UpdateDerivedOutputs,director);
  delta_E_output.Setup(this,InstanceName(),"Delta E","J",0,
              &Oxs_CGEvolve::UpdateDerivedOutputs,director);
  bracket_count_output.Setup(this,InstanceName(),"Bracket count","",0,
              &Oxs_CGEvolve::UpdateDerivedOutputs,director);
  line_min_count_output.Setup(this,InstanceName(),"Line min count","",0,
              &Oxs_CGEvolve::UpdateDerivedOutputs,director);
  cycle_count_output.Setup(this,InstanceName(),
              "Cycle count","",0,
              &Oxs_CGEvolve::UpdateDerivedOutputs,
              director);
  energy_calc_count_output.Setup(this,InstanceName(),
              "Energy calc count","",0,
              &Oxs_CGEvolve::UpdateDerivedOutputs,
              director);

  // Initialize instance variables
  energy_calc_count=0;
  cycle_count=0;
  base_direction_max_mag=0.;  // Safety
  cache_state_id=0;
  bracket_flag=0;
  bracket_count=0;
  scaled_minimum_bracket_step = 0.;
  scaled_maximum_bracket_step = 0.;
  line_minimum_flag=0;
  line_minimum_count=0;
  last_minimum_reduction_ratio=0.0;
  next_to_last_minimum_reduction_ratio=0.0;

  base_state.Release();

  return 1;
}


Oxs_CGEvolve::~Oxs_CGEvolve()
{}

void Oxs_CGEvolve::GetEnergyAndmxHxm(
 const Oxs_SimState* state,          // Import
 Oxs_MeshValue<REAL8m>* export_energy,      // Export
 Oxs_MeshValue<ThreeVector>* export_mxHxm)  // Export
{ // Also fills cache_energy and cache_mxHxm, and
  // sets "Max mxHxm" and "Total energy" derived
  // data in state.
  //   Exports "energy" and "mxHxm" may be set to NULL;
  // Side effects (cache and derived data filling) still
  // occur.
  // Note: (mxH)xm = mx(Hxm)

  UINT4m i;

  // For convenience
  const Oxs_MeshValue<ThreeVector>& spin = state->spin;
  const Oxs_MeshValue<REAL8m>& Ms = *(state->Ms);
  const Oxs_Mesh* mesh = state->mesh;
  UINT4m vecsize = mesh->Size();

  if(cache_state_id != state->Id() ||
     (total_H_field_output.GetCacheRequestCount()>0
      && total_H_field_output.cache.state_id != state->Id())) {
    // Calculate energy & mxHxm from scratch

    ++energy_calc_count;    // Update call count

    // Cache space
    cache_state_id = 0;
    cache_energy.AdjustSize(mesh);
    cache_mxHxm.AdjustSize(mesh);

    // Temp space
    scratch_energy.AdjustSize(mesh);
    scratch_field.AdjustSize(mesh);
    REAL8m max_torque_sq=0.0;

    UINT4m energy_obj_count = director->GetEnergyObjCount();
    UINT4m energy_obj_index = 0;
    if(energy_obj_index<energy_obj_count) {
      Oxs_EnergyData oed(*state);
      oed.energy_buffer = &scratch_energy;
      oed.field_buffer  = &scratch_field;
      director->GetEnergyObj(energy_obj_index)->GetEnergyData(*state,oed);
      if(oed.pE_pt.IsSet() && oed.pE_pt != 0.0) {
	string msg = 
	  string("Oxs_CGEvolve::GetEnergyAndField: Energy object ")
	  + director->GetEnergyObj(energy_obj_index)->InstanceName()
	  + string(" is time varying; this property is not supported "
		   " my minimization evolvers.");
	throw Oxs_Ext::Error(this,msg.c_str());
      }
      cache_energy = *oed.energy;
      if(total_H_field_output.GetCacheRequestCount()>0) {
	total_H_field_output.cache.state_id = 0;
	total_H_field_output.cache.value = *oed.field_buffer;
      }
      for(i=0;i<vecsize;++i) {
	cache_mxHxm[i] = spin[i];
	cache_mxHxm[i] ^= (*oed.field)[i]; // Just mxH at this point
      }
      while(++energy_obj_index<energy_obj_count) {
	Oxs_EnergyData oed(*state);
	oed.energy_buffer = &scratch_energy;
	oed.field_buffer  = &scratch_field;
	director->GetEnergyObj(energy_obj_index)->GetEnergyData(*state,oed);
	cache_energy += *oed.energy;
	if(total_H_field_output.GetCacheRequestCount()>0) {
	  total_H_field_output.cache.value += *oed.field_buffer;
	}
	for(i=0;i<vecsize;++i) {
	  ThreeVector temp = spin[i];
	  temp ^= (*oed.field)[i];
	  cache_mxHxm[i] += temp; // Really just mxH at this point
	}
      }
      // Zero torque at fixed spin sites.
      UpdateFixedSpinList(mesh);
      const UINT4m fixed_count = GetFixedSpinCount();
      for(UINT4m j=0;j<fixed_count;j++) {
	cache_mxHxm[GetFixedSpin(j)].Set(0.,0.,0.);
      }
      // cache_mxHxm holds mxH, so cross against m again, and
      // calculate max |mxHxm|.
      for(i=0;i<vecsize;++i) {
	if(Ms[i]==0.0) { // Ignore zero-moment spins
	  cache_mxHxm[i].Set(0.,0.,0.);
	} else {
	  REAL8m torque_sq = cache_mxHxm[i].MagSq();
	  if(torque_sq>max_torque_sq) max_torque_sq = torque_sq;
	  cache_mxHxm[i] ^= spin[i];
	}
      }
    } else {
      // No energy terms!  (Safety)
      cache_energy = 0;
      cache_mxHxm = ThreeVector(0.,0.,0.);
      if(total_H_field_output.GetCacheRequestCount()>0) {
	total_H_field_output.cache.state_id = 0;
	total_H_field_output.cache.value = ThreeVector(0.,0.,0.);
      }
    }
    cache_state_id = state->Id();
    if(total_H_field_output.GetCacheRequestCount()>0) {
      total_H_field_output.cache.state_id = state->Id();
    }

    // Fill supplemental derived data.
    state->AddDerivedData("Max mxHxm",sqrt(max_torque_sq));
    Nb_Xpfloat total_energy = 0.0;
    for(i=0;i<vecsize;++i) {
      total_energy.Accum(cache_energy[i] * mesh->Volume(i));
    }
    state->AddDerivedData("Total energy",total_energy.GetValue());
    state->AddDerivedData("Bracket count",
			  REAL8m(bracket_count));
    state->AddDerivedData("Line min count",
			  REAL8m(line_minimum_count));
    state->AddDerivedData("Energy calc count",
			  REAL8m(energy_calc_count));
    state->AddDerivedData("Cycle count",
			  REAL8m(cycle_count));
  }

  // At this point cache_energy and cache_mxHxm are filled, either
  // from just above, or by earlier call that had been saved in
  // cache.  It remains to copy from cache into export structures,
  // as requested.
  if(export_mxHxm!=NULL)  *export_mxHxm = cache_mxHxm;
  if(export_energy!=NULL) *export_energy = cache_energy;
}
  
void Oxs_CGEvolve::GetRelativeEnergyAndDerivative(
 const Oxs_SimState* state, // Import
 REAL8m time_offset,        // Import
 REAL8m &relenergy,         // Export
 REAL8m &derivative)        // Export
{ // Calculates total energy relative to that of base_state,
  // and derivative in base_direction (-mu0.H*base_direction).
  // The base_direction and base_energy arrays _must_ be set
  // before calling this function.

  // Uses mxHxm instead of H in calculation of derivative, because
  // energy is calculated off of normalized m, so component of H in m
  // direction doesn't actually have any effect.  Plus, experiments
  // appear to show generally faster convergence with mxHxm than with H.

  GetEnergyAndmxHxm(state,NULL,NULL);

  // Set up some convenience variables
  const Oxs_Mesh* mesh = state->mesh; // For convenience
  const Oxs_MeshValue<REAL8m>& Ms = *(state->Ms);

  UINT4m i;
  UINT4m vecsize = mesh->Size();

  Nb_Xpfloat etemp = 0.0;
  for(i=0;i<vecsize;++i) {
    etemp.Accum((cache_energy[i] - base_energy[i]) * mesh->Volume(i));
  }
  relenergy = etemp.GetValue();
  state->AddDerivedData("Relative energy",relenergy);
  state->AddDerivedData("Energy base state id",
			static_cast<REAL8m>(base_state.GetPtr()->Id()));

  Nb_Xpfloat dtemp = 0.0;
  REAL8m time_offset_sq = time_offset*time_offset;
  for(i=0;i<vecsize;++i) {
    const ThreeVector& vtemp = base_direction[i];
    REAL8m scale_adj = sqrt(1+time_offset_sq * vtemp.MagSq());
    dtemp.Accum((cache_mxHxm[i]*vtemp)
		 * Ms[i] * mesh->Volume(i)/scale_adj);
  }
  derivative = -MU0 * dtemp.GetValue();
  /// See mjd's NOTES II, 29-May-2002, p156, which includes
  /// the derivation of the scale_adj term above.
}

void Oxs_CGEvolve::FillBracket(REAL8m offset,
			       const Oxs_SimState* oldstateptr,
			       Oxs_Key<Oxs_SimState>& statekey,
			       Bracket& endpt)
{
  try {
    Oxs_SimState& workstate
      = statekey.GetWriteReference(); // Write lock
    Oxs_MeshValue<ThreeVector>& spin = workstate.spin;
    const Oxs_MeshValue<ThreeVector>& base_spin
      = base_state.GetPtr()->spin;
    UINT4m vecsize = workstate.mesh->Size();
    for(UINT4m i=0;i<vecsize;i++) {
      ThreeVector temp = base_spin[i] + offset * base_direction[i];
      temp.MakeUnit();  // Do we want this?
      spin[i] = temp;
    }
    workstate.iteration_count = oldstateptr->iteration_count + 1;
    workstate.stage_iteration_count
      = oldstateptr->stage_iteration_count + 1;

    const Oxs_SimState& newstate
      = statekey.GetDepReference(); // Release write lock
    GetRelativeEnergyAndDerivative(&newstate,offset,
				   endpt.E,endpt.Ep);
    endpt.offset=offset;
  }
  catch (...) {
    statekey.GetReadReference(); // Release write lock
    throw;
  }
}


BOOL Oxs_CGEvolve::FindBracketStep(const Oxs_SimState* oldstate,
				   Oxs_Key<Oxs_SimState>& statekey)
{ // Takes one step; assumes rightpt.offset >= leftpt.offset and
  // minimum to be bracketed is to the right of rightpt.offset
  REAL8m offset = rightpt.offset;
  if(offset < bracket_start_step) {
    offset = bracket_start_step;
  } else {
    if(offset<leftpt.offset+scaled_minimum_bracket_step) {
      offset = leftpt.offset*2 + scaled_minimum_bracket_step;
    } else {
      // Try to use derivate information at existing endpoints
      // to estimate location of minimum.
      REAL8m temp1 = (leftpt.offset-rightpt.offset)*leftpt.Ep;
      REAL8m temp2 = rightpt.Ep-leftpt.Ep;
      if(temp2<=0) {
	offset *=4; // Aiee! We're rolling off a cliff!
      } else {
	REAL8m temp3 = 8*rightpt.offset-leftpt.offset;
	if(2*temp1 < temp3*temp2) { // Overflow check
	  offset = leftpt.offset + 2*temp1/temp2;
	} else {
	  offset = 8*rightpt.offset;
	}
      }
    }
  }
  if(offset < 2*rightpt.offset) {
    offset = 2*rightpt.offset;
  }
  if(offset > scaled_maximum_bracket_step) {
    offset = scaled_maximum_bracket_step;
  }

  // Zero span check
  if(offset <= rightpt.offset) {
    // Proposed bracket span increase is floating point 0.
    if(rightpt.offset>0) {
      offset = rightpt.offset*(1+16*REAL8_EPSILON);
    } else {
      if(scaled_maximum_bracket_step>0) {
	offset = scaled_maximum_bracket_step;
      } else {
	// Punt
	offset = REAL8_EPSILON;
      }
    }
  }
  if(offset*base_direction_max_mag<16*REAL8_EPSILON) {
    // If offset*base_direction<REAL8_EPSILON, and we assume
    // that the base spins are unit vectors, then there is a
    // good chance that the proposed offset won't actually
    // change any of the discretized spins.  For bracketing
    // purposes, it shouldn't hurt to bump this up a bit.
    if(base_direction_max_mag >= 0.5 ||
       DBL_MAX*base_direction_max_mag > 16*REAL8_EPSILON) {
      offset = 16*REAL8_EPSILON/base_direction_max_mag;
    }
  }


  Bracket temppt;
  ++bracket_count;
  FillBracket(offset,oldstate,statekey,temppt);

  // Shift leftpt to the right.  This shift is optional,
  // but the resulting decrease in interval size should
  // help subsequent calls to the derivative-based minimum
  // location estimator above, and also help in the line
  // minimizer routine once the minimum is bracketed.
  // OTOH, one might want to add some check to see that
  // E is decreasing as predicted; if not, then it may
  // be likely that we've overstepped the minimum.
  if(rightpt.E<leftpt.E && rightpt.Ep<0) { // This should always be
    /// true, since otherwise (leftpt,rightpt] already bracketed the
    /// minimum, but check anyway for robustness.
    leftpt=rightpt;
  }

  // Classify situation.  We always require leftpt have Ep<0,
  // so there are two possibilities:
  //  1) temppt.E>leftpt.E or temppt.Ep>=0 ==> Minimum is bracketed
  //  2) temppt.E<leftpt.E and temppt.Ep<0 ==> Minimum not bracketed
  if(temppt.E<leftpt.E && temppt.Ep<0) {
    // Minimum not bracketed.  Shift brackets to the right.
    rightpt=temppt;
    if(rightpt.offset >= scaled_maximum_bracket_step) {
      // Unable to bracket minimum inside allowed range.
      // So set bracket_flag true to force acceptance,
      // and shift leftpt to rightpt to disable line
      // minimizer.
      bracket_flag=1;
      leftpt=rightpt;
    }
  } else {
    // Minimum is bracketed.
    rightpt=temppt;
    bracket_flag=1;
  }
  return 1; // FillBracket call fills statekey with new state.
}

BOOL Oxs_CGEvolve::FindLineMinimumStep(const Oxs_SimState* oldstate,
				       Oxs_Key<Oxs_SimState>& statekey)
{ // Compresses bracket span.
  // Assumes coming in and forces on exit
  // requirement that leftpt.Ep<0.
  REAL8m span = rightpt.offset - leftpt.offset; 

  if(span<=0.0 // Empty interval; assume minimum is reached.
     || leftpt.Ep>=0.) { // ==0 means minimum found. >0 is an error,
                        /// but include anyway for robustness.
    line_minimum_flag=1;
    last_minimum_reduction_ratio=0.0;
    next_to_last_minimum_reduction_ratio=0.0;
    minpt = leftpt;
    return 0;  // No step
  }

  if(rightpt.Ep==0.) { // ==0 means minimum found.
    line_minimum_flag=1;
    last_minimum_reduction_ratio=0.0;
    next_to_last_minimum_reduction_ratio=0.0;
    minpt = rightpt;
    return 0;  // No step
  }

  // Pick test point.
  REAL8m test_offset;
  REAL8m lambda = 0.5; // Default
  Bracket testpt;
  REAL8m lEp = leftpt.Ep * span;
  REAL8m rEp = rightpt.Ep * span;
  REAL8m Ediff = rightpt.E - leftpt.E;
  BOOL poor_data = 0;
  if(fabs(Ediff)<=1e-10*(fabs(leftpt.E)+fabs(rightpt.E))) {
    poor_data = 1; // Data is likely unreliable
  }
  if(!poor_data) {
    // Cubic fit.  See mjd's NOTES II, 1-Sep-2001, p134-136.
    REAL8m a = -2*Ediff +   lEp + rEp;
    REAL8m b =  3*Ediff - 2*lEp - rEp;
    REAL8m c = lEp;
    if(a==0.0) {
      // Quadratic
      if(b!=0.0) { // Safety check. b should be >= -c/2 > 0.
	lambda = -c/(2*b);
      }
    } else {
      REAL8m disc = b*b - 3*a*c;
      if(disc>=0.0) { // Safety check.  See NOTES II, 1-Sep-2001, p135.
	disc = sqrt(disc);
	if(b>=0.) {
	lambda = -c/(b + disc);
	} else {
	  lambda = (-b + disc)/(3*a);
	}
      }
    }
  }
	      
  // Check aggressive reductions and cases with suspect
  // data against lower order method.
  const double AGGRESSIVE_FIT = 0.0625;
  if(poor_data ||
     lambda<AGGRESSIVE_FIT || lambda>1-AGGRESSIVE_FIT) {
    // Calculate lambda using lower-order method
    REAL8m alt_lambda = 0.5;
    if(rEp<0.) {
      // Overly large bracket, or bad rightpt.Ep.
      // Guess at minpt using leftpt.E/Ep and rightpt.E
      const REAL8m reduce_limit = AGGRESSIVE_FIT*0.25; // The
      /// situation rightpt.Ep<0 is a bit suspicious, so limit
      /// reduction.
      REAL8m numerator = -1.0 * lEp; // This is > 0.
      REAL8m denominator = 2*(Ediff - lEp);
      // denominator must also be >0, because rightpt.E>=leftpt.E
      // if a minimum is bracketed with rightpt.Ep<0.
      denominator = fabs(denominator); // But play it safe, because
      /// checks below depend on denominator>0.
      if(numerator<reduce_limit*denominator) {
	alt_lambda = reduce_limit;
      } else if(numerator>(1-reduce_limit)*denominator) {
	alt_lambda = 1-reduce_limit;
      } else {
	alt_lambda = numerator/denominator;
      }
    } else {
      // Use just derivative information, i.e., quadratic fit.
      REAL8m Ep_diff = rightpt.Ep - leftpt.Ep; // Since leftpt.Ep<0,
      /// Ep_diff = |rightpt.Ep| + |leftpt.Ep|.
      alt_lambda = -leftpt.Ep/Ep_diff;
      // Provided Ep_diff is a valid floating point number,
      // then lambda should not overflow, and in fact must
      // lay inside [0,1].
    } // matches if(rightpt.Ep<0.)

    // Be conservative
    if(poor_data) {
      lambda = alt_lambda;
    } else if(lambda<=0.5) {
      if (lambda<alt_lambda) lambda = alt_lambda;
    } else {
      if (lambda>alt_lambda) lambda = alt_lambda;
    }
  } // Matches if(lambda<AGGRESSIVE_FIT || lambda>1-AGGRESSIVE_FIT)

  // Restrict reduction.
  REAL8m max_reduce=0.5; // Don't set this any larger than 0.707
  REAL8m last_reduce = 1.0-last_minimum_reduction_ratio;
  REAL8m next_to_last_reduce = 1.0-next_to_last_minimum_reduction_ratio;
  if(last_reduce<max_reduce) max_reduce=last_reduce;
  if(next_to_last_reduce<max_reduce) max_reduce=next_to_last_reduce;
  max_reduce *= max_reduce;
  if(max_reduce<REAL8_EPSILON) max_reduce=REAL8_EPSILON;
  if(span*max_reduce < REAL8_EPSILON * rightpt.offset) {
    REAL8m temp = REAL8_EPSILON * rightpt.offset;
    if(0.5*span>temp) max_reduce = temp/span;
    else              max_reduce = 0.5;
  }
  if(max_reduce>0.5) max_reduce=0.5; // Safety; This shouldn't trigger.
  if(lambda<max_reduce)           lambda = max_reduce;
  else if(lambda>1.0-max_reduce)  lambda = 1.0-max_reduce;

  test_offset = leftpt.offset + lambda * span;
  if(test_offset<=leftpt.offset || test_offset>=rightpt.offset) {
    // Roundoff check
    lambda = 0.5;
    test_offset = 0.5 * (leftpt.offset + rightpt.offset);
    if(test_offset<=leftpt.offset || test_offset>=rightpt.offset) {
      // Interval width effectively machine 0; assume minimum reached
      line_minimum_flag=1;
      last_minimum_reduction_ratio=0.0;
      next_to_last_minimum_reduction_ratio=0.0;
      if(leftpt.E<=rightpt.E) minpt = leftpt;
      else                    minpt = rightpt;
      return 0;  // No step
    }
  }

  ++line_minimum_count;
  FillBracket(test_offset,oldstate,statekey,testpt);

  // Determine which interval, [left,test] or [test,right] to keep.
  next_to_last_minimum_reduction_ratio
    = last_minimum_reduction_ratio;
  if(testpt.E>leftpt.E || testpt.Ep>=0) {
    // leftpt + testpt bracket a minimum, so keep the left
    // interval.  Note: It could happen that the right interval
    // would also pass this test, but we give preference to the
    // lefthand minimum because it is closer to the basept.
    rightpt=testpt;
    last_minimum_reduction_ratio=1.0-lambda;
  } else {
    // Otherwise keep the right interval.  Note that in this
    // branch testpt.Ep<0, so we retain this requirement.
    leftpt=testpt;
    last_minimum_reduction_ratio=lambda;
  }

  // See if testpt is minimum
  if(rightpt.offset-leftpt.offset
     <= line_minimum_relwidth*rightpt.offset) {
    line_minimum_flag=1;
    if(leftpt.E<=rightpt.E) minpt = leftpt;
    else                    minpt = rightpt;
  }

  return 1; // FillBracket call fills statekey with new state.
}

BOOL Oxs_CGEvolve::Step(const Oxs_MinDriver* /* driver */,
			Oxs_ConstKey<Oxs_SimState> current_state_key,
			Oxs_Key<Oxs_SimState>& next_state_key)
{
  // Act 1: Pick line optimization direction
  const Oxs_SimState* cstate = current_state_key.GetPtr();
  const Oxs_SimState* bstate = base_state.GetPtr();
  if(bstate==NULL || bstate->stage_number!=cstate->stage_number
     || line_minimum_flag) {
    // Basept not set, invalid, or in need of updating;
    // set from current_state
    BOOL stage_change = 1;
    if(bstate!=NULL && bstate->stage_number==cstate->stage_number) {
      stage_change = 0;
    }
    base_state = current_state_key; // Copies key
    const Oxs_SimState* sptr
      = &(base_state.GetReadReference()); // Set and hold read lock

    const Oxs_MeshValue<REAL8m>& Ms = *(sptr->Ms); // For convenience
    const Oxs_Mesh* mesh = sptr->mesh; // For convenience
    const UINT4m size = mesh->Size();
    UINT4m i;
    base_mxHxm.AdjustSize(mesh);
    base_direction.AdjustSize(mesh);
    base_energy.AdjustSize(mesh);

    // Derive line optimization direction and initialize brackets
    REAL8m Ep=0.0;
    REAL8m base_direction_max_mag_new=0;
    if(stage_change || base_mxHxm.Size()==0
       || minpt.offset==scaled_maximum_bracket_step 
       || (cycle_count % gradient_reset_count ) == 0 ) {
      // Initialize direction sequence with gradient direction
      // If last attempted step was successful (i.e., reduced
      // total energy), then it will be stored in the cache,
      // and GetEnergyAndmxHxm will just copy cache values
      // into base_energy and base_mxHxm.  However, if minimum
      // value from previous cycle came from an endpoint of the
      // bracket interval other than the one added last, then
      // the cache value won't match sptr, and so the field
      // and energy arrays must be recalculated for that state.
      // Might want to think about caching both endpoints to
      // avoid this recalculation.  Or else always take guess
      // at better minimum estimate inside interval (halfway
      // point?) as final minimum point, in which case the
      // GetEnergyAndmxHxm call at cycle start will always involve
      // a field+energy evaluation, but at least it will be at a
      // new point instead of repeating a calculation from an old
      // point.  This comment applies also to the GetEnergyAndmxHxm
      // call in the conjugate gradient cycle initialization block
      // below.
      GetEnergyAndmxHxm(sptr,&base_energy,&base_mxHxm);
      base_direction = base_mxHxm;
      base_direction_max_mag_new = 0.0;
      base_mxHxm_mag_sq = 0.0;
      for(i=0;i<size;i++) {
	REAL8m magsq = base_mxHxm[i].MagSq();
	if(magsq>base_direction_max_mag_new)
	  base_direction_max_mag_new = magsq;
	base_mxHxm_mag_sq += magsq;
	Ep += magsq * Ms[i] * mesh->Volume(i);
	/// See mjd's NOTES II, 29-May-2002, p156.
      }
      base_direction_max_mag_new = sqrt(base_direction_max_mag_new);
      if(sptr->stage_iteration_count==0)
	base_direction_max_mag = base_direction_max_mag_new;
    } else { // Conjugate gradient
      // Direction[n] = mxHxm[n] + gamma*direction[n-1],
      // where 'n' is the cycle count,
      // and gamma = (mxHxm[n]-mxHxm[n-1])*mxHxm[n]/mxHxm[n-1]^2
      // (Polak-Ribiere; for Fletcher-Reeves drop the mxHxm[n-1]
      // term in the numerator for gamma.  This would eliminate
      // the need to store mxHxm[n-1], since otherwise only
      // |mxHxm[n-1]| is needed.)
      Oxs_MeshValue<ThreeVector>& mxHxm_new
	= scratch_field; // Introduced for code clarity
      // See note preceding GetEnergyAndmxHxm in gradient
      // initialization of cycle block above.  Applies here
      // as well.
      GetEnergyAndmxHxm(sptr,&base_energy,&mxHxm_new);
      REAL8m gamma=0;
      REAL8m mxHxm_mag_sq_new=0.;
      for(i=0;i<size;i++) {
	ThreeVector temp = mxHxm_new[i];
	mxHxm_mag_sq_new += temp.MagSq();
	//	temp -= base_mxHxm[i];  // Polak-Ribiere adjustment
	gamma += temp * mxHxm_new[i];
      }
      gamma /= base_mxHxm_mag_sq;
      const Oxs_MeshValue<ThreeVector>& spin = sptr->spin;
      for(i=0;i<size;i++) {
	ThreeVector temp = base_direction[i];
	temp *= gamma;
	temp += mxHxm_new[i];
	// Make temp orthgonal to spin[i]
	temp ^= spin[i];
	temp ^= spin[i];
	temp *= -1;
	base_direction[i] = temp;
	Ep += (temp*mxHxm_new[i]) * Ms[i] * mesh->Volume(i);
	/// See mjd's NOTES II, 29-May-2002, p156.
	REAL8m magsq = temp.MagSq();
	if(magsq>base_direction_max_mag_new)
	  base_direction_max_mag_new = magsq;
      }
      base_direction_max_mag_new = sqrt(base_direction_max_mag_new);
      base_mxHxm_mag_sq = mxHxm_mag_sq_new;
      base_mxHxm.Swap(mxHxm_new);
    }

    Ep *= -MU0;
    /// See mjd's NOTES II, 29-May-2002, p156.

    bracket_flag=0;
    leftpt.offset = rightpt.offset = 0.;
    leftpt.E      = rightpt.E      = 0.;
    leftpt.Ep     = rightpt.Ep     = Ep;
    REAL8m step_scaling = 1.0;
    REAL8m scaling_ratio=1.0;
    if(base_direction_max_mag_new>0.0) {
      step_scaling = 1.0/base_direction_max_mag_new;
      scaling_ratio = base_direction_max_mag * step_scaling;
    }
    scaled_minimum_bracket_step = step_scaling * minimum_bracket_step;
    scaled_maximum_bracket_step = step_scaling * maximum_bracket_step;
    if(line_minimum_flag && minpt.offset > 0) {
      bracket_start_step = minpt.offset*scaling_ratio;
      if(bracket_start_step>scaled_maximum_bracket_step)
	bracket_start_step = scaled_maximum_bracket_step;
    } else {
      bracket_start_step = scaled_minimum_bracket_step;
    }
    line_minimum_flag=0;
    // To force maximal restriction on initial bracket interval
    // reduction (see FindLineMinimumStep()), set both
    // minimum_reduction_ratio's to 0.  Moderately conservative
    // values are last_mrr=0.5, next_to_last_mrr=0.75.  Or set
    // both values to 1 to allow FindLineMinimumStep complete
    // freedom on first two steps.
    last_minimum_reduction_ratio=1.0;
    next_to_last_minimum_reduction_ratio=1.0;
    base_direction_max_mag = base_direction_max_mag_new;
    ++cycle_count;
  }

  BOOL step_taken=0;

  if(!bracket_flag) {
    // Act 2: Bracket minimum on line
    step_taken = FindBracketStep(cstate,next_state_key);
  } else {
    // Act 3: Find minimum on line
    step_taken = FindLineMinimumStep(cstate,next_state_key);
  }

  // Keep step?
  BOOL keep_step = 0;
  if(step_taken) {
    const Oxs_SimState* nstate = next_state_key.GetPtr();
    // Use "Relative energy" if possible, because that is what
    // the bracket and line minimization code uses, and is
    // presumably more accurate.
    REAL8m new_energy=0.0,last_energy=0.0;
    BOOL good_relative = 0;
    if(nstate->GetDerivedData("Relative energy",new_energy)) {
      // We have relative energy data for new state.  Check if
      // current state is either the base state, or else contains
      // relative energy data against the same base state.
      REAL8m base_id = -1.;
      nstate->GetDerivedData("Energy base state id",base_id);
      if(cstate->Id() == static_cast<UINT4m>(base_id)) {
	good_relative = 1;
	last_energy = 0.0;
      } else {
	REAL8m other_base_id = -2.;
	cstate->GetDerivedData("Energy base state id",other_base_id);
	if(other_base_id == base_id &&
	   cstate->GetDerivedData("Relative energy",last_energy)) {
	  good_relative = 1;
	}
      }
    }
    if(!good_relative) {
      nstate->GetDerivedData("Total energy",new_energy);
      nstate->GetDerivedData("Last energy",last_energy);
      // Is nstate's "Last energy" the same as cstate's
      // "Total energy"? Should we check?  If different,
      // which would we use?
    }
    if(new_energy<=last_energy) keep_step=1;
    //   We might want to check at this point if line minimum
    // has been obtained.  If so, and if step is not being kept,
    // then we could try to take another step inside the
    // bracket interval, because when the new cycle starts
    // we are going to re-calculate energy and field at the
    // old step, since it is not being held in cache_energy,
    // cache_field.  We could be extra conservative, i.e.,
    // make best quess using quadratic fit on Ep as to where
    // minimum is, and then take half the distance between
    // that point and the endpoint with smaller E.  That should
    // yield a point with high probability of even small E.
    //   However, it seems that usually by this point the
    // bracket interval is so small that the E and Ep values
    // are into discretization noise, so chance of success is
    // minimal and meaningless.
  }

  return keep_step;
}

void Oxs_CGEvolve::UpdateDerivedFieldOutputs(const Oxs_SimState& state)
{ // Fill Oxs_VectorOutput's that have CacheRequest enabled.
  GetEnergyAndmxHxm(&state,NULL,NULL);
  if(mxHxm_output.GetCacheRequestCount()>0) {
    mxHxm_output.cache.state_id = 0;
    mxHxm_output.cache.value    = cache_mxHxm;
    mxHxm_output.cache.state_id = cache_state_id;
  }
}

void Oxs_CGEvolve::UpdateDerivedOutputs(const Oxs_SimState& state)
{ // This routine fills all the Oxs_CGEvolve Oxs_ScalarOutput's to
  // the appropriate value based on the import "state".

  energy_calc_count_output.cache.state_id
    = cycle_count_output.cache.state_id
    = max_mxHxm_output.cache.state_id
    = delta_E_output.cache.state_id
    = total_energy_output.cache.state_id
    = bracket_count_output.cache.state_id
    = line_min_count_output.cache.state_id
    = 0;  // Mark change in progress

  REAL8m last_energy;
  if(!state.GetDerivedData("Energy calc count",
			   energy_calc_count_output.cache.value) ||
     !state.GetDerivedData("Cycle count",
			   cycle_count_output.cache.value) ||
     !state.GetDerivedData("Max mxHxm",max_mxHxm_output.cache.value) ||
     !state.GetDerivedData("Last energy",last_energy) ||
     !state.GetDerivedData("Total energy",
			   total_energy_output.cache.value) ||
     !state.GetDerivedData("Bracket count",
			   bracket_count_output.cache.value) ||
     !state.GetDerivedData("Line min count",
			   line_min_count_output.cache.value)) {
    // Missing at least some data
    GetEnergyAndmxHxm(&state,NULL,NULL);
    if(!state.GetDerivedData("Energy calc count",
			     energy_calc_count_output.cache.value) ||
       !state.GetDerivedData("Cycle count",
			     cycle_count_output.cache.value) ||
       !state.GetDerivedData("Max mxHxm",max_mxHxm_output.cache.value) ||
       !state.GetDerivedData("Last energy",last_energy) ||
       !state.GetDerivedData("Total energy",
			     total_energy_output.cache.value) ||
       !state.GetDerivedData("Bracket count",
			     bracket_count_output.cache.value) ||
       !state.GetDerivedData("Line min count",
			     line_min_count_output.cache.value)) {
      throw Oxs_Ext::Error(this,"Missing output data."
			   " Programming error?");
    }
  }
  delta_E_output.cache.value
    = total_energy_output.cache.value - last_energy;

  energy_calc_count_output.cache.state_id
    = cycle_count_output.cache.state_id
    = max_mxHxm_output.cache.state_id
    = delta_E_output.cache.state_id
    = total_energy_output.cache.state_id
    = bracket_count_output.cache.state_id
    = line_min_count_output.cache.state_id
    = state.Id();
}
