/* FILE: imageatlas.cc                 -*-Mode: c++-*-
 *
 * Atlas class derived from Oxs_Atlas class that uses an import
 * image file for region demarcation.
 */

#include <vector>
#include "nb.h"
#include "director.h"
#include "imageatlas.h"

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_ImageAtlas);

/* End includes */

// Support routines
BOOL OxsImageAtlasColorData::SetColor(const string& colorstr)
{
  return Nb_GetColor(colorstr.c_str(),red,green,blue);
}

REAL8m
OxsImageAtlasColorData::DistanceSq(REAL8m r,REAL8m g,REAL8m b) const
{
  r-=red;   r*=r;
  g-=green; g*=g;
  b-=blue;  b*=b;
  return r+g+b;
}

int
Oxs_ImageAtlas::ReadImage(const char* filename,
			  string& imagestring)
{
  FILE* fptr = Nb_FOpen(filename,"rb");
  if(fptr==NULL) return 1; // Unable to open file for reading

  imagestring.erase();

  char buf[16384];
  size_t bytes_read;

  bytes_read = fread(buf,sizeof(char),2,fptr);
  if(bytes_read!=2 || buf[0]!='P' || (buf[1]!='3' && buf[1]!='6')) {
    fclose(fptr);
    return 2; // File is not a PPM file.
  }
  imagestring.append(buf,bytes_read);

  do {
    bytes_read = fread(buf,sizeof(char),sizeof(buf),fptr);
    if(bytes_read>0) {
      imagestring.append(buf,bytes_read);
    }
  } while(bytes_read==sizeof(buf));

  int errcode=0;
  if(!feof(fptr)) errcode=3; // Read error

  fclose(fptr);

  return errcode; // Read error
}

BOOL
Oxs_ImageAtlas::GetNextPPMInteger
(const char* start,
 UINT4m& value,
 const char* &nextptr,
 BOOL binary)
{ // Reads an integer from a string representing a PPM P3 or P6 file.
  // In the case of P6 files, after reading the ascii header, the
  // client should skip past 1 character of whitespace following 
  // maxvalue, and read the rest of the file with binary==1.
  // NOTE: In binary mode, there is no end-of-string test.

  if(binary) {
    unsigned char uch = static_cast<unsigned char>(*start);
    value = static_cast<UINT4m>(uch); // Squelch sign extensions
    nextptr = start + 1;
    return 1; // Success
  }
  // Otherwise, proceed in ASCII mode

  // Skip leading whitespace
  const char* cptr = start + strspn(start," \t\r\n\f");

  // Skip comments
  while(*cptr=='#') {
    const char* cptr2 = strchr(cptr+1,'\n');
    if(cptr2==NULL) return 0; // Comment extends to end of string
    ++cptr2;
    cptr = cptr2 + strspn(cptr2," \t\r\n\f"); // Skip more whitespace
  }

  if(*cptr=='\0') return 0; // End of string

  char* endptr;
  unsigned long int test = strtoul(cptr,&endptr,10);
  if(endptr==cptr) return 0; // Conversion failure

  value=test;
  nextptr=endptr;
  return 1;
}

BOOL
Oxs_ImageAtlas::FindBestMatchRegion
(REAL8m r, REAL8m g, REAL8m b,
 const vector<OxsImageAtlasColorData>& color_data,
 UINT4m& match_id, REAL8m& match_distsq)
{ // Returns 0 if color_data is empty.
  // Otherwise sets match_id and match_distsq to best
  // fit, and returns 1.
  vector<OxsImageAtlasColorData>::const_iterator it
    = color_data.begin();
  if(it==color_data.end()) {
    return 0; // Error
  }
  match_id = it->region_id;
  match_distsq = it->DistanceSq(r,g,b);
  while(++it!=color_data.end()) {
    REAL8m distsq = it->DistanceSq(r,g,b);
    if(distsq < match_distsq) {
      match_distsq = distsq;
      match_id = it->region_id;
    }
  }
  return 1;
}

void
Oxs_ImageAtlas::FillPixels
(const char* filename,
 const vector<OxsImageAtlasColorData>& color_data,
 REAL8m match_error,
 UINT4m default_id)
{
  string msg; // Error message string.  Define it here so we can
  /// use it inside case blocks of switch statement. Otherwise
  /// VC++ 6.0 complains about "skipped" initializations.
  string imagestring; // Raw image buffer space

  const char* cptr = NULL;
  BOOL binaryfile = 0;

  // Read file into imagestring
  switch(ReadImage(filename,imagestring)) {
  case 0: // Success
    cptr = imagestring.c_str();
    binaryfile = (cptr[1]=='6');
    cptr += 2;  // Skip past P3/P6
    break;
  case 1: // Unable to open file
    msg = string("Unable to open input image file ")
      + string(filename) + string(" for reading.");
#if (TCL_MAJOR_VERSION > 8) \
   || (TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION > 0)
    // Tcl_GetCwd first appeared in Tcl 8.1.0
    { // Group this block to keep VC++ 6.0 quiet about dbuf
      // initialization.
      Tcl_DString dbuf;
      if(Tcl_GetCwd(NULL,&dbuf)!=NULL) {
	msg += string("  Current working directory: ")
	  + string(Tcl_DStringValue(&dbuf));
      }
      Tcl_DStringFree(&dbuf);
    }
#endif // End Tcl version check
    throw Oxs_Ext::Error(this,msg.c_str());
  case 2: // Not a PPM file.  Call ReadImageFile(), which
    /// is essentially a wrapper around any2ppm.
    if(director->ReadImageFile(filename,imagestring)!=0) {
      msg = string("Error reading image file ")
	+ string(filename) + string(" : ") + imagestring;
      throw Oxs_Ext::Error(this,msg.c_str());
    }
    cptr = imagestring.c_str(); // NOTE: No "P3" header
    binaryfile = 0;             // "P3" is a text format
    break;
  default: // Read error
    msg = string("Error reading input image file ")
      + string(filename);
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  // Process image header
  UINT4m imaxvalue;
  if(!GetNextPPMInteger(cptr,image_width,cptr)  ||
     !GetNextPPMInteger(cptr,image_height,cptr) ||
     !GetNextPPMInteger(cptr,imaxvalue,cptr)       ) {
    msg = string("Input image file ")
      + string(filename)
      + string(" has an invalid header.");
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  REAL8m maxvalue = static_cast<REAL8m>(imaxvalue);
  if(isspace(*cptr)) cptr++; // Skip 1 trailing whitespace
  /// character.  This is necessary for binary (P6) files,
  /// and doesn't hurt for ASCII (P3) files.

  // Read and categorize pixels
  UINT4m pixelcount = image_width*image_height;
  if(pixelcount<1) { // This check insures image width & height > 0.
    msg = string("Input image file ")
      + string(filename) + string(" is an empty image");
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  if(pixel!=NULL) delete[] pixel;
  pixel = new OxsImageAtlasPixelType[pixelcount];
  if(pixel==NULL) {
    msg = string("Insufficient memory to load input image file ")
      + string(filename);
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  for(UINT4m index=0;index<pixelcount;index++) {
    UINT4m r,g,b;
    if(!GetNextPPMInteger(cptr,r,cptr,binaryfile) ||
       !GetNextPPMInteger(cptr,g,cptr,binaryfile) ||
       !GetNextPPMInteger(cptr,b,cptr,binaryfile)   ) {
      delete[] pixel;
      pixel=NULL;
      char buf[1024];
      Oc_Snprintf(buf,sizeof(buf),
		  "Error parsing data area of input image file %s"
		  " at pixel %u (row=%u/%u, column=%u/%u)",
		  filename,index,
		  1+index/image_width,image_height,
		  1+index%image_width,image_width);
      throw Oxs_Ext::Error(this,buf);
    }
    REAL8m red   = static_cast<REAL8m>(r)/maxvalue;
    REAL8m green = static_cast<REAL8m>(g)/maxvalue;
    REAL8m blue  = static_cast<REAL8m>(b)/maxvalue;
    // Image is read in from upper lefthand corner, in normal English
    // reading order.  Pixel coords are based at bottom lefthand corner.
    // The "offset" calculation here accounts for these differences.
    UINT4m offset = pixelcount-image_width-index+2*(index%image_width);
    UINT4m id;
    REAL8m distsq;
    if(!FindBestMatchRegion(red,green,blue,color_data,id,distsq)
       || distsq>match_error*match_error) {
      id = default_id;
    }
    pixel[offset] = static_cast<OxsImageAtlasPixelType>(id);
  }
}

void
Oxs_ImageAtlas::InitializeRegionBBoxes(const Oxs_Box& world)
{ // Uses world, view and pixel data to setup bounding boxes
  // for each region.  NB: view, pixel, image_width,
  // image_height, and row/column_scale/offset variables
  // _must_ be set up *before* calling this function!

  region_bbox.clear(); // Safety
  region_bbox.push_back(world); // Set "universe" region to world

  // Initialize all non-universe region bboxes to empty boxes
  UINT4m id;
  Oxs_Box empty_box;
  for(id=1;id<region_name.size();id++) {
    region_bbox.push_back(empty_box);
  }

  // For each point in pixel array, expand corresponding region
  // the point in pixel coords.
  UINT4m index=0;
  for(UINT4m j=0;j<image_height;j++) {
    for(UINT4m i=0;i<image_width;i++) {
      UINT4m tempid = static_cast<UINT4m>(pixel[index]);
      if(tempid>0) { // Skip points in "universe" region.
	region_bbox[tempid].Expand(static_cast<REAL8m>(i),
				   static_cast<REAL8m>(j),
				   0.0);
	// Fix the out-of-plane component to 0 here; it
	// will be expanded to full world size when converting
	// to atlas coordinates below.
      }
      ++index;
    }
  }

  // Now convert from pixel coords to atlas coords.
  REAL8m cmult = 1.0/column_scale;
  REAL8m rmult = 1.0/row_scale;
  for(id=1;id<region_bbox.size();id++) {
    REAL8m cmin = region_bbox[id].GetMinX();
    REAL8m cmax = region_bbox[id].GetMaxX()+1.0; // Pad to full pixel
    REAL8m rmin = region_bbox[id].GetMinY();
    REAL8m rmax = region_bbox[id].GetMaxY()+1.0; // Pad to full pixel
    cmin = (cmin*cmult) + column_offset;
    cmax = (cmax*cmult) + column_offset;
    rmin = (rmin*rmult) + row_offset;
    rmax = (rmax*rmult) + row_offset;
    switch(view) {
    case xy:
      region_bbox[id].Set(cmin,cmax,
			  rmin,rmax,
			  world.GetMinZ(),world.GetMaxZ());
      break;
    case zx:
      region_bbox[id].Set(rmin,rmax,
			  world.GetMinY(),world.GetMaxY(),
			  cmin,cmax);
      break;
    case yz:
      region_bbox[id].Set(world.GetMinX(),world.GetMaxX(),
			  cmin,cmax,
			  rmin,rmax);

      break;
    }
  }
}

// Constructor
Oxs_ImageAtlas::Oxs_ImageAtlas
( const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr    // MIF block argument string
  ) : Oxs_Atlas(name,newdtr,argstr),
      pixel(NULL),image_width(0),image_height(0),view(xy),
      row_offset(0.),row_scale(1.),
      column_offset(0.),column_scale(1.)
{
  // Process input string
  vector<REAL8m> xrange; GetRealVectorInitValue("xrange",2,xrange);
  vector<REAL8m> yrange; GetRealVectorInitValue("yrange",2,yrange);
  vector<REAL8m> zrange; GetRealVectorInitValue("zrange",2,zrange);
  Oxs_Box world_box;
  if(!world_box.CheckOrderSet(xrange[0],xrange[1],
			      yrange[0],yrange[1],
			      zrange[0],zrange[1])) {
    throw Oxs_Ext::Error(this,"One of [xyz]range in Specify block"
			 " is improperly ordered.");
  }
  if(world_box.GetVolume()<=0.0) {
    throw Oxs_Ext::Error(this,"Atlas bounding box has zero volume.");
  }

  string viewstr = GetStringInitValue("viewplane");
  if(viewstr.compare("xy")==0)      view = xy;
  else if(viewstr.compare("zx")==0) view = zx;
  else if(viewstr.compare("yz")==0) view = yz;
  else {
    string msg = "Invalid viewplane init string: "
      + viewstr + "; should be one of xy, zx or yz.";
    throw Oxs_Ext::Error(this,msg.c_str());
  }

  REAL8m match_error = GetRealInitValue("matcherror",3.);
  // Default is no limit

  // Parse colormap info from Specify block
  vector<string> colormap;
  FindRequiredInitValue("colormap",colormap);
  if(colormap.size()%2!=0) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"Colormap entry in Specify block has"
		" %d entries, which is not an even number",
		colormap.size());
    throw Oxs_Ext::Error(this,buf);
  }
  vector<OxsImageAtlasColorData> color_data;
  UINT4m default_id=0;
  region_name.push_back("universe"); // Required special region, id 0.
  for(UINT4m i=0;i<colormap.size();i+=2) {
    // Is label already defined?
    const string& label = colormap[i+1];
    UINT4m id;
    for(id=0;id<region_name.size();id++) {
      if(region_name[id].compare(label)==0) break;
    }
    if(id>=region_name.size()) {
      region_name.push_back(label); // New label
    }
    // id is the index of the label specified by colormap[i+1]
    if(Nb_StrCaseCmp(colormap[i].c_str(),"default")==0) {
      default_id = id;
    } else {
      OxsImageAtlasColorData tempcolor;
      tempcolor.region_id=id;
      if(!tempcolor.SetColor(colormap[i])) {
	char buf[1024];
	Oc_Snprintf(buf,sizeof(buf),
		    "Unrecognized color in colormap"
		    " in Specify block: %s",
		    colormap[i].c_str());
	throw Oxs_Ext::Error(this,buf);
      }
      color_data.push_back(tempcolor); // Add color mapping to color_data
    }
  }
  if(region_name.size() >= (1<<(8*sizeof(OxsImageAtlasPixelType))) ) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"Too many regions: %u (max allowed is %u)",
		region_name.size(),
		1<<(8*sizeof(OxsImageAtlasPixelType)));
    throw Oxs_Ext::Error(this,buf);
  }
  DeleteInitValue("colormap");

  string image_file = GetStringInitValue("image");

  VerifyAllInitArgsUsed();

  // Initialize pixel map
  FillPixels(image_file.c_str(),color_data,match_error,default_id);

  // Initialize transformation coefficients
  switch(view) {
  case xy:
    column_offset = world_box.GetMinX();
    column_scale  =  static_cast<REAL8m>(image_width)
      /(world_box.GetMaxX()-world_box.GetMinX());
    row_offset = world_box.GetMinY();
    row_scale  = static_cast<REAL8m>(image_height)
      /(world_box.GetMaxY()-world_box.GetMinY());
    break;
  case zx:
    column_offset = world_box.GetMinZ();
    column_scale  = static_cast<REAL8m>(image_width)
      /(world_box.GetMaxZ()-world_box.GetMinZ());
    row_offset = world_box.GetMinX();
    row_scale  = static_cast<REAL8m>(image_height)
      /(world_box.GetMaxX()-world_box.GetMinX());
    break;
  case yz:
    column_offset = world_box.GetMinY();
    column_scale  = static_cast<REAL8m>(image_width)
      /(world_box.GetMaxY()-world_box.GetMinY());
    row_offset = world_box.GetMinZ();
    row_scale  = static_cast<REAL8m>(image_height)
      /(world_box.GetMaxZ()-world_box.GetMinZ());
    break;
  }

  // Iterate through points to find bboxes for each region
  InitializeRegionBBoxes(world_box);
}


// Destructor
Oxs_ImageAtlas::~Oxs_ImageAtlas()
{
  if(pixel!=NULL) delete[] pixel;
}

BOOL Oxs_ImageAtlas::GetRegionExtents(UINT4m id,Oxs_Box &mybox) const
{ // If 0<id<GetRegionCount, then sets mybox to bounding box
  // for the specified region, and returns 1.  If id is 0,
  // sets mybox to atlas bounding box, and returns 1.
  // Otherwise, leaves mybox untouched and returns 0.
  if(id>=GetRegionCount()) return 0;
  mybox = region_bbox[id]; // Atlas bbox is stored in region_bbox[0].
  return 1;
}

INT4m Oxs_ImageAtlas::GetRegionId(const ThreeVector& point) const
{ // Returns the id number for the region containing point.
  // The return value is 0 if the point is not contained in
  // the atlas, i.e., belongs to the "universe" region.

  if(!region_bbox[0].IsIn(point)) return 0; // Outside atlas

  REAL8m mapx=0,mapy=0;
  switch(view) {
  case xy: mapx = point.x; mapy = point.y; break;
  case zx: mapx = point.z; mapy = point.x; break;
  case yz: mapx = point.y; mapy = point.z; break;
  }
  
  REAL8m tempc = floor((mapx-column_offset)*column_scale);
  UINT4m column_index = 0;
  if(tempc>=image_width) column_index = image_width-1;
  else if(tempc>0) column_index = static_cast<UINT4m>(tempc);

  REAL8m tempr = floor((mapy-row_offset)*row_scale);
  UINT4m row_index = 0;
  if(tempr>=image_height) row_index = image_height-1;
  else if(tempr>0) row_index = static_cast<UINT4m>(tempr);

  UINT4m index = row_index*image_width+column_index;

  UINT4m id = pixel[index];
  if(id>=GetRegionCount()) {
    char buf[1024];
    Oc_Snprintf(buf,sizeof(buf),
		"Programming error detected in "
		"Oxs_ImageAtlas::GetRegionId(const ThreeVector&) const."
		"  Region id %u discovered in pixel array;"
		" max value should be %u",id,GetRegionCount()-1);
    throw Oxs_Ext::Error(this,buf);
  }

  return static_cast<INT4m>(id);
}

INT4m Oxs_ImageAtlas::GetRegionId(const string& name) const
{ // Given a region id string (name), returns
  // the corresponding region id index.  If
  // "name" is not included in the atlas, then
  // -1 is returned.  Note: If name == "universe",
  // then the return value will be 0.
  UINT4m count = GetRegionCount();
  for(UINT4m i=0;i<count;i++) {
    if(region_name[i].compare(name)==0) return i;
  }
  return -1;
}

BOOL Oxs_ImageAtlas::GetRegionName(UINT4m id,string& name) const
{ // Given an id number, fills in "name" with
  // the corresponding region id string.  Returns
  // 1 on success, 0 if id is invalid.  If id is 0,
  // then name is set to "universe", and the return
  // value is 1.
  name.erase();
  if(id>=GetRegionCount()) return 0;
  name = region_name[id];
  return 1;
}
