/* FILE: planerandomvectorfield.cc      -*-Mode: c++-*-
 *
 * Vector field object, random in a specified plane.
 * Derived from Oxs_VectorField class.
 *
 */

#include "oc.h"
#include "nb.h"

#include "threevector.h"
#include "meshvalue.h"
#include "director.h"
#include "planerandomvectorfield.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_PlaneRandomVectorField);

/* End includes */


// Constructor
Oxs_PlaneRandomVectorField::Oxs_PlaneRandomVectorField(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_VectorField(name,newdtr,argstr)
{
  // Process arguments.
  min_norm=GetRealInitValue("min_norm");
  max_norm=GetRealInitValue("max_norm");
  norm_range=max_norm-min_norm;

  plane_normal=GetThreeVectorInitValue("plane_normal");
  plane_normal.SetMag(1.0);

  VerifyAllInitArgsUsed();
}

BOOL Oxs_PlaneRandomVectorField::Init()
{
  return 1;
}

void
Oxs_PlaneRandomVectorField::Value
(const ThreeVector& /* pt */,
 ThreeVector& value) const
{
  // Produce vector perpendicular to plane
  int j=0;
  REAL8m dot=0;
  do {
    value.Random(1.0);
    dot = plane_normal*value;
  } while(fabs(dot)>0.96875 && ++j<32);
  if(fabs(dot)>0.96875) {
    string msg=
      string("Programming error?  Unable to obtain vector"
	     " in specified plane in"
	     " Oxs_PlaneRandomVectorField::Value()"
	     " method of instance ")
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }
  value -= dot*plane_normal;

  // Adjust to proper size
  REAL8m mag = min_norm;
  if(norm_range!=0.0) {
    // Vary magnitude
    mag += norm_range*Oc_UnifRand();
  }
  value.SetMag(mag);
}

void
Oxs_PlaneRandomVectorField::FillMeshValueOrthogonal
(const Oxs_Mesh* mesh,
 const Oxs_MeshValue<ThreeVector>& orth_array,
 Oxs_MeshValue<ThreeVector>& array) const
{
  if(!orth_array.CheckMesh(mesh)) {
    string msg=
      string("Mesh size incompatibility detected in"
	     " Oxs_PlaneRandomVectorField::FillMeshValueOrthogonal()"
	     " method of instance ")
      + string(InstanceName());
    throw Oxs_Ext::Error(msg.c_str());
  }

  REAL8m dot=0.0;
  ThreeVector vec;
  UINT4m i,size=mesh->Size();
  array.AdjustSize(mesh);
  for(i=0;i<size;i++) {
    // We need a vector that is orthogonal to both plane and
    // orth_array[i].  First try 'plane x orth_array[i]'
    vec = orth_array[i];
    vec.SetMag(1.0);
    vec ^= plane_normal;
    if(vec.MagSq()>0.0) {
      // Randomly select +/-
      if(Oc_UnifRand()<0.5) vec *= -1;
    } else {
      // orth_array[i] is either 0 or parallel to plane,
      // so just produce a vector orthogonal to plane from
      // scratch.
      int j=0;
      while(1) {
	vec.Random(1.0);
	dot = plane_normal*vec;
	if(fabs(dot)<=0.96875) break;
	if(++j>32) {
	  string msg=
	    string("Programming error?  Unable to obtain vector"
		   " in specified plane in"
		   " Oxs_PlaneRandomVectorField::FillMeshValueOrthogonal()"
		   " method of instance ")
	    + string(InstanceName());
	  throw Oxs_Ext::Error(msg.c_str());
	}
      }
      vec -= dot*plane_normal;
    }

    // Adjust to proper size
    REAL8m mag = min_norm;
    if(norm_range!=0.0) {
      // Vary magnitude
      mag += norm_range*Oc_UnifRand();
    }
    vec.SetMag(mag);
    array[i] = vec;
  }
}
