/* FILE: timedriver.cc            -*-Mode: c++-*-
 *
 * Example concrete Oxs_Driver class.
 *
 */

#include "nb.h"
#include "timedriver.h"
#include "director.h"
#include "simstate.h"
#include "timeevolver.h"
#include "key.h"
#include "mesh.h"
#include "scalarfield.h"
#include "vectorfield.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_TimeDriver);

/* End includes */

// Constructor
Oxs_TimeDriver::Oxs_TimeDriver(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_Driver(name,newdtr,argstr), max_dm_dt_obj_ptr(NULL)
{
  // Process arguments
  OXS_GET_INIT_EXT_OBJECT("evolver",Oxs_TimeEvolver,evolver_obj);
  evolver_key.Set(evolver_obj.GetPtr());
  // Dependency lock on Oxs_TimeEvolver object is
  // held until *this is destroyed.

  if(!HasInitValue("stopping_dm_dt")) {
    stopping_dm_dt.push_back(0.0); // Default is no control
  } else {
    GetGroupedRealListInitValue("stopping_dm_dt",stopping_dm_dt);
  }

  if(!HasInitValue("stopping_time")) {
    stopping_time.push_back(0.0); // Default is no control
  } else {
    GetGroupedRealListInitValue("stopping_time",stopping_time);
  }

  OXS_GET_INIT_EXT_OBJECT("mesh",Oxs_Mesh,mesh_obj);
  mesh_key.Set(mesh_obj.GetPtr());  // Sets a dep lock

  Oxs_OwnedPointer<Oxs_ScalarField> Msinit;
  OXS_GET_INIT_EXT_OBJECT("Ms",Oxs_ScalarField,Msinit);

  OXS_GET_INIT_EXT_OBJECT("m0",Oxs_VectorField,m0);

  VerifyAllInitArgsUsed();

  // Fill Ms and Ms_inverse array, and verify that Ms is non-negative.
  Msinit->FillMeshValue(mesh_obj.GetPtr(),Ms);
  Ms_inverse.AdjustSize(mesh_obj.GetPtr());
  for(UINT4m i=0;i<mesh_obj->Size();i++) {
    if(Ms[i]<0.0) {
      char buf[1024];
      Oc_Snprintf(buf,sizeof(buf),
		  "Negative Ms value (%g) detected at mesh index %d.",
		  Ms[i],i);
      throw Oxs_Ext::Error(this,buf);
    } else if(Ms[i]==0.0) {
      Ms_inverse[i]=0.0; // Special case handling
    } else {
      Ms_inverse[i]=1.0/Ms[i];
    }
  }

  // Reserve space for initial state (see GetInitialState() below)
  director->ReserveSimulationStateRequest(1);
}

Oxs_ConstKey<Oxs_SimState>
Oxs_TimeDriver::GetInitialState() const
{
  Oxs_Key<Oxs_SimState> initial_state;
  director->GetNewSimulationState(initial_state);
  Oxs_SimState& istate = initial_state.GetWriteReference();
  istate.previous_state_id     = 0;
  istate.iteration_count       = 0;
  istate.stage_number          = 0;
  istate.stage_iteration_count = 0;
  istate.stage_start_time      = 0.0;
  istate.stage_elapsed_time    = 0.0;
  istate.last_timestep         = 0.0;

  istate.mesh = mesh_key.GetPtr();

  istate.Ms = &Ms;
  istate.Ms_inverse = &Ms_inverse;
  m0->FillMeshValue(istate.mesh,istate.spin);

  // Insure that all spins are unit vectors
  UINT4m size = istate.spin.Size();
  for(UINT4m i=0;i<size;i++) istate.spin[i].MakeUnit();

  initial_state.GetReadReference();  // Release write lock.
  /// The read lock will be automatically released when the
  /// key "initial_state" is destroyed.

  return initial_state;
}


BOOL Oxs_TimeDriver::Init()
{ 
  last_timestep_output.Setup(
           this,InstanceName(),"Last time step","s",0,
	   &Oxs_TimeDriver::Fill__last_timestep_output,
	   director);
  simulation_time_output.Setup(
	   this,InstanceName(),"Simulation time","s",0,
	   &Oxs_TimeDriver::Fill__simulation_time_output,
	   director);

  Oxs_Driver::Init();  // Run init routine in parent.
  /// This will call Oxs_TimeDriver::GetInitialState().

  // Get pointer to output object providing max dm/dt data
  const Oxs_TimeEvolver* evolver = evolver_key.GetPtr();
  if(evolver==NULL) {
    throw Oxs_Ext::Error(this,"PROGRAMMING ERROR: No evolver found?");
  }
  string output_name = string(evolver->InstanceName());
  output_name += string(":Max dm/dt");
  max_dm_dt_obj_ptr =  director->FindOutputObject(output_name.c_str());
  if(max_dm_dt_obj_ptr==NULL) {
    throw Oxs_Ext::Error(this,"Unable to identify unique"
                         " Max dm/dt output object");
  }

  return 1;
}

Oxs_TimeDriver::~Oxs_TimeDriver()
{}

void Oxs_TimeDriver::StageRequestCount
(unsigned int& min,
 unsigned int& max) const
{ // Number of stages wanted by driver

  Oxs_Driver::StageRequestCount(min,max);

  unsigned int count = stopping_dm_dt.size();
  if(count>min) min=count;
  if(count>1 && count<max) max=count;
  // Treat length 1 lists as imposing no upper constraint.

  count =  stopping_time.size();
  if(count>min) min=count;
  if(count>1 && count<max) max=count;
  // Treat length 1 lists as imposing no upper constraint.
}

BOOL
Oxs_TimeDriver::ChildIsStageDone(const Oxs_SimState& state) const
{
  UINT4m stage_index = state.stage_number;

  // Stage time check
  REAL8m stop_time=0.;
  if(stage_index >= stopping_time.size()) {
    stop_time = stopping_time[stopping_time.size()-1];
  } else {
    stop_time = stopping_time[stage_index];
  }
  if(stop_time>0.0 && state.stage_elapsed_time >= stop_time) {
    return 1; // Stage done
  }

  // dm_dt check
  Tcl_Interp* mif_interp = director->GetMifInterp();
  if(max_dm_dt_obj_ptr==NULL ||
     max_dm_dt_obj_ptr->Output(&state,mif_interp,0,NULL) != TCL_OK) {
    string msg=string("Unable to obtain Max dm/dt output: ");
    if(max_dm_dt_obj_ptr==NULL) {
      msg += string("PROGRAMMING ERROR: max_dm_dt_obj_ptr not set."
		    " Driver Init() probably not called.");
    } else {
      msg += string(Tcl_GetStringResult(mif_interp));
    }
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  BOOL err;
  REAL8m max_dm_dt = Nb_Atof(Tcl_GetStringResult(mif_interp),err);
  if(err) {
    string msg=string("Error detected in StageDone method"
		      " --- Invalid Max dm/dt output: ");
    msg += string(Tcl_GetStringResult(mif_interp));
    throw Oxs_Ext::Error(this,msg.c_str());
  }
  REAL8m stop_dm_dt=0.;
  if(stage_index >= stopping_dm_dt.size()) {
    stop_dm_dt = stopping_dm_dt[stopping_dm_dt.size()-1];
  } else {
    stop_dm_dt = stopping_dm_dt[stage_index];
  }
  if(stop_dm_dt>0.0 && max_dm_dt <= stop_dm_dt) {
    return 1; // Stage done
  }

  // If control gets here, then stage not done
  return 0;
}

BOOL
Oxs_TimeDriver::ChildIsRunDone(const Oxs_SimState& /* state */) const
{
  // No child-specific checks at this time...
  return 0; // Run not done
}

void Oxs_TimeDriver::FillStateSupplemental(Oxs_SimState& work_state) const
{
  REAL8m work_step = work_state.last_timestep;
  REAL8m base_time = work_state.stage_elapsed_time - work_step;

  // Insure that step does not go past stage stopping time
  UINT4m stop_index = work_state.stage_number;
  REAL8m stop_value=0.0;
  if(stop_index >= stopping_time.size()) {
    stop_value = stopping_time[stopping_time.size()-1];
  } else {
    stop_value = stopping_time[stop_index];
  }
  if(stop_value>0.0
     && base_time + work_step > stop_value) {
    work_step = stop_value - base_time;
  }

  // Reflect changes into work_state, as needed.
  if(work_step != work_state.last_timestep) {
    work_state.last_timestep = work_step;
    work_state.stage_elapsed_time = base_time + work_step;
  }

}

BOOL
Oxs_TimeDriver::Step
(Oxs_ConstKey<Oxs_SimState> base_state,
 Oxs_Key<Oxs_SimState>& next_state)
{ // Returns true if step was successful, false if
  // unable to step as requested.

  // Put write lock on evolver in order to get a non-const
  // pointer.  Use a temporary variable, temp_key, so
  // write lock is automatically removed when temp_key
  // is destroyed.
  Oxs_Key<Oxs_TimeEvolver> temp_key = evolver_key;
  Oxs_TimeEvolver& evolver = temp_key.GetWriteReference();

  return evolver.Step(this,base_state,next_state);
}

////////////////////////////////////////////////////////////////////////
// State-based outputs, maintained by the driver.  These are
// conceptually public, but are specified private to force
// clients to use the output_map interface in Oxs_Director.

#define OSO_FUNC(NAME) \
void Oxs_TimeDriver::Fill__##NAME##_output(const Oxs_SimState& state) \
{ NAME##_output.cache.state_id=state.Id(); \
  NAME##_output.cache.value=state.NAME; }

OSO_FUNC(last_timestep)

void
Oxs_TimeDriver::Fill__simulation_time_output(const Oxs_SimState& state)
{
  simulation_time_output.cache.state_id = state.Id();
  simulation_time_output.cache.value =
    state.stage_start_time + state.stage_elapsed_time;
}
