/* FILE: functions.cc                   -*-Mode: c++-*-
 *
 * Non-class C++ functions defined by the Nb extension.
 *
 * NOTICE: Please see the file ../../LICENSE
 *
 * Last modified on: $Date: 2002/08/05 19:06:43 $
 * Last modified by: $Author: dgp $
 */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "functions.h"
#include "errhandlers.h"

/* End includes */     

//////////////////////////////////////////////////////////////////////////
// Greatest common divisor via Euclid's algorithm
int Nb_Gcd(int m,int n)
{
  m=abs(m); n=abs(n);
  if(n==0) return 0;
  int temp;
  while((temp=(m%n))>0) { m=n; n=temp; }
  return n;
}

// Tcl wrapper for the above
int NbGcdCmd(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  char buf[1024];
  Tcl_ResetResult(interp);
  if(argc!=3) {
    Oc_Snprintf(buf,sizeof(buf),"Nb_Gcd must be called with 2 arguments,"
	    " m and n (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  int n=atoi(argv[1]);
  int m=atoi(argv[2]);

  if(m<1 || n<1) {
    Oc_Snprintf(buf,sizeof(buf),"Nb_Gcd input must be two positive integers"
	    " (got %d, %d)", n,m);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Oc_Snprintf(buf,sizeof(buf),"%d",Nb_Gcd(m,n));
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

//////////////////////////////////////////////////////////////////////////
// Simple continued fraction like rational approximator.  Unlike the usual
// continued fraction expansion, here we allow +/- on each term (rounds to
// nearest integer rather than truncating).  This converges faster (by
// producing fewer terms).
//   A more useful rat approx routine would probably take an error
// tolerance rather than a step count, but we'll leave that until the time
// when we have an actual use for this routine.  (There is a 4 term
// recursion relation for continued fractions in "Numerical Recipes in C"
// that can probably be used, and also remove the need for the coef[]
// array.)  -mjd, 23-Jan-2000.

void Nb_RatApprox(double x,int steps,int &num,int &denom)
{
  if(steps<1) steps=1;
  int *coef=new int[steps];
  coef[0]=int(rint(x));
  double remainder=fabs(x)-fabs(double(coef[0]));
  // Expand
  int i=0;
  while( ++i < steps ) {
    if(fabs(remainder)<=1.0/double(INT_MAX)) break;
    remainder=1.0/remainder;
    coef[i]=int(rint(remainder));
    remainder=fabs(remainder)-fabs(double(coef[i]));
  }
  // Collect
  int a=1,b=coef[--i];
  while( --i >= 0 ) {
    int sign=1;
    int c=coef[i];
    if(c<0) { sign=-1; c*=-1; }
    int temp=b;
    b=b*c+a;
    a=sign*temp;
  }
  if(a<0) { a*=-1; b*=-1; }
  int div=1;
  if(a!=0 && b!=0) {
    div=Nb_Gcd(a,abs(b));
  }
  num=b/div;  denom=a/div;
  delete[] coef;
}

// Tcl wrapper for the above
int NbRatApprox(ClientData,Tcl_Interp *interp,int argc,CONST84 char** argv)
{
  char buf[1024];
  Tcl_ResetResult(interp);
  if(argc!=3) {
    Oc_Snprintf(buf,sizeof(buf),"Nb_RatApprox must be called with 2 arguments,"
	    " x and steps (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }
  double x=Nb_Atof(argv[1]);
  int steps=atoi(argv[2]);

  int num,denom;
  Nb_RatApprox(x,steps,num,denom);

  Oc_Snprintf(buf,sizeof(buf),"%d %d",num,denom);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}

//////////////////////////////////////////////////////////////////////////
// Version of atof that allows 'D' or 'd' to be used to denote exponent
double Nb_Atof(const char *nptr)
{ // NB: Mimicking the system atof(), this routine will seg fault
  // of nptr==NULL.  It is the resposibility of the calling routine
  // to insure this doesn't occur.  OTOH, the Nb_Atof(const char*,BOOL&)
  // below checks and catches this error.
  const char *cptr;
  if((cptr=strchr(nptr,'d'))==NULL &&
      (cptr=strchr(nptr,'D'))==NULL  )
    return atof(nptr);  // Usual, C-compatible string
  /// The above checks seem to add about 10% overhead for short
  /// strings on a 486/66dx2, and it is faster there to check for
  /// each of 'd' and 'D' separately than to call strcspn(nptr,"dD");
  /// The base atof time on this machine is about 85 microseconds
  /// to start with (compared to about 6 microseconds for atoi).

  // Otherwise, assume we are working with a FORTRAN-style
  // "double precision" string.
  //   On the aforementioned 486, with gcc 2.7.2, this code costs
  // on average about an additional 10% over the above, if statbuf
  // is used.  Calls to new[] and delete[] add another 20%.  NOTE:
  // the strncpy function is really slow on this machine (it takes
  // as long as the atof call itself), so instead of using strncpy
  // to ensure we don't write past the end of statbuf, we take a
  // more flexible approach: see how big the string is, and if it
  // is too big (which is unlikely, but possible), then allocate
  // temporary storage just large enough to hold the string.  The
  // latter is a performance hit, so if it happens frequently one
  // should enlarge the size of statbuf.
  size_t exp_off=cptr-nptr;
  static char statbuf[256]; // I don't expect many numeric strings this long
  char *buf=statbuf;
  size_t num_len=strlen(nptr)+1;
  if(num_len>sizeof(statbuf)) buf=new char[num_len];

  strcpy(buf,nptr);
  buf[exp_off]='e';         // Replace 'd' or 'D' with 'e'
  double result=atof(buf);

  if(buf!=statbuf) delete[] buf;
  return result;
}

// Version of Nb_Atof with error checking.  Export value 'error'
// is set true if an error occurred, false otherwise.
double Nb_Atof(const char *nptr,BOOL& error)
{
  if(nptr==NULL) { error=1; return 0.0; } // Null pointer check.
  /// Note that the Nb_Atof(const char*) function above fails
  /// if nptr==NULL.

  // Replace first 'd' or 'D', if any, with 'e'
  static char statbuf[256]; // I don't expect many num. strs this long
  char *buf=statbuf;
  const char *cptr;
  if((cptr=strchr(nptr,'d'))!=NULL ||
     (cptr=strchr(nptr,'D'))!=NULL  ) {
    size_t exp_off=cptr-nptr;
    size_t num_len=strlen(nptr)+1;
    if(num_len>sizeof(statbuf)) buf=new char[num_len];
    strcpy(buf,nptr);
    buf[exp_off]='e';         // Replace 'd' or 'D' with 'e'
    cptr=buf;
  } else {
    cptr=nptr;
  }

  // Convert to double
  char* endptr;
  double result=strtod(cptr,&endptr);

  // Check for errors
  if(cptr[0]=='\0' || *endptr!='\0') {
    error=1; // Error!
  } else {
    error=0; // Full conversion
  }

  // If extra buffer space was allocated above, then release it
  if(buf!=statbuf) delete[] buf;

  // All done!
  return result;
}

//////////////////////////////////////////////////////////////////////////
// Atan2 inverse, in degrees.  If angle is an integer, then we can do
// range reduction on it without loss of precision.  This routine could
// undoubtably be shortened, but the overriding concern here is accuracy.
// NOTE: The import "angle" is in DEGREES!
void degcossin(double angle,double& cosine,double& sine)
{
  int sign=1;
  if(angle<0) { sign=-1; angle*=-1; }
  int base=int(floor(angle));
  angle-=base;
  base%=360;
  int quadrant=1;
  if(base>=270)      quadrant=4;
  else if(base>=180) quadrant=3;
  else if(base>=90)  quadrant=2;
  angle+=(base%90);
  int flip=0;
  if(angle>45) { angle=90.-angle; flip=1; }
  if(quadrant==2 || quadrant==4) flip=(flip+1)%2;
  double x,y;
  if(angle==0.0) {
    y=0.0; x=1.0;
  } else {
    angle*=(PI/180.);
    x=cos(angle);
    y=sin(angle);
    // x=sqrt(1.0-y*y);
    // x += (1-(x*x+y*y))/(2*x);  // Tiny linear correction
  }
  if(flip) { double temp=x; x=y; y=temp; }
  switch(quadrant)
    {
    case 2: x*=-1; break;
    case 3: x*=-1; y*=-1; break;
    case 4: y*=-1; break;
    }
  cosine=x;
  sine=y*sign;
}

//////////////////////////////////////////////////////////////////////////
// Routines to detect IEEE floating point NAN's and infinities.
int Nb_IsFinite(REAL4 x)
{
  // Inf's and nan's are indicated by all exponent bits (8) being set. 
  // The top bit is the sign bit.
  unsigned char *cptr = (unsigned char *)(&x);
#if (BYTEORDER == 4321)  // Little endian
  unsigned int code = (((unsigned int)cptr[3])<<8) + ((unsigned int)cptr[2]);
#else // Otherwise assume big endian
  unsigned int code = (((unsigned int)cptr[0])<<8) + ((unsigned int)cptr[1]);
#endif
  unsigned mask = 0x7F80;
  code &= mask;
  return (code!=mask);
}

int Nb_IsFinite(REAL8 x)
{
  // Inf's and nan's are indicated by all exponent bits (11) being set.
  // The top bit is the sign bit.
  unsigned char *cptr = (unsigned char *)(&x);
#if (BYTEORDER == 4321)  // Little endian
  unsigned int code = (((unsigned int)cptr[7])<<8) + ((unsigned int)cptr[6]);
#else // Otherwise assume big endian
  unsigned int code = (((unsigned int)cptr[0])<<8) + ((unsigned int)cptr[1]);
#endif
  unsigned mask = 0x7FF0;
  code &= mask;
  return (code!=mask);
}

//////////////////////////////////////////////////////////////////////////
// Routine to detect string containing nothing but whitespace.
int Nb_StrIsSpace(const char* str)
{
  if(str==NULL) return 1;
  do {
    char ch = *str;
    if(ch=='\0') break;
    if(!isspace(ch)) return 0; // Non-space, not '\0' character
  } while(++str);
  return 1; // Each string element satisfies isspace()==TRUE
}

//////////////////////////////////////////////////////////////////////////
// Portable case insensitive string comparison, modeled after
// strcasecmp().
int Nb_StrCaseCmp(const char *s1, const char *s2)
{
  const char *cptr1=s1,*cptr2=s2;
  int c1,c2;
  do {
    c1=tolower(*(cptr1++));
    c2=tolower(*(cptr2++));
  } while(c1==c2 && c1!='\0');
  if(c1<c2) return -1;
  if(c1>c2) return  1;
  return 0;
}

//////////////////////////////////////////////////////////////////////////
// Portable thread-safe string tokenizer, modeled after strsep().
// Compare also to strtok().
char* Nb_StrSep(char **stringp, const char *delim)
{
  char* token_start = *stringp + strspn(*stringp,delim);
  if(*token_start == '\0') {
    *stringp = token_start;
    return NULL;
  }
  char* token_end = token_start + strcspn(token_start,delim);
  if(*token_end=='\0') {
    *stringp = token_end;
  } else {
    *token_end='\0';
    *stringp = token_end + 1;
  }
  return token_start;
}

//////////////////////////////////////////////////////////////////////////
// Map color strings to real [0,1] rgb values.
// Returns 1 on success, 0 on error.
BOOL Nb_GetColor(const char* color,
		 REAL8m& red,REAL8m& green,REAL8m& blue)
{
  if(color==NULL || color[0]=='\0') return 0;

  // Translate "color" to lowercase and remove whitespace
  size_t size = strlen(color);
  char* buf = new char[size+1]; // Certainly big enough
  if(buf==NULL) return 0; // Safety
  const char* cptr1=color;
  char* cptr2=buf;
  char ch='\0';
  do {
    ch = *cptr1;
    ++cptr1;
    if(!isspace(ch)) {
      *cptr2 = static_cast<char>(tolower(ch));
      ++cptr2;
    }
  } while(ch!='\0');
  size = cptr2 - buf - 1; // Resize

  BOOL success=0;

  // Check to see if color is in Tk numeric color format
  if(buf[0] == '#') {
    size_t match_len = strspn(buf+1,"0123456789abcdef");
    if(match_len+1 == size) {
      unsigned int r=0,g=0,b=0;
      switch(size) {
      case  4:
	if(sscanf(buf,"#%1x%1x%1x",&r,&g,&b)==3) {
	  r*=4096; g*=4096; b*=4096;
	  success=1;
	}
	break;
      case  7:
	if(sscanf(buf,"#%2x%2x%2x",&r,&g,&b)==3) {
	  r*=256; g*=256; b*=256;
	  success=1;
	}
	break;
      case 10:
	if(sscanf(buf,"#%3x%3x%3x",&r,&g,&b)==3) {
	  r*=16; g*=16; b*=16;
	  success=1;
	}
	break;
      case 13:
	if(sscanf(buf,"#%4x%4x%4x",&r,&g,&b)==3) {
	  success=1;
	}
	break;
      default:
	success=0; // Safety
	break;
      }
      if(success) {
	red   = static_cast<REAL8m>(r)/65535.;
	green = static_cast<REAL8m>(g)/65535.;
	blue  = static_cast<REAL8m>(b)/65535.;
      }
    }
  } else {
    // Check for gray# or grey# variants
    if(size>4 && size<8
       && buf[0]=='g' && buf[1]=='r' && (buf[2]=='a' || buf[2]=='e')
       && buf[3]=='y' && strspn(buf+4,"0123456789")==size-4) {
      int shade = atoi(buf+4);
      if(shade<=100) {
	red=green=blue=static_cast<REAL8m>(shade)/100.;
 	success=1;
      }
    }
  }

  Tcl_Interp* interp=NULL;
  if(!success && (interp=Oc_GlobalInterpreter())!=NULL) {
    // If we get to here, then input string was either not in
    // Tk numeric color format, or else an error occurred
    // reading the string.  Assume symbolic name, and do a
    // lookup in the color database
    Tcl_SavedResult saved;
    Tcl_SaveResult(interp, &saved);

    Oc_AutoBuf ab1,ab2,ab3;
    const char* cptr=NULL;

    // First check to see if nbcolordatabase array has been
    // initialized.
    if(Tcl_GetVar2(interp,ab1("nbcolordatabase"),
		   ab2("black"),TCL_GLOBAL_ONLY)==NULL) {
      // Array not yet filled.  Evaluate a Tcl script to
      // source the file(s) specified in the Oc_Option
      // database.

      if(Tcl_GlobalEval(interp,
			ab1("Oc_Option Get Color filename _fl ;"
			    " foreach _f $_fl { source $_f }"))
	 != TCL_OK) {
	PlainWarning("Error sourcing color database file(s): %s",
		     Tcl_GetStringResult(interp));
	goto finished;
      }

      // Add black entry just to be sure that we don't
      // initialize again next time through.
      Tcl_SetVar2(interp,ab1("nbcolordatabase"),
		  ab2("black"),ab3("0 0 0"),TCL_GLOBAL_ONLY);
    }

    // Try to read color from database
    cptr = Tcl_GetVar2(interp,ab1("nbcolordatabase"),
		       buf,TCL_GLOBAL_ONLY);
    if(cptr!=NULL) {
      // Success; Convert from hex triplet into floats
      unsigned int r,g,b;
      if(sscanf(cptr,"%X%X%X",&r,&g,&b)!=3) {
	PlainWarning("Invalid triplet in color database for color %s: %s",
		     buf,cptr);
      } else {
	red   = static_cast<REAL8m>(r)/255.;
	green = static_cast<REAL8m>(g)/255.;
	blue  = static_cast<REAL8m>(b)/255.;
 	success=1;
      }
    }

  finished:
    Tcl_RestoreResult(interp, &saved);
  }

  delete[] buf;
  return success;
}


//////////////////////////////////////////////////////////////////////////
// C interface to the Tcl 'file nativename' command.
void Nb_TclNativeFilename(Nb_DString &filename)
{
  // Similar to the Tcl command 'file nativename', this routine provides
  // an interface to the Tcl C function Tcl_TranslateFileName(), using
  // NB_DString's instead of Tcl_DString's.  It converts "filename" from
  // Tcl standard format (with /'s) to a platform specific form.  If the
  // call is successful, then filename is overwritten with the returned
  // result. Otherwise filename is left unchanged.

  Oc_AutoBuf ab(filename);
  Tcl_DString buf;
  Tcl_DStringInit(&buf); // Safety init
  const char *newname=Tcl_TranslateFileName(NULL,(char *)ab,&buf);
  if(newname!=NULL) {
    filename=newname;
    Tcl_DStringFree(&buf);
  }
}

// A slightly portable version of fopen
FILE *Nb_FOpen(const char *path,const char *mode)
{
  Nb_DString filename(path);
  Nb_TclNativeFilename(filename);
  return fopen(filename.GetStr(),mode);
}

int
Nb_Remove(const char *path)
{
    Nb_DString filename(path);
    Nb_TclNativeFilename(filename);
    return remove(filename.GetStr());
}

// Interfaces to Tcl_Write.  The first in intended mainly
// as an interface layer in case we decide in the future to
// use Tcl_WriteObj in place of Tcl_Write.  NB: The buffer
// in Nb_WriteChannel is 'char*', not 'const char*'.  Blame
// Tcl_Write for this.
UINT4m
Nb_WriteChannel(Tcl_Channel channel,const char* buf,UINT4m bytecount)
{ // NOTE: For consistency with fwrite-style interfaces, bytecount
  // is an unsigned type.  However, this routine may fail if bytecount
  // is too large to fit in an int, because Tcl_Write takes an int.

  // We'll just *ASSUME* that Tcl_Write doesn't _really_ modify buf.
  return static_cast<UINT4m>(Tcl_Write(channel,
				       const_cast<char*>(buf),
				       static_cast<int>(bytecount)));
}

UINT4m
Nb_FprintfChannel(Tcl_Channel channel,char* buf,UINT4m bufsize,
		      const char* format, ...)
{ // NOTE: For consistency with fwrite-style interfaces, bufsize
  // is an unsigned type.  However, this routine may fail if
  // bufsize is too large to fit in an int, because snprintf
  // returns an int, and Tcl_Write takes an int.

  // If buf==NULL, then use local buffer.
  static char* localbuf=NULL;
  static UINT4m localbuf_size=0;
  if(buf==NULL) {
    UINT4m worksize=bufsize;
    if(worksize<1024) worksize=1024; // Min localbuf size
    if(worksize>localbuf_size) {
      // Allocate new buffer
      if(localbuf!=NULL) delete[] localbuf;
      localbuf_size=worksize;
      localbuf = new char[localbuf_size];
    }
    buf=localbuf;
  }

  // Fill buf with formatted string
  va_list arg_ptr;
  va_start(arg_ptr,format);
  int len = Oc_Vsnprintf(buf,bufsize,format,arg_ptr);
  va_end(arg_ptr);
  if(UINT4m(len)>bufsize) len=static_cast<int>(bufsize);
  /// NOTE: According to the Linux sprintf man page, as of glibc 2.1 the
  /// return value is the number of bytes (excluding the trailing '\0')
  /// that would have been written if enough space had been available.

  if(len<1) return 0;

  // Write formatted string to channel
  return static_cast<UINT4m>(Tcl_Write(channel,buf,len));
}
